//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 102
// for test fourier changed
template<class T> void Function_102(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    // use resolution -resolution 10
    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m,IV()+1,10);
    e.reflection_bc=true;
    e.reflection_bc_flags=-1;
    e.Add_Walls(-1,MPM::BC_PERIODIC,0,0,false);
    T density = 2*e.unit_rho*e.scale_mass;
    T eps=1e-6;

    TV u;
    e.random.Fill_Uniform(u,-eps*(e.m/e.s),eps*(e.m/e.s));
    typedef MATRIX<T,2,2> M22;
    M22 F = M22(1,0,0,1);
    RANGE<TV> box(TV(0,0)*e.m,TV(10,10)*e.m);
    GRID<TV> seed_grid(e.grid.numbers_of_cells*2,e.grid.domain);
    e.Seed_Particles_Uniform(box,
        [&](const TV& X)
        {
            e.random.Fill_Uniform(u,-eps*(e.m/e.s),eps*(e.m/e.s));
            return u;
        },
        0,density,seed_grid,r0);
    for(int p=0;p<e.particles.X.m;p++)
    {
        M22 rF;
        e.random.Fill_Uniform(rF,-eps,eps);
        // e.random.Fill_Uniform(rF,-0.1,0.1);
        // std::cout<<F+rF<<"test F"<<std::endl;
        e.particles.F(p)=F+rF;
    }
    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);

    e.Add_Callbacks(false,"time-step",[&e]()
        {
            T scale=e.particles.V.Magnitude();
            e.particles.V/=scale*1000;
            for(auto&f:e.particles.F) f=(f-1)/(scale*1000)+1;
            LOG::printf("scale: %p\n",scale);
        });


}
template void Function_102(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_102(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
