//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Deformables/Constitutive_Models/COROTATED_FIXED.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 103
// for test fourier
template<class T> void Function_103(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    // use resolution -resolution 10
    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m,IV()+1,10);
    e.reflection_bc=true;
    e.reflection_bc_flags=-1;
    e.Add_Walls(-1,MPM::BC_PERIODIC,0,0,false);
    // e.Add_Walls(-1,MPM::BC_SEP,0,0,false);
    T density = 2*e.unit_rho*e.scale_mass;
    //input numbers
    T eps=1e-6;(void)eps;
    TV c = TV(3.174,2.542);// imag
    TV Z = TV(-1.57035,-1.27015);
    TV k=c;// imag

    T young = 1e3*e.unit_p*e.scale_E; T poissons = 0.3;
    COROTATED_FIXED<T,TV::m> corotated(young,poissons);
    std::cout<<"test c_mu"<<corotated.constant_mu<<" c_lambda "<<corotated.constant_lambda<<"mu"<<corotated.mu<<"lambda"<<corotated.lambda<<"\n";
    // get lambda and mu
    // T lam=Lambda(0); Tmu=Mu(0);
    T lam=corotated.constant_lambda; T mu=corotated.constant_mu;
    typedef MATRIX<T,2,2> M22;
//            T c_ms=c.Magnitude_Squared();
    // c is first eigenvector (lam+2*mu)ms(c)is eigenvalue
    // T eigval1 = (lam+2*mu)*c_ms;
    // k=c F=fk^H f=c
//            T lambda=(lam+2*mu)*c_ms;
    TV f = c;// real
    M22 F = Outer_Product(f,-k); // imag
    // second eigenvector? todo
    // get K
    M22 K = lam*F.Trace()+2*mu*F; // K=M:F K= lam*I*tr(F)+2*mu*F; imaginary
    // TODO: g
    T g=1;//TODO
    TV u = -g*K*c; // real
    // TV u = -g*(lam*F.Trace()+2*mu*F)*c; // real
    // TV u = -g*(lam*F.Trace()*c+2*mu*F*c); // real
    // TV u = -g*lam*F.Trace()*c-2*g*mu*F*c); // real
    // TV u = -g*lam*Outer_Product(f,-k).Trace()*c-2*g*mu*Outer_Product(f,-k)*c; // real
    // TV u = g*lam*f.Dot(k)*c+2*g*mu*f*k.Dot(c); // real
    // TV u = g*lam*c.Dot(c)*c+2*g*mu*f*c.Dot(c); // real
    // TV u = g*lambda*c; // real

    // g = dt*V0/(m*b*(Q-1))
    // d = -dt*g*b*Q/(w*(Q-1))
    // d = -g*b*dt*V0/(m*b*(Q-1))*m*b*Q/w/V0
    // -(d*V0)/(m*Q) = g^2

    std::cout << u << " get u test \n";
    RANGE<TV> box(TV(0,0)*e.m,TV(10,10)*e.m);
    GRID<TV> seed_grid(e.grid.numbers_of_cells*2,e.grid.domain);
    Z=TV(1,2)*pi*2;
    // u and F
    e.Seed_Particles_Uniform(box,
//                [&e](const TV& X){return eps*cos(X.Dot(Z))*u;},
        [=](const TV& X)
        {
            return eps*cos(X.Dot(Z))*u;
        },
        0,density,seed_grid,r0);
    for(int p=0;p<e.particles.X.m;p++)
    {
        e.particles.F(p)=eps*sin(e.particles.X(p).Dot(Z))*F;
    }
    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    // set Fp originally I   quantity with p on it- set them some where
    e.Add_Callbacks(false,"time-step",[&e]()
        {
            T scale=e.particles.V.Magnitude();
            e.particles.V/=scale*1000;
            for(auto&f:e.particles.F) f=(f-1)/(scale*1000)+1;
            LOG::printf("scale: %p\n",scale);
        });

}
template void Function_103(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_103(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
