//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 105
// two particle balls collide and collide
template<class T> void Function_105(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    // range, scale, resolution - range defines dx dy dz, scale*resolution = num of cell in direction
    e.Set_Grid(RANGE<TV>(TV(),TV(0.5,0.5))*e.m,IV()+1,50);
    e.Add_Walls(-1,MPM::BC_SEP,.49,.05*e.m,false);
    std::cout<<e.m<<"m\n";
    SPHERE<TV> sphere1(SPHERE<TV>(TV(0.15,0.15)*e.m,0.1*e.m));
    VECTOR<T,1> angular_velocity(0.5*e.scale_speed/e.s);
    T density = 2*e.unit_rho*e.scale_mass;
    e.Seed_Particles(sphere1,
        [=](const TV& X){return angular_velocity.Cross(X-sphere1.center);},
        [=](const TV&){return MATRIX<T,2>::Cross_Product_Matrix(angular_velocity);},
        density, e.particles_per_cell, r0);
    SPHERE<TV> sphere2(SPHERE<TV>(TV(0.3,0.3)*e.m,0.03*e.m));
    e.Seed_Particles(sphere2,[&e](const TV& X){return TV(-0.5*(e.m/e.s),-0.5*(e.m/e.s));},0,
        density,e.particles_per_cell,r0);
    //this velocity ball go towards one rotating ball
    ARRAY<int> mpm_particles(IDENTITY_ARRAY<>(e.particles.number));
    e.Add_Neo_Hookean(2*e.unit_p*e.scale_E,0.425,&mpm_particles);

}
template void Function_105(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_105(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
