//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 11
// freefall circle, rising ground
template<class T> void Function_11(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    SPHERE<TV> sphere(TV(.5,.5)*e.m,.3*e.m);
    T density=2*e.unit_rho*e.scale_mass;
    e.Seed_Particles(sphere,0,0,density,e.particles_per_cell,r0);
    ARRAY<int> foo(IDENTITY_ARRAY<>(e.particles.number));
    e.Add_Fixed_Corotated(1*e.unit_p*e.scale_E,0.3,&foo);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.8));
    e.Add_Walls(-1,MPM::BC_SEP,.3,.1*e.m,false);
    MPM_COLLISION_OBJECT<TV>* bottom=dynamic_cast<MPM_COLLISION_OBJECT<TV>*>(e.collision_objects(3));
    bottom->func_frame=[&e](T time){return FRAME<TV>(TV(0,(T).75-abs((T).15*time*e.scale_speed-(T).75)));};
    bottom->func_twist=[&e](T time){return TWIST<TV>(TV(0,-sign((T).15*time*e.scale_speed-(T).75)*(T).15*e.scale_speed),typename TV::SPIN());};
}
template void Function_11(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_11(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
