//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 111
// box smash static ball with gravity
template<class T> void Function_111(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    RANGE<TV> box(TV(-0,0)*e.m,TV(0.3,0.75)*e.m);
    T density=2*e.unit_rho*e.scale_mass;
    e.Seed_Particles(box,
        [&e](const TV X){return TV(-1,0);},
        0,density,e.particles_per_cell,r0);
    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    e.opt.dilation_only=true;
    e.particles.mu*=0;
    e.particles.mu0*=0;
    e.Add_Collision_Object(SPHERE<TV>(TV(-0.3,0.2)*e.m,.1*e.m),MPM::BC_SEP,.1);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-1.8));
    e.Add_Walls(-1,MPM::BC_NOSLIP,0,e.grid.dX.Min()*e.m,false);

}
template void Function_111(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_111(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
