//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Arrays/ARRAY.h>
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 115b
// balls with fluid with gravity // SS-FAIL simplified to only one ball
// command: cs && ./mpm -explicit -o output -last_frame 1 -particles_per_cell 50 -print_stats 115 | grep "grid total energy"
template<class T> void Function_115b(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    // cs && ./mpm -affine -order 3 -cfl_c .99 -last_frame 1000 -use_reflect -resolution 32 -explicit -threads 4 -o output -particles_per_cell 40 115 | grep velo
    LOG::printf("TEST");
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    // add liquid
    // RANGE<TV> box(TV(-0.8,0.75)*e.m,TV(0.8,0.9)*e.m);
    RANGE<TV> box(TV(-0.8,0.75)*e.m,TV(-0.6,0.9)*e.m);
    // T density=2*e.unit_rho*e.scale_mass;
    // e.Seed_Particles(box,0,
    //     // [&e](const TV X){return TV(0,.2);},
    //     0,density,e.particles_per_cell,r0);
    // e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    int particle_count = e.particles.number;
    // e.particles.mu*=0;
    // e.particles.mu0*=0;
    // e.dilation_only=true; // question: this also defines fluid/nonfluid how to separate between fluid and nonfluid  // doesn't matter with mu dilation or not

    // add more balls with random velocity.
    ARRAY<SPHERE<TV> > spheres;
    ARRAY<TV> v0;
    ARRAY<T> r;
    for(T i=-0.8;i<=-0.7;i+=0.2){
        // for(T j=0.65;j>=0.6;j-=0.2){
        for(T j=0.5;j>=0.5;j-=0.2){
            // for(T i=-0.8;i<=0.8;i+=0.2){
            //     for(T j=0.7;j>=0.3;j-=0.2){
            spheres.Append(SPHERE<TV>(TV(i,j)*e.m,0.04*e.m));
            spheres.Last().center.x+=4*e.grid.dX(0);
            T a = e.random.Get_Uniform_Number(-0.4,0.4);
            T b = e.random.Get_Uniform_Number(-0.4,0.4);
            v0.Append(TV(a,b)*(e.m/e.s));
            r.Append(0.03*e.m);
        }
    }
    for(int s=0;s<spheres.m;s++){
        SPHERE<TV>& sphere=spheres(s);
        T density=1010*e.unit_rho*e.scale_mass;
        int last=e.particles.number;
        e.Seed_Particles(sphere,[=,&e](const TV& X){return v0(s);},0,density,e.particles_per_cell,r0);
        for(int k=last;k<e.particles.number;k++){
            if((e.particles.X(k)-sphere.center).Magnitude_Squared()<sqr(r(s))){
                e.particles.deletion_list.Append(k);}}}// delete particles in mid circle
    e.particles.Delete_Elements_On_Deletion_List();
    ARRAY<int> affected_particles;
    for (int i=particle_count;i<e.particles.number;i++){
        affected_particles.Append(i);
    }
    e.Add_Neo_Hookean(0.073e9*e.unit_p*e.scale_E,0.4,&affected_particles);
    // e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-1.8));
    e.Add_Walls(-1,MPM::BC_SLIP,0.01,e.grid.dX.Min()*e.m,false);

}
template void Function_115b(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_115b(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
