//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 116
// mix of a lot of different fluid
template<class T> void Function_116(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    LOG::printf("FLAGS: %i %i\n",e.reflection_bc,e.reflection_bc_flags);
    int gap=e.reflection_bc?0:4;
    RANGE<TV> box(TV(e.grid.dX*gap-1*e.m),TV(1*e.m-e.grid.dX(0)*gap,0.1*e.m));
    T density=2*e.unit_rho*e.scale_mass;
    e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);
    ARRAY_VIEW<VECTOR<T,3> >* colors=e.particles.template Get_Array<VECTOR<T,3> >("color");
    T fluidCount = 10;
    for(int i=0;i<e.particles.X.m;i++)
        if(e.particles.X(i).y>box.Center().y){
            T count = 0;
            for(T j=box.Minimum_Corner().x;j<=box.Maximum_Corner().x;j+=box.Edge_Lengths().x/fluidCount){
                if(e.particles.X(i).x<j){
                    e.particles.mass(i)*=1+count;
                    (*colors)(i)=VECTOR<T,3>(1-count/fluidCount,count/fluidCount,0);
                    LOG::printf("TEST min %P max %P edg %P j %P mass %P count %P color %P \n",box.Minimum_Corner().x,box.Maximum_Corner().x,box.Edge_Lengths().x,j,e.particles.mass(i),count,(*colors)(i));
                    break;
                }
                count++;
            }
        }
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-1.8));
    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    e.opt.dilation_only=true;
    e.particles.mu*=0;
    e.particles.mu0*=0;
    if(e.reflection_bc) e.reflection_bc_flags=-1;
    e.Add_Walls(-1,MPM::BC_SEP,.2,e.grid.dX(0)*gap,false);

}
template void Function_116(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_116(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
