//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 117
// ball hits water
template<class T> void Function_117(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    LOG::printf("FLAGS: %i %i\n",e.reflection_bc,e.reflection_bc_flags);
    int gap=e.reflection_bc?0:4;
    RANGE<TV> box(TV(e.grid.dX*gap-1*e.m),TV(1*e.m-e.grid.dX(0)*gap,0));
    T density=2*e.unit_rho*e.scale_mass;
    e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-1.8));
    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    e.opt.dilation_only=true; // how to separate dilation from two particle groups , implement callback properly.
    e.particles.mu*=0;
    e.particles.mu0*=0;
    e.Seed_Particles(SPHERE<TV>(TV(0.8,0.8)*e.m,0.04*e.m),
        [&e](const TV& X){return TV(-1,-1);},
        0,density,e.particles_per_cell,r0);
    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    if(e.reflection_bc) e.reflection_bc_flags=-1;
    e.Add_Walls(-1,MPM::BC_SEP,.2,e.grid.dX(0)*gap,false);

}
template void Function_117(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_117(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
