//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 121
// two balls hanging from one spot
template<class T> void Function_121(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    e.Add_Walls(-1,MPM::BC_SEP,.3,.1*e.m,false);
    e.Add_Collision_Object(SPHERE<TV>(TV(0,0)*e.m,.1*e.m),MPM::BC_NOSLIP,0);
    SPHERE<TV> sphere1(TV(.5,0)*e.m,.06*e.m);
    T density=2*e.unit_rho*e.scale_mass;
    e.Seed_Particles(sphere1,[&e](const TV& X){return TV(0,1)*(e.m/e.s);},0,
        density,e.particles_per_cell,r0);
    SPHERE<TV> sphere2(TV(-0.5,0)*e.m,.06*e.m);
    e.Seed_Particles(sphere2,[&e](const TV& X){return TV(0,-1)*(e.m/e.s);},0,
        density,e.particles_per_cell,r0);
    RANGE<TV> box(TV(-0.5,-0.01)*e.m,TV(0.5,0.01)*e.m);
    e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);
    e.Add_Fixed_Corotated(20*e.unit_p*e.scale_E,0.3);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-1.8));

}
template void Function_121(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_121(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
