//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 124
// rotation of two balls
template<class T> void Function_124(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    SPHERE<TV> sphere(TV(.5,0)*e.m,.3*e.m);
    VECTOR<T,1> angular_velocity(0.4*e.scale_speed/e.s);
    T density=2*e.unit_rho*e.scale_mass;
    e.Seed_Particles(sphere,
        [=](const TV& X){return angular_velocity.Cross(X);},
        [=](const TV&){return MATRIX<T,2>::Cross_Product_Matrix(angular_velocity);},
        density,e.particles_per_cell,r0);
    SPHERE<TV> sphere2(TV(-.5,0)*e.m,.3*e.m);
    e.Seed_Particles(sphere2,
        [=](const TV& X){return angular_velocity.Cross(X);},
        [=](const TV&){return MATRIX<T,2>::Cross_Product_Matrix(angular_velocity);},
        density,e.particles_per_cell,r0);
    RANGE<TV> box(TV(-0.5,-0.01)*e.m,TV(0.5,0.01)*e.m);
    e.Seed_Particles(box,[=](const TV& X){return angular_velocity.Cross(X);},
        0,density,e.particles_per_cell,r0);
    T total_mass=e.particles.mass.Sum();
    TV total_momentum=e.particles.V.Weighted_Sum(e.particles.mass);
    TV dV=total_momentum/total_mass;
    e.particles.V-=dV; // make it no linear momentum only angular momentum
    T E=1e3*e.unit_p*e.scale_E;
    T nu=0.3;
    e.Add_Neo_Hookean(E,nu);

}
template void Function_124(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_124(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
