//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Meshes/SEGMENTED_CURVE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 127d
template<class T> void Function_127d(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    SEGMENTED_CURVE<TV>* sc=SEGMENTED_CURVE<TV>::Create();
    T density=1*e.unit_rho*e.scale_mass;
    int N=800;
    sc->particles.Add_Elements(N);
    sc->Update_Number_Nodes();
    for(int n=0;n<N;n++){
        TV X=TV();
        X(0)+=e.random.Get_Uniform_Number(-0.4,-0.2);
        X(1)+=e.random.Get_Uniform_Number(-0.4,-0.2);
        sc->particles.X(n)=X;
        int np1=(n!=N-1)?(n+1):0;
        sc->mesh.elements.Append(IV(n,np1));
    }
    e.Seed_Lagrangian_Particles(*sc,
        [&e](const TV& X){TV v;e.random.Fill_Uniform(v,-200,200); return v*(e.m/e.s);},0,density,r0);
    SEGMENTED_CURVE<TV>* sc2=SEGMENTED_CURVE<TV>::Create();
    sc2->particles.Add_Elements(N);
    sc2->Update_Number_Nodes();
    for(int n=0;n<N;n++){
        TV X=TV();
        X(0)+=e.random.Get_Uniform_Number(0.2,0.4);
        X(1)+=e.random.Get_Uniform_Number(0.2,0.4);
        sc2->particles.X(n)=X;
        int np1=(n!=N-1)?(n+1):0;
        sc2->mesh.elements.Append(IV(n,np1));
    }
    e.Seed_Lagrangian_Particles(*sc2,
        [&e](const TV& X){TV v;e.random.Fill_Uniform(v,-200,200); return v*(e.m/e.s);},0,density,r0);
    // e.Add_Fixed_Corotated(10000*e.unit_p*e.scale_E,0.001);
    e.Add_Fixed_Corotated(100000000*e.unit_p*e.scale_E,0.001);
    // e.Add_Fixed_Corotated(100*e.unit_p*e.scale_E,0.001);
    // e.Add_Fixed_Corotated(10*e.unit_p*e.scale_E,0.49);
    e.Add_Walls(-1,MPM::BC_SLIP,.2,e.grid.dX(0),false);
}
template void Function_127d(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_127d(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
