//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 129b
template<class T> void Function_129b(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    T vel=0.07;
    T density=1000*e.unit_rho*e.scale_mass;
    T volume=e.grid.dX.Product()/e.particles_per_cell;
    T mass=density*volume;
    for(T theta=0;theta<4*pi;theta+=0.01){
        for(T bias=0;bias<0.04;bias+=0.01){
            e.Add_Particle(TV(vel*theta*cos(theta)+bias*cos(theta),vel*theta*sin(theta)+bias*sin(theta)),0,0,mass,volume,r0);
        }
    }
    T rot_col=-0.1*M_PI;
    e.Add_Collision_Object(RANGE<TV>(TV(-.15,-.15),TV(.15,.15))*e.m,MPM::BC_NOSLIP,1
        , [&e,rot_col](T time){return FRAME<TV>(ROTATION<TV>::From_Angle(rot_col*time));},
        [&e,rot_col](T time){return TWIST<TV>(TV(),typename TV::SPIN(rot_col));}); // rotating collision object
    e.Add_Fixed_Corotated(1000*e.unit_p*e.scale_E,0.3);
    e.Add_Walls(-1,MPM::BC_SLIP,0.2,0,false);
}
template void Function_129b(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_129b(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
