//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/ORIENTED_BOX.h>
#include <Geometry/Geometry_Particles/DEBUG_PARTICLES.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 18
// friction test
template<class T> void Function_18(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    T angle=e.extra_T.m>0?e.extra_T(0):0.1;
    T initial_velocity=e.extra_T.m>1?e.extra_T(1)*(e.m/e.s):1*(e.m/e.s);
    T coefficient_of_friction=e.extra_T.m>2?e.extra_T(2):0.3;
    T g=9.8*e.m/(e.s*e.s);
    if(!e.no_regular_seeding) e.regular_seeding=true;
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    RANGE<TV> box(TV(-.8,0)*e.m,TV(-.2,.2)*e.m);
    T density=e.unit_rho*e.scale_mass;
    e.Seed_Particles(box,[=,&e](const TV& X){return TV(initial_velocity,0);},0,density,e.particles_per_cell,r0);
    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    FRAME<TV> frame(TV(),ROTATION<TV>::From_Angle(-angle));
    ORIENTED_BOX<TV> ground(RANGE<TV>(TV(-5,-5),TV(5,0))*e.m,frame);
    e.Add_Collision_Object(ground,MPM::BC_SLIP,coefficient_of_friction);
    for(int i=0;i<e.particles.X.m;i++){
        e.particles.X(i)=frame*e.particles.X(i);
        e.particles.V(i)=frame.r.Rotate(e.particles.V(i));}
    e.Add_Gravity(TV(0,-g));
    e.write_output_files.Append([=,&e]()
        {
            T c=sin(angle)*g;
            T d=coefficient_of_friction*cos(angle)*g;
            T v0=initial_velocity;
            T v,x;
            T t=e.time;
            T st=FLT_MAX;
            T a=c-sign(v0)*d;
            if(v0*a<0) st=-v0/a;
            if(t<st){
                v=v0+a*t;
                x=v0*t+.5*a*t*t;}
            else{
                x=.5*v0*st;
                if(abs(c)>d){
                    a=c+sign(v0)*d;
                    t-=st;
                    v=a*t;
                    x+=.5*a*t*t;}
                else v=0;}
            Add_Arrow(frame*(box.max_corner+TV(x,0)),frame.r.Rotate(TV(v,0)),VECTOR<T,3>(1,0,0));
        });
}
template void Function_18(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_18(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
