//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 20
template<class T> void Function_20(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>(TV(0,0),TV(4,6))*e.m,IV(2,3));
    int jet_freq=3;
    T init_vel=7*e.m;
    auto func=[&e,jet_freq,init_vel,r0](int frame)
        {
            if (frame%jet_freq==0){
                int old_m=e.particles.X.m;
                T density=2*e.unit_rho*e.scale_mass;
                RANGE<TV> box1(TV(.1,3.75)*e.m,TV(.3,3.77)*e.m), box2(TV(.1,3.5)*e.m,TV(.3,3.57));
                e.Seed_Particles(box1,[=](const TV& X){return TV(init_vel,0);},0,
                    density,e.particles_per_cell,r0);
                e.Seed_Particles(box2,[=](const TV& X){return TV(init_vel,0);},0,
                    density,e.particles_per_cell,r0);
                box1+=RANGE<TV>(TV(3.6,0),TV(3.6,0))*e.m;
                box2+=RANGE<TV>(TV(3.6,0),TV(3.6,0))*e.m;
                e.Seed_Particles(box1,[=](const TV& X){return TV(-init_vel,0);},0,
                    density,e.particles_per_cell,r0);
                e.Seed_Particles(box2,[=](const TV& X){return TV(-init_vel,0);},0,
                    density,e.particles_per_cell,r0);
                ARRAY<int> new_particles(IDENTITY_ARRAY<>(e.particles.X.m-old_m)+old_m);
                e.Set_Lame_On_Particles(1e3*e.unit_p*e.scale_E,.3,&new_particles);
                e.particles.mu*=0;
                e.particles.mu0*=0;
            }
        };
    e.begin_frame.Append(func);
    e.Add_Collision_Object(RANGE<TV>(TV(0.,3.74),TV(.7,3.75))*e.m,MPM::BC_SLIP,0.);
    e.Add_Collision_Object(RANGE<TV>(TV(1.3,3.49),TV(2.,3.5))*e.m,MPM::BC_SLIP,0.);
    e.Add_Collision_Object(RANGE<TV>(TV(1.3,3.74),TV(2.,3.75))*e.m,MPM::BC_SLIP,0.);
    e.Add_Collision_Object(RANGE<TV>(TV(0.,3.49),TV(.7,3.5))*e.m,MPM::BC_SLIP,0.);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-1.8));
    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    e.Add_Walls(-1,MPM::BC_SLIP,.3,.1*e.m,false);
}
template void Function_20(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_20(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
