//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 31
// Light fluid on heavy fluid
template<class T> void Function_31(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    RANGE<TV> box(e.grid.dX*4,TV(1*e.m-e.grid.dX(0)*4,0.5*e.m));
    T density=2*e.unit_rho*e.scale_mass;
    e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);
    ARRAY_VIEW<VECTOR<T,3> >* colors=e.particles.template Get_Array<VECTOR<T,3> >("color");
    for(int i=0;i<e.particles.X.m;i++)
        if(e.particles.X(i).y>box.Center().y)
        {
            e.particles.mass(i)*=.5;
            (*colors)(i)=VECTOR<T,3>(1,0,0);
        }
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-1.8));
    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    e.opt.dilation_only=true;
    e.particles.mu*=0;
    e.particles.mu0*=0;
    e.Add_Walls(-1,MPM::BC_SEP,.2,e.grid.dX(0)*4,false);
}
template void Function_31(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_31(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
