//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 32
// colliding
template<class T> void Function_32(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>(TV(-1,-1),TV(2,2))*e.m);
    T d_small_1=1*e.unit_rho*e.scale_mass,d_small_2=1*e.unit_rho*e.scale_mass,d_large_1=d_small_1*.9,d_large_2=d_small_2*.9;
    SPHERE<TV> large1(TV(.2,.5)*e.m,.1*e.m);
    e.Seed_Particles(large1,[&e](const TV& X){return TV(0.75,0)*(e.m/e.s);},0,d_large_1,e.particles_per_cell,r0);
    SPHERE<TV> large2(TV(.6,.5)*e.m,.1*e.m);
    e.Seed_Particles(large2,0,0,d_large_2,e.particles_per_cell,r0);
    for(int k=0;k<e.particles.number;k++)
        if((e.particles.X(k)-large1.center).Magnitude_Squared()<sqr(large1.radius*0.6)
            || (e.particles.X(k)-large2.center).Magnitude_Squared()<sqr(large2.radius*0.6))
            e.particles.deletion_list.Append(k);
    ARRAY_VIEW<VECTOR<T,3> >* color_attribute=e.particles.template Get_Array<VECTOR<T,3> >("color");
    for(int i=0;i<e.particles.X.m;i++) (*color_attribute)(i)=VECTOR<T,3>(0,1,1);
    e.particles.Delete_Elements_On_Deletion_List();
    ARRAY<int> mpm_particles(IDENTITY_ARRAY<>(e.particles.number));
    e.Add_Neo_Hookean(2*e.unit_p*e.scale_E,0.425,&mpm_particles);
    int old_pn=e.particles.number;
    SPHERE<TV> small1(large1.center,large1.radius*.6);
    e.Seed_Particles(small1,[&e](const TV& X){return TV(0.75,0)*(e.m/e.s);},0,d_small_1,e.particles_per_cell,r0);
    SPHERE<TV> small2(large2.center,large2.radius*.6);
    e.Seed_Particles(small2,0,0,d_small_2,e.particles_per_cell,r0);
    ARRAY<int> mpm_particles2;for(int i=old_pn;i<e.particles.number;i++) mpm_particles2.Append(i);
    e.Add_Neo_Hookean(2000*e.unit_p*e.scale_E,0.3,&mpm_particles2);
    for(int i=old_pn;i<e.particles.X.m;i++) (*color_attribute)(i)=VECTOR<T,3>(1,1,1);
}
template void Function_32(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_32(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
