//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Deformables/Constitutive_Models/COROTATED_FIXED.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 33
// sand box drop
template<class T> void Function_33(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.particles.Store_Fp(true);

    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    e.Add_Collision_Object(RANGE<TV>(TV(-0.5,-1),TV(1.5,.1))*e.m,MPM::BC_SEP,10);

    T density=(T)1281*e.unit_rho*e.scale_mass;
    T E=5000*e.unit_p*e.scale_E,nu=.4;
    if(!e.use_theta_c) e.theta_c=0.01;
    if(!e.use_theta_s) e.theta_s=.00001;
    if(!e.use_hardening_factor) e.hardening_factor=80;
    if(!e.use_max_hardening) e.max_hardening=5;
    e.Add_Clamped_Plasticity(*new COROTATED_FIXED<T,TV::m>(E,nu),e.theta_c,e.theta_s,e.max_hardening,e.hardening_factor,0);
    RANGE<TV> box(TV(.45,.11)*e.m,TV(.55,.31)*e.m);
    e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.8));
}
template void Function_33(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_33(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
