//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/ORIENTED_BOX.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include <Deformables/Constitutive_Models/COROTATED_FIXED.h>
#include "../STANDARD_TESTS_BASE.h"
#include <fstream>
namespace PhysBAM{
// TEST-NAME 35
// snow wedge
template<class T> void Function_35(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    // ./mpm 35 -flip 0.95 -max_dt .005 -cfl .1 -resolution 200
    e.particles.Store_Fp(true);

    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    ORIENTED_BOX<TV> wedge(RANGE<TV>(TV(),TV(0.2,0.2))*e.m,ROTATION<TV>::From_Angle(0.25*M_PI),TV(0.5,0.4-sqrt(2.0)*0.1));
    RANGE<TV> ground(TV(-1,0)*e.m,TV(2,0.1)*e.m);
    if(e.use_penalty_collisions){
        e.Add_Penalty_Collision_Object(wedge);
        e.Add_Penalty_Collision_Object(ground);}
    else{
        e.Add_Collision_Object(ground,MPM::BC_SEP,1);
        e.Add_Collision_Object(new ANALYTIC_IMPLICIT_OBJECT<ORIENTED_BOX<TV> >(wedge),MPM::BC_SEP,1);}

    T density=(T)2*e.unit_rho*e.scale_mass;
    int number_of_particles=20000;
    T E=40*e.unit_p*e.scale_E,nu=.2;
    if(!e.use_theta_c) e.theta_c=0.015;
    if(!e.use_theta_s) e.theta_s=.005;
    if(!e.use_hardening_factor) e.hardening_factor=7;
    e.Add_Clamped_Plasticity(*new COROTATED_FIXED<T,TV::m>(E,nu),e.theta_c,e.theta_s,e.max_hardening,e.hardening_factor,0);
    RANGE<TV> box(TV(.3,.7)*e.m,TV(.7,.9)*e.m);
    std::ifstream ifs("particles.dat");
    if(ifs.is_open()){
        std::string foobar;
        ifs>>foobar>>foobar>>foobar>>foobar;
        T x,y,mass,volume;
        while(ifs>>x>>y>>mass>>volume && !ifs.eof())
            e.Add_Particle(TV(x,y),0,0,mass,volume,r0);
        PHYSBAM_ASSERT(e.particles.number==number_of_particles);}
    else{
        PHYSBAM_WARNING("Couldn't open 'particles.dat'. Falling back to using random particle positions.");
        e.Seed_Particles(box,0,0,density,
            number_of_particles*e.grid.dX.Product()/box.Size(),r0);}
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-2));
}
template void Function_35(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_35(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
