//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Arrays/ARRAY.h>
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 4
// colliding of two rings
template<class T> void Function_4(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>(TV(),TV(0.48,0.48))*e.m,IV()+1,48);
    ARRAY<SPHERE<TV> > spheres;
    ARRAY<TV> v0;
    ARRAY<T> r;
    spheres.Append(SPHERE<TV>(TV(0.1,0.24)*e.m,0.04*e.m));
    spheres.Append(SPHERE<TV>(TV(0.4,0.24)*e.m,0.04*e.m));
    v0.Append(TV(50,0)*(e.m/e.s));
    v0.Append(TV(-50,0)*(e.m/e.s));
    r.Append(0.03*e.m);
    r.Append(0.03*e.m);
    for(int s=0;s<spheres.m;s++){
        SPHERE<TV>& sphere=spheres(s);
        T density=1010*e.unit_rho*e.scale_mass;
        int last=e.particles.number;
        e.Seed_Particles(sphere,[=](const TV& X){return v0(s);},0,density,e.particles_per_cell,r0);
        for(int k=last;k<e.particles.number;k++){
            if((e.particles.X(k)-sphere.center).Magnitude_Squared()<sqr(r(s))){
                e.particles.deletion_list.Append(k);}}}
    e.particles.Delete_Elements_On_Deletion_List();
    e.Add_Neo_Hookean(0.073e9*e.unit_p*e.scale_E,0.4);
}
template void Function_4(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_4(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
