//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Deformables/Constitutive_Models/COROTATED_FIXED.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 49a 49b 49c 49d 49e 49f 49g 49h 49i 49j
// sand column collapse (drucker prager v.s. snow-style)
template<class T> void Function_49(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    //  ./mpm 42 -use_exp_F -max_dt 7.5e-4 -scale_E 0.1 -resolution 100 -T 0(1)
    // fooT1=1:  use drucker prager
    // fooT1=0:  use snow-style plasticity
    e.particles.Store_Fp(true);
    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    if(e.use_penalty_collisions) e.Add_Penalty_Collision_Object(RANGE<TV>(TV(-0.5,-1),TV(1.5,.1))*e.m,0.9);
    else e.Add_Collision_Object(RANGE<TV>(TV(-0.5,-1),TV(1.5,.1))*e.m,MPM::BC_NOSLIP,0);
    T l0=0.05*e.m;
    T h0=l0*8;
    T gap=e.grid.dX(1)*0.01;
    T density=(T)2200*e.unit_rho*e.scale_mass;
    T E=35.37e6*e.unit_p*e.scale_E,nu=.3;
    int test_num=e.test_name[2]-'a';
    RANGE<TV> box(TV(.5*e.m-l0,.1*e.m+gap),TV(.5*e.m+l0,.1*e.m+gap+h0));
    if(e.extra_T(0)){
        if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
        e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);
        e.Add_Drucker_Prager_Case(E,nu,test_num);}
    else{
        if(!e.use_theta_c) e.theta_c=0.015;
        if(!e.use_theta_s) e.theta_s=.000001;
        if(!e.use_hardening_factor) e.hardening_factor=20;
        e.Add_Clamped_Plasticity(*new COROTATED_FIXED<T,TV::m>(E,nu),e.theta_c,e.theta_s,e.max_hardening,e.hardening_factor,0);
        e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);}
    e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);
    e.Add_Drucker_Prager_Case(E,nu,test_num);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.81));
}
template void Function_49(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_49(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
