//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 5
// rebound of an elastic cylinder
template<class T> void Function_5(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    if(!e.user_resolution) e.resolution=10;
    T dx=(T)5/e.resolution*e.m;
    e.Set_Grid(RANGE<TV>(TV(),TV(15,5)*e.m).Thickened(dx*(T)4.5),IV(3,1),IV()+9,1,10);
    e.Add_Collision_Object(RANGE<TV>(TV(-5,-5)*e.m,TV(0+dx/2,15*e.m)),MPM::BC_SLIP,0);
    e.Add_Collision_Object(RANGE<TV>(TV(15*e.m-dx/2,-5*e.m),TV(20,15)*e.m),MPM::BC_SLIP,0);
    SPHERE<TV> sphere(TV(2.5,2.5)*e.m,1.5*e.m);
    T density=4*e.unit_rho*e.scale_mass;
    e.Seed_Particles(sphere,[&e](const TV& X){return TV(0.5*e.m/e.s,0);},0,
        density,e.particles_per_cell,r0);
    e.Add_Neo_Hookean(85.5*e.unit_p*e.scale_E,0.425); //solve({E/(2*(1+r))=30,E*r/((1+r)*(1-2*r))=170},{E,r});
}
template void Function_5(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_5(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
