//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Hybrid_Methods/Forces/MPM_VISCOSITY.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 52a 52b
template<class T> void Function_52(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
//lambda particles
    //usage:./mpm 51 -threads 8 -use_exp_F -max_dt 1e-3 -resolution 100 -scale_E 10 -T 10 -T 1000 -T 4 -last_frame 20
    PHYSBAM_ASSERT(e.extra_T(3)<=e.particles_per_cell,"You can't have more water particles than total number of particles.");
    e.particles.Store_Fp(true);
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
    //e.particles.Store_Plastic_Deformation(true);
    e.Set_Grid(2*RANGE<TV>::Unit_Box()*e.m);
    if(e.use_penalty_collisions){
        e.Add_Penalty_Collision_Object(RANGE<TV>(TV(-0.5,-1),TV(2.5,.1))*e.m,0.9);
        e.Add_Penalty_Collision_Object(RANGE<TV>(TV(-0.5,-1),TV(.1,1.5))*e.m,0.9);}
    else{
        e.Add_Collision_Object(RANGE<TV>(TV(-0.5,-1),TV(2.5,.1))*e.m,MPM::BC_NOSLIP,0);
        e.Add_Collision_Object(RANGE<TV>(TV(-0.5,-1),TV(.1,1.5))*e.m,MPM::BC_NOSLIP,0);}

    T density=(T)2200*e.unit_rho*e.scale_mass;
    T E=35.37e5*e.unit_p*e.scale_E,nu=.3;
    T gap=e.grid.dX(1)*0.01;
    RANGE<TV> box(TV(.1*e.m+gap,.1*e.m+gap),TV(.3,.75)*e.m);
    //seed sand particles
    if(e.extra_T(3)<e.particles_per_cell)
        e.Seed_Particles(box,0,0,density,e.particles_per_cell-e.extra_T(3),r0);
    ARRAY<int> sand_particles(e.particles.X.m);
    for(int p=0;p<e.particles.X.m;p++) sand_particles(p)=p;
    e.Add_Drucker_Prager(E,nu,(T)35,&sand_particles);
    //seed lambda particles
    if(e.test_name=="51b"){
        T El=500*e.extra_T(0),nul=.1*e.extra_T(2);
        e.Add_Lambda_Particles(&sand_particles,El,nul,e.extra_T(1),r0,true);}
    else if(e.test_name=="52a" && e.extra_T(3)!=0){
        T El=500*e.extra_T(0),nul=.1*e.extra_T(2);
        T lambdal=El*nul/((1+nul)*(1-2*nul));
        int ns=e.particles.X.m;
        e.Seed_Particles(box,0,0,e.extra_T(1),e.extra_T(3),r0);
        ARRAY<int> lambda_particles(e.particles.X.m-ns);
        ARRAY_VIEW<VECTOR<T,3> >* color_attribute=e.particles.template Get_Array<VECTOR<T,3> >("color");
        //fix volume and mass
        T volume=e.grid.dX.Product()/e.particles_per_cell;
        T mass_sand=density*volume;
        T mass_lambda=e.extra_T(1)*volume;
        for(int p=0;p<ns;p++){
            e.particles.mass(p)=mass_sand;
            e.particles.volume(p)=volume;}
        for(int k=0;k<lambda_particles.m;k++){
            int p=ns+k;
            lambda_particles(k)=p;
            e.particles.mass(p)=mass_lambda;
            e.particles.volume(p)=volume;
            e.particles.mu(p)=(T)0;
            e.particles.mu0(p)=(T)0;
            e.particles.lambda(p)=lambdal;
            e.particles.lambda0(p)=lambdal;
            (*color_attribute)(p)=VECTOR<T,3>(0,0,1);}
        e.Add_Fixed_Corotated(El,nul,&lambda_particles,true);}

    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.81));
    if(e.extra_T.m>=2){
        MPM_VISCOSITY<TV>* visc=new MPM_VISCOSITY<TV>(e.force_helper,e.gather_scatter,0,0*e.unit_mu);
        visc->Use_Variable_Viscosity();
        for(int i=0;i<e.particles.X.m;i++)
            visc->viscosity(i)=e.extra_T(e.particles.X(i).y>.4);
        e.Add_Force(*visc);}
}
template void Function_52(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_52(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
