//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 55
// Moving collision object
template<class T> void Function_55(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.particles.Store_Fp(true);

    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    if(e.use_penalty_collisions)
        PHYSBAM_NOT_IMPLEMENTED();
    e.Add_Collision_Object(RANGE<TV>(TV(0,.2),TV(1,.3))*e.m,MPM::BC_NOSLIP,0,
        [&e](T time){return FRAME<TV>(TV(0,max(0.2/e.s*(time-.4),0.0))*e.m);},
        [&e](T time){return TWIST<TV>(TV(0,0.2/e.s*(time-.4)>0?0.2*e.s:0)*e.m,typename TV::SPIN());});
    T density=(T)2200*e.unit_rho*e.scale_mass;
    T E=1e5*e.unit_p*e.scale_E,nu=.4;
    e.Add_Fixed_Corotated(E,nu);
    T l0=0.05*e.m;
    T h0=l0*8;
    T gap=e.grid.dX(1)*0.01;
    RANGE<TV> box(TV(.5*e.m-l0,.3*e.m+gap),TV(.5*e.m+l0,.3*e.m+gap+h0));
    e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.81));
}
template void Function_55(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_55(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
