//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 59
// sand falling into a pile.
template<class T> void Function_59(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.particles.Store_Fp(true);
    e.Set_Grid(RANGE<TV>(TV(-2,0),TV(2,1))*e.m,IV(4,1));
    RANGE<TV> ground(TV(-10,-10)*e.m,TV(10,.1)*e.m);
    if(e.use_penalty_collisions) e.Add_Penalty_Collision_Object(ground);
    else e.Add_Collision_Object(ground,MPM::BC_SLIP,0.3);
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
    int case_num=e.use_hardening_mast_case?e.hardening_mast_case:2;
    T E=1e4*e.unit_p*e.scale_E,nu=.3;
    e.Add_Drucker_Prager_Case(E,nu,case_num);
    TV spout(0,1*e.m);
    T density=(T)2200*e.unit_rho*e.scale_mass;
    T spout_width=.05*e.m;
    T pour_speed=.4*e.m/e.s;
    TV gravity=TV(0,-9.8*e.m/(e.s*e.s));
    e.Add_Source(spout,TV(0,-1),spout_width/2,pour_speed,gravity,density,
        E,nu,0,e.extra_T(2),r0);
    e.Add_Gravity(gravity);
}
template void Function_59(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_59(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
