//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 6
// skew impact of two elastic cylinders
template<class T> void Function_6(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    if(!e.user_resolution) e.resolution=12;
    int third_resolution=(e.resolution+2)/3;
    T dx=(T)4/third_resolution;
    e.Set_Grid(RANGE<TV>(TV(),TV(20,12)*e.m).Thickened(dx*(T)4.5),IV(5,3),IV()+9,3,12);
    T density=5*e.unit_rho*e.scale_mass;
    SPHERE<TV> sphere1(TV(3,3)*e.m,2*e.m);
    e.Seed_Particles(sphere1,[&e](const TV& X){return TV(0.75*(e.m/e.s),0);},0,density,e.particles_per_cell,r0);
    SPHERE<TV> sphere2(TV(16,5)*e.m,2*e.m);
    e.Seed_Particles(sphere2,[&e](const TV& X){return TV(-0.75*(e.m/e.s),0);},0,density,e.particles_per_cell,r0);
    e.Add_Neo_Hookean(31.685*e.unit_p*e.scale_E,0.44022); //solve({E/(2*(1+r))=11,E*r/((1+r)*(1-2*r))=81},{E,r});
}
template void Function_6(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_6(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
