//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 60
template<class T> void Function_60(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
//cohesion sanity check
    e.particles.Store_Fp(true);

    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    if(e.use_penalty_collisions)
        e.Add_Penalty_Collision_Object(RANGE<TV>(TV(-0.5,-1),TV(1.5,.1))*e.m,0.9);
    else
        e.Add_Collision_Object(RANGE<TV>(TV(-0.5,-1),TV(1.5,.1))*e.m,MPM::BC_NOSLIP,0);
    T density=(T)2200*e.unit_rho*e.scale_mass;
    T E=35.37e6*e.unit_p*e.scale_E,nu=.3;
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
    T l0=0.05*e.m;
    T h0=l0*8;
    T gap=e.grid.dX(1)*0.01;
    RANGE<TV> box(TV(.5*e.m-l0,.1*e.m+gap),TV(.5*e.m+l0,.1*e.m+gap+h0));
    e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);
    ARRAY<int> sand_particles(e.particles.X.m);
    for(int p=0;p<e.particles.X.m;p++) sand_particles(p)=p;
    if(!e.use_cohesion) e.sigma_Y=1;
    e.Add_Drucker_Prager(E,nu,(T)35,0,0,0,&sand_particles,false,e.sigma_Y);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.81));
}
template void Function_60(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_60(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
