//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_UNION.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 61a 61b
// sand cup pull
template<class T> void Function_61(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    // ./mpm 54 -threads 10 -use_exp_F -max_dt 7.5e-4  -resolution 90 -last_frame 200
    e.particles.Store_Fp(true);
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
    e.Set_Grid(RANGE<TV>(TV(-1,-0.5),TV(1,0.5))*e.m,IV(2,1));
    RANGE<TV> cupbottom(TV(-0.3,-0.25),TV(0.3,-0.2));
    RANGE<TV> cupleft(TV(-0.3,-0.25),TV(-0.25,0.25));
    RANGE<TV> cupright(TV(0.25,-0.25),TV(0.3,0.25));
    e.Add_Collision_Object(
        new IMPLICIT_OBJECT_UNION<TV>(
            new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(cupbottom),
            new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(cupleft),
            new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(cupright)),MPM::BC_SEP,0);
    e.Add_Walls(-1,MPM::BC_NOSLIP,0,0.04,false);
    T density=(T)2200*e.unit_rho*e.scale_mass;
    T E=35.37e5*e.unit_p*e.scale_E,nu=.3;
    ARRAY<RANGE<TV> > columns;
    columns.Append(RANGE<TV>(TV(-0.24,-0.2),TV(-0.2,0.1)));
    columns.Append(RANGE<TV>(TV(-0.15,-0.2),TV(-0.05,-0.1)));
    columns.Append(RANGE<TV>(TV(-0.02,-0.2),TV(0.1,0.02)));
    columns.Append(RANGE<TV>(TV(0.12,-0.2),TV(0.15,0.12)));
    columns.Append(RANGE<TV>(TV(0.18,-0.2),TV(0.23,-0.15)));
    for(int k=0;k<columns.m;k++) e.Seed_Particles(columns(k),0,0,density,e.particles_per_cell,r0);
    ARRAY<int> sand_particles(e.particles.X.m);
    for(int p=0;p<e.particles.X.m;p++) sand_particles(p)=p;
    e.Add_Drucker_Prager(E,nu,(T)35,&sand_particles,false,e.test_name=="61b"?e.sigma_Y:0);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.81));
    ARRAY_VIEW<VECTOR<T,3> >* color_attribute=e.particles.template Get_Array<VECTOR<T,3> >("color");
    for(int i=0;i<e.particles.X.m;i++) (*color_attribute)(i)=VECTOR<T,3>(.8,.7,.7);
    //Add water particles for case 58
    if(e.test_name=="61b"){
        if(e.extra_T.m<=3) e.extra_T(3)=(T)1;
        e.Add_Lambda_Particles(&sand_particles,E*e.extra_T(3),nu,(T)1000*e.unit_rho,r0,true,(T)0.3,(T)1);}
    auto func=[&e](int frame){if(frame==10) e.Add_Gravity(TV(0,20));};
    e.begin_frame.Append(func);
}
template void Function_61(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_61(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
