//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/HOURGLASS.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_INTERSECTION.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_INVERT.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 62
// hourglass
template<class T> void Function_62(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.particles.Store_Fp(true);
    e.Set_Grid(RANGE<TV>(TV(-0.2,-0.45)*e.m,TV(0.2,0.45)*e.m),IV(4,9),IV(),4);
    LOG::cout<<"GRID DX: " <<e.grid.dX<<std::endl;
    IMPLICIT_OBJECT<TV>* hg=new ANALYTIC_IMPLICIT_OBJECT<HOURGLASS<TV> >(HOURGLASS<TV>(TV::Axis_Vector(1),TV(),(T).16,(T).0225,(T).8,(T).0225));
    IMPLICIT_OBJECT<TV>* inv=new IMPLICIT_OBJECT_INVERT<TV>(hg);
    if(e.use_penalty_collisions) e.Add_Penalty_Collision_Object(inv);
    else e.Add_Collision_Object(inv,MPM::BC_SEP,.3);
    T density=(T)2200*e.unit_rho*e.scale_mass;
    T E=35.37e6*e.unit_p*e.scale_E,nu=.3;
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
    RANGE<TV> fill_part=e.grid.domain;
    fill_part.min_corner.y=0;
    fill_part.max_corner.y=.1;
    ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> > io_fill_part(fill_part);
    IMPLICIT_OBJECT_INTERSECTION<TV> ioi(&io_fill_part,hg);
    ioi.owns_io.Fill(false);
    e.Seed_Particles(ioi,0,0,density,e.particles_per_cell,r0);
    LOG::printf("added %i particles.",e.particles.X.m);
    e.Add_Drucker_Prager_Case(E,nu,2);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.81));
}
template void Function_62(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_62(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
