//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Geometry_Particles/DEBUG_PARTICLES.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 69
// Reflection bc test
template<class T> void Function_69(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>(TV(),TV(4,1))*e.m,IV(4,1));
    // e.reflection_bc=true;
    // e.use_full_reflection=true;
    T E=1e4*e.unit_p*e.scale_E,nu=.3;
    T density=(T)2*e.unit_rho*e.scale_mass;
    TV gravity=TV(0,-9.8*e.m/(e.s*e.s));
    T vel=1*e.m/e.s;
    RANGE<TV> box(TV(.25,0)*e.m,TV(1.25,.5)*e.m);
    if(!e.reflection_bc)
    {
        box.min_corner.y+=3*e.grid.dX.Min();
        box.max_corner.y+=3*e.grid.dX.Min();
    }
    VECTOR<TV,4> C;
    box.Corners(C);
    e.Seed_Particles(box,[=](const TV& X){return TV(vel,0);},0,density,e.particles_per_cell,r0);
    e.Add_Fixed_Corotated(E,nu);
    e.Add_Gravity(gravity);
    T friction=.02;
    e.Add_Walls(-1,MPM::BC_SEP,friction,3*e.grid.dX.Min(),false);
    T a=friction*-gravity.y;
    PHYSBAM_ASSERT(vel>=0);
    e.write_output_files.Append([&e,vel,a,C]()
        {
            TV v,x;
            if(a*e.time<=vel)
            {
                v.x=vel-a*e.time;
                x.x=vel*e.time-a*e.time*e.time/2;
            }
            else x.x=vel*vel/(a*2);
            Add_Segment(C(0)+x,C(1)+x,VECTOR<T,3>(1,0,0));
            Add_Segment(C(0)+x,C(2)+x,VECTOR<T,3>(1,0,0));
            Add_Segment(C(3)+x,C(1)+x,VECTOR<T,3>(1,0,0));
            Add_Segment(C(3)+x,C(2)+x,VECTOR<T,3>(1,0,0));
        });

}
template void Function_69(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_69(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
