//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/ORIENTED_BOX.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 72
template<class T> void Function_72(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>(TV(),TV(1,2))*e.m,IV(1,2));
    T density=2*e.unit_rho*e.scale_mass;
    T gravity=9.8*e.m/(e.s*e.s);
    T half_edge=.05*e.m;
    int seed_freq=4;
    int ob_per_frame=2;
    RANGE<TV> seed_box=e.grid.domain.Thickened(-.1*e.m);
    seed_box.min_corner.y=seed_box.max_corner.y-sqr(seed_freq*e.frame_dt)/2*gravity;
    seed_box=seed_box.Thickened(-sqrt((T)TV::m)*half_edge*e.m);
    seed_box.max_corner.x=seed_box.min_corner.x+seed_box.Edge_Lengths().x/ob_per_frame;
    auto func=[&e,ob_per_frame,seed_freq,half_edge,density,seed_box,r0](int frame)
        {
            int num_part=e.particles.X.m;
            if(frame%seed_freq==0 && frame<200){
                for(int i=0;i<ob_per_frame;i++){
                    RANGE<TV> obj=RANGE<TV>::Centered_Box()*half_edge*e.m;
                    ROTATION<TV> rot;
                    e.random.Fill_Uniform(rot);
                    ANALYTIC_IMPLICIT_OBJECT<ORIENTED_BOX<TV> > ob(ORIENTED_BOX<TV>(obj,rot));
                    TV center;
                    e.random.Fill_Uniform(center,seed_box);
                    center.x+=i*seed_box.Edge_Lengths().x;
                    ob.analytic.corner+=center;
                    e.Seed_Particles(ob,0,0,density,e.particles_per_cell,r0);}}
            ARRAY<int> new_particles(IDENTITY_ARRAY<>(e.particles.X.m-num_part)+num_part);
            e.Set_Lame_On_Particles(1e2*e.unit_p*e.scale_E,0.3,&new_particles);
        };
    e.begin_frame.Append(func);
    e.Add_Fixed_Corotated(1e2*e.unit_p*e.scale_E,0.3);
    e.Add_Gravity(TV(0,-gravity));
    e.Add_Walls(-1,MPM::BC_SEP,.3,.1*e.m,false);

}
template void Function_72(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_72(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
