//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 74
// Testing for surface tension unmixing?
template<class T> void Function_74(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    if(e.extra_T.m<=0) e.extra_T.Append(4);
    if(e.extra_T.m<=1) e.extra_T.Append(4);
    if(e.extra_int.m<=0) e.extra_int.Append(0);
    if(e.extra_int.m<=1) e.extra_int.Append(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*2*e.m);
    LOG::printf("FLAGS: %i %i\n",e.reflection_bc,e.reflection_bc_flags);
    int gap=e.reflection_bc?0:4;
    RANGE<TV> box(e.grid.dX*gap,TV(1*e.m-e.grid.dX(0)*gap,0.5*e.m));
    box=e.grid.domain;
    if(e.extra_T.m>2) box.Scale_About_Center(e.extra_T(2));
    else box.Scale_About_Center(0.6);

    T density=2*e.unit_rho*e.scale_mass;            
    e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);

    if(e.extra_int(0)) e.particles.Resize(e.extra_int(0));
    
    ARRAY_VIEW<VECTOR<T,3> >* colors=e.particles.template Get_Array<VECTOR<T,3> >("color");
    T fluidCount=e.extra_T(0);
    T scale=1;
    if(e.extra_T.m>1)scale=e.extra_T(1);
    if(fluidCount==1)
    {
        if(e.extra_T.m>1)for(int i=0;i<e.particles.X.m;i++) e.particles.X(i)*=scale;
    }
    if(fluidCount==2)
    {
        for(int i=0;i<e.particles.X.m;i++)
        {
            if(e.extra_T.m>1)e.particles.X(i)*=scale;
            if(e.particles.X(i).y>box.Center().y)
            {
                e.particles.mass(i)*=2;
                (*colors)(i)=VECTOR<T,3>(1,0,0);
            }
        }
    }
    else if (fluidCount==3)
    {
        for(int i=0;i<e.particles.X.m;i++)
        {
            if(e.extra_T.m>1)e.particles.X(i)*=scale;
            if(e.particles.X(i).y>box.Center().y)
            {
                if(e.particles.X(i).x>box.Center().x)
                {
                    e.particles.mass(i)*=2;
                    (*colors)(i)=VECTOR<T,3>(1,0,0);
                }
                else
                {
                    e.particles.mass(i)*=3;
                    (*colors)(i)=VECTOR<T,3>(1,1,0);
                }
            }
        }

    } 
    if(!e.extra_int(1)) e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-1.8));
    //e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    e.opt.dilation_only=true;
    e.particles.mu*=0;
    e.particles.mu0*=0;
    if(e.reflection_bc) e.reflection_bc_flags=-1;
    e.Add_Walls(-1,MPM::BC_SEP,.2,e.grid.dX(0)*gap,false);
    e.st_radius=e.extra_T(0);
}
template void Function_74(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_74(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
