//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Tools/Interpolation/INTERPOLATION_CURVE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 79
// smash sphere between plates
template<class T> void Function_79(STANDARD_TESTS_BASE<VECTOR<T,2> >& e)
{
    typedef VECTOR<T,2> TV;
    typedef VECTOR<int,2> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m*2);
    SPHERE<TV> sphere(TV()*e.m,.5*e.m);
    T density=2*e.unit_rho*e.scale_mass;
    T eps=1.001;
    DIAGONAL_MATRIX<T,3> D(eps,1/eps,1/eps);
    e.Seed_Particles(sphere,0,0,density,e.particles_per_cell,r0);
    e.Add_St_Venant_Kirchhoff_Hencky_Strain(1e3*e.unit_p*e.scale_E,0.3);
    INTERPOLATION_CURVE<T,FRAME<TV> > ic0,ic1;
    ic0.Add_Control_Point(0,FRAME<TV>());
    ic0.Add_Control_Point(2,FRAME<TV>(TV(0,.25)));
    ic0.Add_Control_Point(2.01,FRAME<TV>(TV(0,0)));
    ic1.Add_Control_Point(0,FRAME<TV>());
    ic1.Add_Control_Point(2,FRAME<TV>(TV(0,-.25)));
    ic1.Add_Control_Point(2.01,FRAME<TV>(TV(0,0)));
    e.Add_Collision_Object(RANGE<TV>(TV(-2,-1.5),TV(2,-.5))*e.m,
        MPM::BC_SEP,.1,
        [=](T time){return ic0.Value(time);},
        [=](T time){return ic0.Derivative(time);});
    e.Add_Collision_Object(RANGE<TV>(TV(-2,.5),TV(2,1.5))*e.m,
        MPM::BC_SEP,.1,
        [=](T time){return ic1.Value(time);},
        [=](T time){return ic1.Derivative(time);});
}
template void Function_79(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_79(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
}
