//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 1
template<class T> void Function_1(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    SPHERE<TV> sphere(TV(.5,.5,.5)*e.m,.1*e.m);
    VECTOR<T,3> angular_velocity(TV(0.4,0,0)/e.s);
    if(e.extra_T.m) angular_velocity.x=e.extra_T(0)/e.s;
    T density=2*e.unit_rho*e.scale_mass;
    e.Seed_Particles(sphere,[=](const TV& X){return angular_velocity.Cross(X-sphere.center);},
        [=](const TV&){return MATRIX<T,3>::Cross_Product_Matrix(angular_velocity);}
        ,density,e.particles_per_cell,r0);
    T total_mass=e.particles.mass.Sum();
    TV total_momentum=e.particles.V.Weighted_Sum(e.particles.mass);
    TV dV=total_momentum/total_mass;
    e.particles.V-=dV;
    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    e.Add_Walls(-1,MPM::BC_SEP,.2,0,false);
    e.reflection_bc=true;
    e.reflection_bc_friction=.2;
}
template void Function_1(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_1(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
