//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/TORUS.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 10
// torus into a box one by one
template<class T> void Function_10(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>(TV(-1,0,-1),TV(2,2,2))*e.m,IV(3,2,3),IV()+1);
    RANGE<TV> ym(TV(0.2,0.0,0.2)*e.m,TV(0.8,0.2,0.8)*e.m);
    RANGE<TV> xm(TV(0.2,0.2,0.2)*e.m,TV(0.3,0.5,0.8)*e.m);
    RANGE<TV> xM(TV(0.7,0.2,0.2)*e.m,TV(0.8,0.5,0.8)*e.m);
    RANGE<TV> zm(TV(0.3,0.2,0.2)*e.m,TV(0.7,0.5,0.3)*e.m);
    RANGE<TV> zM(TV(0.3,0.2,0.7)*e.m,TV(0.7,0.5,0.8)*e.m);
    e.Add_Penalty_Collision_Object(ym);
    e.Add_Penalty_Collision_Object(xm);
    e.Add_Penalty_Collision_Object(xM);
    e.Add_Penalty_Collision_Object(zm);
    e.Add_Penalty_Collision_Object(zM);
    e.Add_Walls(-1,MPM::BC_SEP,.3,.1*e.m,true);
    T density=5*e.unit_rho*e.scale_mass;
    ANALYTIC_IMPLICIT_OBJECT<TORUS<T> > torus(TORUS<T>(TV(),TV(1,0,0)*e.m,0.02*1.5*e.m,0.04*1.5*e.m));
    e.Seed_Particles(torus,0,0,density,e.particles_per_cell,r0);
    for(int i=0;i<e.particles.number;i++) e.particles.valid(i)=false;
    int foo_int1=e.particles.number;
    LOG::cout<<"one torus particle #: "<<foo_int1<<std::endl;
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.8,0));
    auto func=[&e,r0,foo_int1](int frame)
        {
            if(frame%8==0 && frame<200){
                TV center;
                e.random.Fill_Uniform(center,TV(.4,1,.4)*e.m,TV(.6,1,.6)*e.m);
                T angle=e.random.Get_Uniform_Number((T)0,(T)pi*2);
                ROTATION<TV> rotation(angle,TV(0,1,0));
                int old_m=e.particles.number;
                for(int k=0;k<foo_int1;k++)
                    e.Add_Particle(center+rotation.Rotate(e.particles.X(k)),
                        0,0,e.particles.mass(k),e.particles.volume(k),r0);
                ARRAY<int> mpm_particles;
                for(int k=old_m;k<old_m+foo_int1;k++) mpm_particles.Append(k);
                e.Add_Fixed_Corotated(150*e.unit_p*e.scale_E,0.3,&mpm_particles);}
        };
    e.begin_frame.Append(func);
}
template void Function_10(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_10(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
