//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Core/Math_Tools/RANGE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 110
// box fall onto static ball
template<class T> void Function_110(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    RANGE<TV> box(TV(-0.2,0,0)*e.m,TV(0.2,0.75,.2)*e.m);
    T density=2*e.unit_rho*e.scale_mass*10;
    e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);
    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    e.particles.mu*=0;
    e.particles.mu0*=0;
    e.Add_Collision_Object(SPHERE<TV>(TV(0,-0.2,0)*e.m,.1*e.m),MPM::BC_SLIP,.1);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-1.8,0));
    e.Add_Walls(47,MPM::BC_SLIP,0,e.grid.dX.Min()*e.m,false);
}
template void Function_110(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_110(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
