//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Core/Math_Tools/RANGE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 112
// box fall onto static woods // SS-FAIL with huge E
template<class T> void Function_112(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    RANGE<TV> box(TV(-0,0,0)*e.m,TV(0.3,0.75,.3)*e.m);
    T density=2000*e.unit_rho*e.scale_mass;
    e.Seed_Particles(box,
        [=](const TV X){return TV(-1,0,0);},
        0,density,e.particles_per_cell,r0);
    // e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    e.Add_Fixed_Corotated(0.073e9*e.unit_p*e.scale_E,0.4);
    e.particles.mu*=0;
    e.particles.mu0*=0;
    for(T i=-0.9;i<=0.9;i+=0.2){
        for (T j=0;j<0.4;j+=0.1){
            e.Add_Collision_Object(SPHERE<TV>(TV(i+(fmod(j,0.2)==0?0:0.1),-0.5-j,0)*e.m,.05*e.m),MPM::BC_SLIP,.1);
        }
    }
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-1.8,0));
    e.Add_Walls(47,MPM::BC_NOSLIP,0,e.grid.dX.Min()*e.m,false);
}
template void Function_112(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_112(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
