//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 114
// lot of balls collide SS-FAIL
// command: cs && ./mpm -explicit -o output -threads 8 -last_frame 1 -particles_per_cell 50 -print_stats -3d 114 | grep "grid total energy"
template<class T> void Function_114(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    ARRAY<SPHERE<TV> > spheres;
    ARRAY<TV> v0;
    ARRAY<T> r;
    // add more balls with random velocity.
    for(T i=-0.8;i<=0.8;i+=0.2){
        for(T j=0.8;j>=0.2;j-=0.2){
            for(T k=0.8;k>=0.2;k-=0.2){
                spheres.Append(SPHERE<TV>(TV(i,j,k)*e.m,0.04*e.m));
                T a = e.random.Get_Uniform_Number(-4,4);
                T b = e.random.Get_Uniform_Number(-4,4);
                T c = e.random.Get_Uniform_Number(-4,4);
                v0.Append(TV(a,b,c)*(e.m/e.s));
                r.Append(0.03*e.m);
            }
        }
    }
    for(int i=0;i<spheres.m;i++){
        SPHERE<TV>& sphere=spheres(i);
        T density=1010*e.unit_rho*e.scale_mass;
        int last=e.particles.number;
        e.Seed_Particles(sphere,[=](const TV& X){return v0(i);},0,density,e.particles_per_cell,r0);
        for(int k=last;k<e.particles.number;k++){
            if((e.particles.X(k)-sphere.center).Magnitude_Squared()<sqr(r(i))){
                e.particles.deletion_list.Append(k);}}}// delete particles in mid circle
    e.particles.Delete_Elements_On_Deletion_List();
    e.Add_Neo_Hookean(0.073e9*e.unit_p*e.scale_E,0.4);// property of circles
    e.Add_Walls(47,MPM::BC_NOSLIP,0,e.grid.dX.Min()*e.m,false);
}
template void Function_114(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_114(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
