//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 115
// balls with fluid with gravity
template<class T> void Function_115(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    // add liquid
    RANGE<TV> box(TV(-0.8,0.75,-0.8)*e.m,TV(0.8,0.9,0.8)*e.m);
    T density=2*e.unit_rho*e.scale_mass;
    e.Seed_Particles(box,0,
        // [=](const TV X){return TV(0,.2);},
        0,density,e.particles_per_cell,r0);
    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    int particle_count = e.particles.number;
    LOG::printf("TEST particle1 mass %P volume %P \n",e.particles.mass(e.particles.number-1),e.particles.volume(e.particles.number-1));
    // e.particles.mu*=0;
    // e.particles.mu0*=0;
    // dilation_only=true; // not working in 3d

    // add more balls with random velocity.
    ARRAY<SPHERE<TV> > spheres;
    ARRAY<TV> v0;
    ARRAY<T> r;
    for(T i=-0.8;i<=0.8;i+=0.2){
        for(T j=0.8;j>=0.2;j-=0.2){
            for(T k=-0.8;k<=0.8;k+=0.2){
                spheres.Append(SPHERE<TV>(TV(i,j,k)*e.m,0.04*e.m));
                T a = e.random.Get_Uniform_Number(-40,40);
                T b = e.random.Get_Uniform_Number(-40,40);
                T c = e.random.Get_Uniform_Number(-40,40);
                v0.Append(TV(a,b,c)*(e.m/e.s));
                r.Append(0.03*e.m);
            }
        }
    }
    for(int i=0;i<spheres.m;i++){
        SPHERE<TV>& sphere=spheres(i);
        T density=1010*e.unit_rho*e.scale_mass;
        int last=e.particles.number;
        e.Seed_Particles(sphere,[=](const TV& X){return v0(i);},0,density,e.particles_per_cell,r0);
        for(int k=last;k<e.particles.number;k++){
            if((e.particles.X(k)-sphere.center).Magnitude_Squared()<sqr(r(i))){
                e.particles.deletion_list.Append(k);}}}// delete particles in mid circle
    e.particles.Delete_Elements_On_Deletion_List();
    ARRAY<int> affected_particles;
    for (int i=particle_count;i<e.particles.number;i++){
        affected_particles.Append(i);
    }
    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3); // still explode with same material. apic 2-spline, pic 2/3-spline
    LOG::printf("TEST particle2 mass %P volume %P \n",e.particles.mass(e.particles.number-1),e.particles.volume(e.particles.number-1));
    // e.Add_Neo_Hookean(0.073e9*e.unit_p*e.scale_E,0.4,&affected_particles);
    // e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-1.8,0));
    e.Add_Walls(47,MPM::BC_SLIP,0.01,e.grid.dX.Min()*e.m,false);
}
template void Function_115(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_115(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
