//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 11shell
// skew impact of two elastic spheres with initial angular velocity // -T angularv angularv cut
template<class T> void Function_11shell(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    // TODO maybe need angular velocity
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    T cut=.02;
    if(e.extra_T.m>=1)
        cut=e.extra_T(0);
    T density=2200*e.unit_rho*e.scale_mass;

    SPHERE<TV> sphere1(TV(10,15,15)/30*e.m,2*e.m/30);
    VECTOR<T,3> angular_velocity1(TV(0,0,0));
    ARRAY_VIEW<VECTOR<T,3> >* color_attribute=e.particles.template Get_Array<VECTOR<T,3> >("color");
    e.Seed_Particles(sphere1,[=,&e](const TV& X){return angular_velocity1.Cross(X-sphere1.center)+TV(5,0,0)*(e.m/e.s);},
        [=](const TV&){return MATRIX<T,3>::Cross_Product_Matrix(angular_velocity1);},density,e.particles_per_cell,r0);
    for(int i=0;i<e.particles.X.m;i++)
        if((e.particles.X(i)-sphere1.center).Magnitude()<sphere1.radius-cut)
            e.particles.deletion_list.Append(i);
    T count=e.particles.X.m;
    color_attribute->Fill(TV(0,1,0));

    SPHERE<TV> sphere2(TV(20,15,15)/30*e.m,2*e.m/30);
    VECTOR<T,3> angular_velocity2(TV(0,0,0));
    e.Seed_Particles(sphere2,[=,&e](const TV& X){return angular_velocity2.Cross(X-sphere2.center)+TV(-5,0,0)*(e.m/e.s);},
        [=](const TV&){return MATRIX<T,3>::Cross_Product_Matrix(angular_velocity2);},density,e.particles_per_cell,r0);
    for(int i=count;i<e.particles.X.m;i++)
        if((e.particles.X(i)-sphere2.center).Magnitude()<sphere2.radius-cut)
            e.particles.deletion_list.Append(i);
    e.particles.Delete_Elements_On_Deletion_List();

    T nu=.3;
    if(e.input_nu!=-1) nu=e.input_nu;
    e.Add_Neo_Hookean(1e6*e.unit_p*e.scale_E,nu); //solve({E/(2*(1+r))=11,E*r/((1+r)*(1-2*r))=81},{E,r});
    e.Add_Walls(-1,MPM::BC_SEP,.3,0,false);
    auto func=[&e]()
              {
                  T max_v=0,max_B=0,max_F=0;
                  for(int p=0;p<e.particles.V.m;p++)
                  {
                      max_v=std::max(max_v,e.particles.V(p).Magnitude_Squared());
                      if(e.particles.C.m) max_B=std::max(max_B,e.particles.C(p).Frobenius_Norm_Squared());
                      auto F=e.particles.F(p);
                      max_F=std::max(max_F,(F.Transpose_Times(F)-1).Frobenius_Norm_Squared());
                  }
                  max_v=sqrt(max_v);
                  max_B=sqrt(max_B)/((6-e.order)*e.grid.one_over_dX.Min());
                  max_F=sqrt(max_F);
                  LOG::printf("PLOT %P %P %P %P\n",e.dt,max_v,max_B,max_F);
              };
    e.Add_Callbacks(false,"time-step",func);
    int exit_early=0;
    if(e.extra_int.m) exit_early=e.extra_int(0);
    T * global_max_up= new T();
    *global_max_up=e.particles.V.Maximum_Magnitude();
    e.Add_Callbacks(false,"g2p",[&e,global_max_up,exit_early](){
                                    T max_up=0;
                                    auto& gmu=*global_max_up;
                                    max_up=e.particles.V.Maximum_Magnitude();
                                    LOG::printf("max_up %P %P \n",max_up,gmu);
                                    for(int i=0;i<e.particles.number;i++){
                                        if(e.grid.Outside(e.particles.X(i))) {
                                            LOG::printf("red unstable\n");
                                            if(exit_early) exit(0);
                                            break;}}
                                    if(max_up>2.0*gmu){
                                        LOG::printf("red unstable\n");
                                        if(exit_early) exit(0);}
                                    gmu=max_up;
                                });
    e.destroy=[]()
              {
                  LOG::printf("FINISHED\n");
              };
}
template void Function_11shell(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_11shell(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
