//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 133
// formal name splat
template<class T> void Function_133(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    // using 1 million+ particle to test 1p instability
    // one solid ball splash on one collision object
    //grep vel output/common/log.txt | sort -g -k3 | tail
    // ./mpm -explicit 133 -resolution 64 -last_frame 50 -3d -T 10 -cfl_c 1 -E .1 -nu .01

    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    T density=1000*e.unit_rho*e.scale_mass;
    SPHERE<TV> sphere1(TV(-.5,0,0)*e.m,.2*e.m);
    T vel=2;
    if(e.extra_T.m>0) vel=e.extra_T(0);
    e.Seed_Particles(sphere1,[=,&e](const TV& X){return TV(vel,0,0)*(e.m/e.s);},0,density,e.particles_per_cell,r0);
    SPHERE<TV> sphere2(TV(0,0,0)*e.m,.2*e.m);
    // RANGE<TV> box(TV(-.1,-.1,-.1)*e.m,TV(.1,.1,.1)*e.m);
    e.Add_Collision_Object(sphere2,MPM::BC_SEP,0);
    T nu=.3; if(e.input_nu!=-1) nu=e.input_nu;
    T E=1e4; if(e.input_E!=-1) E=e.input_E;
    e.Add_Neo_Hookean(E*e.unit_p*e.scale_E,nu);
    if(!e.friction_is_set) e.friction=.2;
    e.reflection_bc=true;
    e.reflection_bc_friction=0;
    e.use_full_reflection=true;
    e.Add_Walls(-1,MPM::BC_SEP,e.friction,0,false);
    // e.Add_Walls(-1,MPM::BC_PERIODIC,0,0,false);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.8,0));
}
template void Function_133(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_133(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
