//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 134
// formal name TODO ball hit wall with different constitutive model
template<class T> void Function_134(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    // using 1 million+ particle to test 1p instability
    // one solid ball splash on wall
    // cs && ./mpm -explicit 134 -resolution 32 -3d -last_frame 50 -T 15 -cfl_c .9  -threads 8 | grep -e actual -e Ratio
    // cs && ./mpm -explicit 134 -resolution 32 -3d -last_frame 50 -T 20 -cfl_c .9  -threads 8 | grep -e vel
    // density * 10


    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);

    T nu=.3; if(e.input_nu!=-1) nu=e.input_nu;
    T E=1e4; if(e.input_E!=-1) E=e.input_E;

    T vel=2; if(e.extra_T.m>0) vel=e.extra_T(0);
    T balldensity=1000;if(e.extra_T.m>1) balldensity=e.extra_T(1);
    T density=balldensity*e.unit_rho*e.scale_mass;
    SPHERE<TV> sphere1(TV(-.6,0,0)*e.m,.1*e.m);
    e.Seed_Particles(sphere1,[=,&e](const TV& X){return TV(vel,0,0)*(e.m/e.s);},0,density,e.particles_per_cell,r0);
    e.Add_Neo_Hookean(E*e.unit_p*e.scale_E,nu);

    int current_pnum=e.particles.number;

    T wallthickness=.03;if(e.extra_T.m>2) wallthickness=e.extra_T(2);
    RANGE<TV> box(TV(-wallthickness,-.7,-.7)*e.m,TV(wallthickness,.7,.7)*e.m);
    density=1000*e.unit_rho*e.scale_mass;
    e.Seed_Particles(box,[=,&e](const TV& X){return TV(0,0,0)*(e.m/e.s);},0,density,e.particles_per_cell,r0);
    ARRAY<int> affected_particles;
    for(int i=current_pnum;i<e.particles.number;i++){
        affected_particles.Append(i);
    }
    // e.Add_Neo_Hookean(1e6*e.unit_p*e.scale_E,.3,&affected_particles);
    //e.Add_Neo_Hookean(E*e.unit_p*e.scale_E,.499,&affected_particles);
    //  e.Add_Corotated(E*e.unit_p*e.scale_E,.499,&affected_particles);
    e.Add_Fixed_Corotated(E*e.unit_p*e.scale_E,.499,&affected_particles);

    // e.Add_Collision_Object(RANGE<TV>(TV(-.05,-.7,-.7),TV(.05,-.6,-.6)),MPM::BC_NOSLIP,0);
    // e.Add_Collision_Object(RANGE<TV>(TV(-.05,.6,-.7), TV(.05,.7,-.6)),MPM::BC_NOSLIP,0);
    // e.Add_Collision_Object(RANGE<TV>(TV(-.05,-.7,.6), TV(.05,-.6,.7)),MPM::BC_NOSLIP,0);
    // e.Add_Collision_Object(RANGE<TV>(TV(-.05,.6,.6),  TV(.05,.7,.7)),MPM::BC_NOSLIP,0);

    e.Add_Collision_Object(RANGE<TV>(TV(-.05,-.7,-.7),TV(.05,-.6,.7)),MPM::BC_NOSLIP,0);
    e.Add_Collision_Object(RANGE<TV>(TV(-.05,.6,-.7),TV(.05,.7,.7)),MPM::BC_NOSLIP,0);
    e.Add_Collision_Object(RANGE<TV>(TV(-.05,-.7,-.7),TV(.05,.7,-.6)),MPM::BC_NOSLIP,0);
    e.Add_Collision_Object(RANGE<TV>(TV(-.05,-.7,.6),TV(.05,.7,.7)),MPM::BC_NOSLIP,0);

    if(!e.friction_is_set) e.friction=.2;
    e.reflection_bc=true;
    e.reflection_bc_friction=0;
    e.use_full_reflection=true;
    e.Add_Walls(-1,MPM::BC_SEP,e.friction,0,false);
    // e.Add_Walls(-1,MPM::BC_PERIODIC,0,0,false);
}
template void Function_134(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_134(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
