//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 136 136fail1 136fail2
// test obj smash on ground
template<class T> void Function_136(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    // using 1 million+ particle to test 1p instability

    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    T vel=0; if(e.extra_T.m>0) vel=e.extra_T(0);
    T density=1000; if(e.extra_T.m>1) density=e.extra_T(1);
    density*=e.unit_rho*e.scale_mass;
    // SPHERE<TV> sphere1(TV(0,0,0.5)*e.m,0);
    // e.Seed_Particles(sphere1,[=,&e](const TV& X){return TV(0,vel,0)*(e.m/e.s);},0,density,e.particles_per_cell,r0);
    // RANGE<TV> box1(TV(-.7,-.05,-.7)*e.m,TV(.7,.05,.7)*e.m);
    if(e.test_name=="136fail1")
    {
        // -o ~/sca-data/output_136/E1e7T-2
        // FAIL at first farme ./mpm -3d -explicit 136fail1 -resolution 64 -min_dt 1e-7 -threads 16 -T 2 -framerate 96 -last_frame 200 -E 1e7
        // if -T -2 also fail at 30 frame
        // test: ./mpm -3d -explicit 136fail1 -resolution 256 -min_dt 1e-7 -threads 16 -T -2 -framerate 96 -last_frame 200 -E 1e7 -o ~/sca-data/output_136/E1e7T-2
        // TODO adjust E nu next
        T ind=0.02; if(e.extra_T.m>2) ind=e.extra_T(2);
        RANGE<TV> box1(TV(-.7,-.3-ind,-.7)*e.m,TV(.7,-.3+ind,.7)*e.m);
        e.Seed_Particles(box1,[=,&e](const TV& X){return TV(0,vel,0)*(e.m/e.s);},0,density,e.particles_per_cell,r0);
    }
    else if(e.test_name=="136fail2")
    {
        T rad=0.1; if(e.extra_T.m>2) rad=e.extra_T(2);
        SPHERE<TV> sphere1(TV(0,-0.5,0)*e.m,rad);
        e.Seed_Particles(sphere1,[=,&e](const TV& X){return TV(0,vel,0)*(e.m/e.s);},0,density,e.particles_per_cell,r0);
        T ind=0.02; if(e.extra_T.m>3) ind=e.extra_T(3);
        RANGE<TV> box1(TV(-.7,-1,-.7)*e.m,TV(.7,-1+ind,.7)*e.m);
        e.Seed_Particles(box1,[=,&e](const TV& X){return TV(0,0,0)*(e.m/e.s);},0,density,e.particles_per_cell,r0);
    }
    else {
        RANGE<TV> box1(TV(-.9,-.9,-.1)*e.m,TV(.9,.7,.1)*e.m);
        e.Seed_Particles(box1,[=,&e](const TV& X){return TV(0,vel,0)*(e.m/e.s);},0,density,e.particles_per_cell,r0);
    }
    T E=1e4; if(e.input_E!=-1) E=e.input_E;
    T nu=.3; if(e.input_nu!=-1) nu=e.input_nu;
    e.Add_Neo_Hookean(E*e.unit_p*e.scale_E,nu);
    if(!e.friction_is_set) e.friction=.2;
    e.reflection_bc=true;
    e.reflection_bc_friction=0;
    e.use_full_reflection=true;
    e.Add_Walls(-1,MPM::BC_SEP,e.friction,0,false);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.8,0));
}
template void Function_136(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_136(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
