//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 138 138set1 138set2 138set3
// two sand object splash towards each other
template<class T> void Function_138(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    // using 1 million+ particle to test 1p instability
    // ../mpm -explicit 138 -min_dt 1e-8 -resolution 256 -3d -last_frame 200 -threads 16 -framerate 120 -o ~/sca-data/output_138/sand -single_particle_cfl

    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    T density=2000*e.unit_rho*e.scale_mass;

    T vel=5; if(e.extra_T.m>0) vel=e.extra_T(0);
    if(e.test_name=="138set1"){
        // not working when particles touch the wall they spread in one direction; has to do with the boundary force or conditions.
        SPHERE<TV> sphere1(TV(-.3,-.3,0)*e.m,.3*e.m);
        e.Seed_Particles(sphere1,[=,&e](const TV& X){return TV(vel*1.5,vel*1.5,0)*(e.m/e.s);},0,density,e.particles_per_cell,r0);
        SPHERE<TV> sphere2(TV(.5,.5,0)*e.m,.2*e.m);
        e.Seed_Particles(sphere2,[=,&e](const TV& X){return TV(-vel,-vel,0)*(e.m/e.s);},0,density,e.particles_per_cell,r0);
    }
    else if(e.test_name=="138set2")
    {
        SPHERE<TV> sphere1(TV(.3,.3,0)*e.m,.27*e.m);
        e.Seed_Particles(sphere1,[=,&e](const TV& X){return TV(-vel,-vel,0)*(e.m/e.s);},0,density,e.particles_per_cell,r0);
        SPHERE<TV> sphere2(TV(-.2,-.2,0)*e.m,.15*e.m);
        e.Add_Collision_Object(sphere2,MPM::BC_SEP,0);
    }
    // else if(e.test_name=="138set3") // no different constitutive model for now
    // {
    //     SPHERE<TV> sphere1(TV(-.3,-.3,0)*e.m,.2*e.m);
    //     e.Seed_Particles(sphere1,[=,&e](const TV& X){return TV(vel,vel,0)*(e.m/e.s);},0,density,e.particles_per_cell,r0);
    // }

    T E=35.37e6*e.unit_p*e.scale_E,nu=.3;
    e.Add_Drucker_Prager(E,nu,(T)35,9,0.3,10);
    e.Set_Lame_On_Particles(E,nu);

    // if(e.test_name=="138set3")
    // {
    //     SPHERE<TV> sphere2(TV(.3,.3,0)*e.m,.15*e.m);
    //     e.Seed_Particles(sphere2,[=,&e](const TV& X){return TV(-vel,-vel,0)*(e.m/e.s);},0,density,e.particles_per_cell,r0);
    //     ARRAY<int> solid_particles(e.particles.X.m-sand_particles.m);
    //     for(int p=sand_particles.m;p<e.particles.X.m;p++) solid_particles(p-sand_particles.m)=p;
    //     e.Add_Neo_Hookean(E,nu,&solid_particles);
    // }

    if(!e.friction_is_set) e.friction=.2;
    e.reflection_bc=true;
    e.reflection_bc_friction=0;
    e.use_full_reflection=true;
    e.Add_Walls(-1,MPM::BC_SEP,e.friction,0,false);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.8,0));
}
template void Function_138(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_138(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
