//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include <Geometry/Meshes/TRIANGULATED_SURFACE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 139
// formal name billiards
template<class T> void Function_139(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    // using 1 million+ particle to test 1p instability
    // four ball left , one ball right
    // TODO apply mesh look 3d/87

    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    T density=1000*e.unit_rho*e.scale_mass;
    T r=.1;
    SPHERE<TV> sphere1(TV(-.3,-1+r,0)*e.m,r*e.m);
    SPHERE<TV> sphere2(TV(-.1,-1+r,0)*e.m,r*e.m);
    SPHERE<TV> sphere3(TV(.1, -1+r,0)*e.m,r*e.m);
    SPHERE<TV> sphere4(TV(.3, -1+r,0)*e.m,r*e.m);
    SPHERE<TV> sphere5(TV(.8, -1+r,0)*e.m,r*e.m);

    T rs=(T)e.resolution/128;
    int lg_rs=rint(log(rs)/log(2));
    int res_s=6+lg_rs;
    auto save=[&e](auto* ts,std::string file) {
            e.viewer_dir.Make_Common_Directory();
            ts->Write_Obj(e.viewer_dir.Common_File(file+".obj"));
            delete ts;};
    save(Generate_Triangles(sphere1,res_s),"sphere-orig");

    TWIST<TV> t;
    auto V_0=[](const TV& X) {return TV(0,0,0);};
    auto dV=[t](const TV& X) {return MATRIX<T,TV::m>::Cross_Product_Matrix(t.angular);};
    e.Seed_Particles_To_Analytic_Surface(sphere1,V_0,dV,density,e.particles_per_cell,r0++,true,res_s);
    e.Seed_Particles_To_Analytic_Surface(sphere2,V_0,dV,density,e.particles_per_cell,r0++,true,res_s);
    e.Seed_Particles_To_Analytic_Surface(sphere3,V_0,dV,density,e.particles_per_cell,r0++,true,res_s);
    e.Seed_Particles_To_Analytic_Surface(sphere4,V_0,dV,density,e.particles_per_cell,r0++,true,res_s);
    T vel=-2; if(e.extra_T.m>0) vel=e.extra_T(0);
    e.Seed_Particles_To_Analytic_Surface(sphere5,[vel,&e](const TV& X){return TV(vel,0,0)*(e.m/e.s);},dV,density,e.particles_per_cell,r0++,true,res_s);

    T nu=.3; if(e.input_nu!=-1) nu=e.input_nu;
    T E=1e7; if(e.input_E!=-1) E=e.input_E;
    e.Add_Neo_Hookean(E*e.unit_p*e.scale_E,nu);
    if(!e.friction_is_set) e.friction=.2;
    e.reflection_bc=true;
    e.reflection_bc_friction=0;
    e.use_full_reflection=true;
    e.Add_Walls(-1,MPM::BC_SEP,e.friction,0,false);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.8,0));
}
template void Function_139(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_139(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
