//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 140
// formal name flying particles/ shooting particles/ dust ()test the scene to make sense)
template<class T> void Function_140(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    // using 1 million+ particle to test 1p instability
    // How to simulate air to make particles keep floating, it is easy to sample them, use one callback to add force to them

    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    T density=1000*e.unit_rho*e.scale_mass;

    RANGE<TV> box(TV(-.5,-.5,-.5)*e.m,TV(.5,.5,.5)*e.m);
    e.Seed_Particles_Poisson(box,0,0,density,.5,r0);

    T nu=.3; if(e.input_nu!=-1) nu=e.input_nu;
    T E=1e4; if(e.input_E!=-1) E=e.input_E;
    e.Add_Neo_Hookean(E*e.unit_p*e.scale_E,nu);
    if(!e.friction_is_set) e.friction=.2;
    e.reflection_bc=true;
    e.reflection_bc_friction=0;
    e.use_full_reflection=true;
    e.Add_Walls(-1,MPM::BC_SEP,e.friction,0,false);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.8,0));

    // TODO add a pipe so p.v+=(wind_vel-p.v)*.1 simulate the friction of air
    // e.Add_Callbacks()
}
template void Function_140(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_140(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
