//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 19
template<class T> void Function_19(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    e.particles.Store_Fp(true);

    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    if(e.use_penalty_collisions)
        e.Add_Penalty_Collision_Object(RANGE<TV>(TV(-0.5,-1,-0.5),TV(1.5,.1,1.5))*e.m,0.9);
    else
        e.Add_Collision_Object(RANGE<TV>(TV(-0.5,-1,-0.5),TV(1.5,.1,1.5))*e.m,MPM::BC_NOSLIP,0);

    T density=(T)1281*e.unit_rho*e.scale_mass; // source: Sand, dry http://www.engineeringtoolbox.com/density-materials-d_1652.html
    T E=35.37e6*e.unit_p*e.scale_E,nu=.3;
    e.Add_St_Venant_Kirchhoff_Hencky_Strain(E,nu);
    T gap=e.grid.dX(1)*0.1;
    RANGE<TV> box(TV(.4*e.m,.1*e.m+gap,.4*e.m),TV(.45*e.m,.1*e.m+gap+0.5*e.m,.45*e.m));
    e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);
    e.Set_Lame_On_Particles(E,nu);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.81,0));
//// an elastic sphere collide with a collision object 
}
template void Function_19(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_19(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
