//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_INTERSECTION.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_INVERT.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 33
// dry sand notch dam break
template<class T> void Function_33(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    // ./mpm 33 -3d -threads 8 -resolution 30 -last_frame 40 -framerate 48 -T 4 -T 30 -max_dt 1e-4 -scale_E 0.01 -explicit -no_implicit_plasticity -o notch30
    if(e.extra_T.m<=0) e.extra_T.Append(0);
    if(e.extra_T.m<=1) e.extra_T.Append(0);
    e.particles.Store_Fp(true);
    e.Set_Grid(RANGE<TV>(TV(),TV(1.5,1,1.5))*e.m,IV(3,2,3));
    LOG::cout<<"GRID dx: "<<e.grid.dX<<std::endl;
    RANGE<TV> ground(TV(-10,-5,-5)*e.m,TV(10,0.05,10)*e.m);
    RANGE<TV> left_wall(TV(-5,-5,-5)*e.m,TV(0.05,10,10)*e.m);
    RANGE<TV> back_wall(TV(-5,-1,-10)*e.m,TV(10,10,0.05)*e.m);
//            IMPLICIT_OBJECT<TV>* bounds=new IMPLICIT_OBJECT_DILATE<TV>(new IMPLICIT_OBJECT_INVERT<TV>(new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(e.grid.domain.Thickened(-.05))),-.01);

    
    // IMPLICIT_OBJECT_UNION<TV>* bounds=new IMPLICIT_OBJECT_UNION<TV>(
    //     ,
    //     new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(left_wall),
    //     new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(back_wall));
    e.Add_Collision_Object(new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(ground),MPM::BC_SLIP,e.extra_T(0));
    e.Add_Collision_Object(new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(left_wall),MPM::BC_SLIP,e.extra_T(0));
    e.Add_Collision_Object(new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(back_wall),MPM::BC_SLIP,e.extra_T(0));
    T density=(T)1600*e.unit_rho*e.scale_mass;
    T E=35.37e6*e.unit_p*e.scale_E,nu=.3;
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
    IMPLICIT_OBJECT_INTERSECTION<TV>* fill_sand=new IMPLICIT_OBJECT_INTERSECTION<TV>(
        new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(RANGE<TV>(TV()+0.05,TV(0.5,0.8,0.5)+0.05)),
        new IMPLICIT_OBJECT_INVERT<TV>(
            new ANALYTIC_IMPLICIT_OBJECT<SPHERE<TV> >(SPHERE<TV>(TV(0.5,0.4,0.5)+0.05,0.225))));
    e.Seed_Particles(*fill_sand,0,0,density,e.particles_per_cell,r0);
    LOG::cout<<"Particle count: "<<e.particles.number<<std::endl;
    e.Set_Lame_On_Particles(E,nu);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-1,0));
    //e.extra_T(1) is the friction angle
    e.Add_Drucker_Prager(E,nu,e.extra_T(1));

}
template void Function_33(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_33(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
