//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 34
// sand dam break
template<class T> void Function_34(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    // usage:./mpm 34 -3d -use_exp_F -max_dt 1e-3 -unit_p*e.scale_E 10 -T 10 -T 1000 -T 3 -last_frame 20 
    // usage: ./siggraph-2020/run_sand_dam_break.sh
    // usage:./mpm 34 -3d -use_exp_F -last_frame 30 -resolution 32 -threads 16 -explicit -use_reflect -use_reflect_friction -no_implicit_plasticity -sand_color -float
    // -friction 1 -use_reflect_friction -use_reflect
    e.particles.Store_Fp(true);

    T shift=.1;
    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    if(e.reflection_bc)
    {
        shift=0;
        if(!e.friction_is_set) e.friction=.3;
        if(e.extra_int.m){
            if(e.extra_int(0)==0)
                e.Add_Walls(-1,MPM::BC_SLIP,.3,.1*e.m,false);
            if(e.extra_int(0)==1)
                e.Add_Walls(-1,MPM::BC_NOSLIP,.3,.1*e.m,false);
            if(e.extra_int(0)==2)
                e.Add_Walls(-1,MPM::BC_SEP,e.friction,.1*e.m,false);}
        else
            e.Add_Walls(-1,MPM::BC_SEP,e.friction,.1*e.m,false);
    }
    else{
        RANGE<TV> ground(TV(-0.1,0,-0.1)*e.m,TV(1.1,shift,1.1)*e.m);
        RANGE<TV> left_wall(TV(-0.1,0,-0.1)*e.m,TV(shift,1.1,1.1)*e.m);
        RANGE<TV> back_wall(TV(-0.1,0,-0.1)*e.m,TV(1.1,1.1,shift)*e.m);
        if(e.use_penalty_collisions){
            e.Add_Penalty_Collision_Object(ground);
            e.Add_Penalty_Collision_Object(left_wall);
            e.Add_Penalty_Collision_Object(back_wall);}
        else{
            e.Add_Collision_Object(ground,MPM::BC_NOSLIP,0);
            e.Add_Collision_Object(left_wall,MPM::BC_NOSLIP,0);
            e.Add_Collision_Object(back_wall,MPM::BC_NOSLIP,0);}}

    T density=(T)2200*e.unit_rho*e.scale_mass;
    T E=35.37e6*e.unit_p*e.scale_E,nu=.3;
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
    T gap=e.grid.dX(1)*shift;
    RANGE<TV> box(TV(shift*e.m+gap,shift*e.m+gap,shift*e.m+gap),TV(.3,.75,.3)*e.m);
    e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);
    ARRAY<int> sand_particles(e.particles.X.m);
    
    for(int p=0;p<e.particles.X.m;p++) sand_particles(p)=p;
    e.Add_Drucker_Prager(E,nu,(T)35,9,0.3,10,&sand_particles);
    //int case_num=e.use_hardening_mast_case?e.hardening_mast_case:2;
    //e.Add_Drucker_Prager_Case(E,nu,case_num);
    e.Set_Lame_On_Particles(E,nu);
    
    // if(test_name=="34"){
    //     T El=500*e.unit_p*e.extra_T(0),nul=0.1*e.extra_T(2);
    //     e.Add_Lambda_Particles(&sand_particles,El,nul,e.extra_T(1),true);}
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.81,0));
    LOG::printf("TEST E %P nu %P lam %P mu %P \n",E,nu,e.particles.lambda(0),e.particles.mu(0));
}
template void Function_34(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_34(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
