//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/CONE.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_INTERSECTION.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 35
// cup
template<class T> void Function_35(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    e.particles.Store_Fp(true);
    int r=(e.resolution+9)/10;
    IV res(r*10,r*3,r*10);
    TV extent(res*e.m);
    extent /= 2*e.resolution;
    e.Set_Grid(RANGE<TV>(TV(0.25,0,0.25)*e.m,TV(0.25,0,0.25)*e.m+extent),IV(10,3,10),IV(),10);
    RANGE<TV> ground(TV(-0.1,0,-0.1)*e.m,TV(1.1*e.m,e.grid.dX(1),1.1*e.m));
    if(e.use_penalty_collisions)
        e.Add_Penalty_Collision_Object(ground);
    else
        e.Add_Collision_Object(ground,MPM::BC_NOSLIP,0);

    T density=(T)2200*e.unit_rho*e.scale_mass;
    T E=35.37e6*e.unit_p*e.scale_E,nu=.3;
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
    int case_num=e.use_hardening_mast_case?e.hardening_mast_case:2;
    e.Add_Drucker_Prager_Case(E,nu,case_num);
    CONE<T> cone(TV(0.5,e.grid.dX(1)/e.m*1.1,0.5)*e.m,TV(0.5,0.235+e.grid.dX(1)/e.m*1.1,0.5)*e.m,0.042*e.m);
    RANGE<TV> box(TV(0.4,0,0.4)*e.m,TV(0.6*e.m,0.095*e.m+e.grid.dX(1)*1.1,0.6*e.m));
    IMPLICIT_OBJECT_INTERSECTION<TV> cup(new ANALYTIC_IMPLICIT_OBJECT<CONE<T>>(cone),new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV>>(box));
    e.Seed_Particles(cup,0,0,density,e.particles_per_cell,r0);
    e.Set_Lame_On_Particles(E,nu);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.81,0));
}
template void Function_35(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_35(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
