//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/HOURGLASS.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_INTERSECTION.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_INVERT.h>
#include <Geometry/Meshes/TRIANGULATED_SURFACE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 36
// hourglass
template<class T> void Function_36(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    if(e.extra_int.m<1) e.extra_int.Append(150);
    if(e.extra_int.m<2) e.extra_int.Append(80);
    if(e.extra_T.m<1) e.extra_T.Append(.16);
    if(e.extra_T.m<2) e.extra_T.Append(.0225);
    if(e.extra_T.m<3) e.extra_T.Append(.8);
    if(e.extra_T.m<4) e.extra_T.Append(.0225);
    if(e.extra_T.m<5) e.extra_T.Append(.1);
    if(e.extra_T.m<6) e.extra_T.Append((T).3);
    e.particles.Store_Fp(true);
    e.Set_Grid(RANGE<TV>(TV(-0.2,-0.45,-0.2)*e.m,TV(0.22,0.44,0.22)*e.m),IV(1,2,1));
    LOG::cout<<"GRID DX: " <<e.grid.dX<<std::endl;
    HOURGLASS<TV> hourglass(TV::Axis_Vector(1),TV(),e.extra_T(0),e.extra_T(1),e.extra_T(2),e.extra_T(3));
    IMPLICIT_OBJECT<TV>* hg=new ANALYTIC_IMPLICIT_OBJECT<HOURGLASS<TV> >(hourglass);
    IMPLICIT_OBJECT<TV>* inv=new IMPLICIT_OBJECT_INVERT<TV>(hg);
    if(e.use_penalty_collisions) e.Add_Penalty_Collision_Object(inv);
    else e.Add_Collision_Object(inv,MPM::BC_SEP,e.extra_T(5));
    T density=(T)2200*e.unit_rho*e.scale_mass;
    T E=35.37e6*e.unit_p*e.scale_E,nu=.3;
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
    RANGE<TV> fill_part=e.grid.domain;
    fill_part.min_corner.y=0;
    fill_part.max_corner.y=e.extra_T(4);
    ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> > io_fill_part(fill_part);
    IMPLICIT_OBJECT_INTERSECTION<TV> ioi(&io_fill_part,hg);
    ioi.owns_io.Fill(false);
    e.Seed_Particles(ioi,0,0,density,e.particles_per_cell,r0);
    LOG::printf("added %i particles.",e.particles.X.m);
    e.Set_Lame_On_Particles(E,nu);
    e.Add_Drucker_Prager_Case(E,nu,2);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.81,0));
    if(e.dump_collision_objects){
        TRIANGULATED_SURFACE<T>* ts=Tessellate_Boundary(hourglass,e.extra_int(0),e.extra_int(1));
        Write_To_File(e.stream_type,LOG::sprintf("hourglass-%d-%d.tri.gz",e.extra_int(0),e.extra_int(1)),*ts);
        delete ts;}
}
template void Function_36(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_36(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
