//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/ORIENTED_BOX.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 37a 37b
// sand on wedge
template<class T> void Function_37(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    if(e.extra_T.m<=0) e.extra_T.Append(0);
    if(e.extra_T.m<=1) e.extra_T.Append(0);
    if(e.extra_T.m<=2) e.extra_T.Append(0);
    if(e.extra_T.m<=3) e.extra_T.Append(1);
    e.particles.Store_Fp(true);

    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    ORIENTED_BOX<TV> wedge(RANGE<TV>(TV(0,0,-0.1),TV(0.2,0.2,1.1))*e.m,ROTATION<TV>::From_Euler_Angles(TV(0,0,0.25*M_PI)),TV(0.5,0.4-sqrt(2.0)*0.1,0)*e.m);
    RANGE<TV> ground(TV(-0.1,0,-0.1)*e.m,TV(1.1,0.1,1.1)*e.m);
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;

    if(e.use_penalty_collisions){
        e.Add_Penalty_Collision_Object(wedge);
        e.Add_Penalty_Collision_Object(ground);}
    else{
        e.Add_Collision_Object(ground,MPM::BC_NOSLIP,0);
        e.Add_Collision_Object(new ANALYTIC_IMPLICIT_OBJECT<ORIENTED_BOX<TV> >(wedge),MPM::BC_SEP,1);}

    T density=(T)2200*e.unit_rho*e.scale_mass;
    T E=35.37e5*e.unit_p*e.scale_E,nu=.3;
    if(e.test_name=="37a"){
        int case_num=e.use_hardening_mast_case?e.hardening_mast_case:2;
        e.Add_Drucker_Prager_Case(E,nu,case_num);}
    //add sand particles
    RANGE<TV> box(TV(.3,.7,.3),TV(.7,.9,.7));
    e.Seed_Particles(box,0,0,density,e.particles_per_cell,r0);
    e.Set_Lame_On_Particles(E,nu);

    if(e.test_name=="37b"){
        ARRAY<int> sand_particles(e.particles.X.m);
        for(int p=0;p<e.particles.X.m;p++) sand_particles(p)=p;
        e.Add_Drucker_Prager(E,nu,(T)35,&sand_particles);
        e.Add_Lambda_Particles(&sand_particles,E*e.extra_T(3),nu,e.extra_T(0)*e.unit_rho,r0,true,e.extra_T(1),e.extra_T(2));}

    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.81,0));
}
template void Function_37(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_37(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
