//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/CYLINDER.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 43
// Rotating table
template<class T> void Function_43(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    e.particles.Store_Fp(true);
    e.Set_Grid(RANGE<TV>(TV(),TV(1,0.25,1)*e.m),IV(4,1,4));
    LOG::printf("REAL GRID: %P\n",e.grid);

    if(!e.friction_is_set)e.friction=0.5;
    const T omega=0.5*pi/e.s;
    e.Add_Collision_Object(
            CYLINDER<T>(TV(0,-0.1,0),TV(0,-0.1,0),0.5),
            MPM::BC_SEP,e.friction,
            [=](T time){return FRAME<TV>(TV(),ROTATION<TV>::From_Euler_Angles(0,omega*time,0));},
            [=](T time){return TWIST<TV>(TV(),typename TV::SPIN(0,omega,0));});
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.8,0));

// ./mpm -3d 44 -last_frame 20 -T 35 -T 1 -resolution 10 -use_exp_F -max_dt 2e-4 -explicit 
}
template void Function_43(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_43(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
