//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include <Deformables/Constitutive_Models/COROTATED_FIXED.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 69
// Magic balls
template<class T> void Function_69(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    if(e.extra_T.m<=0) e.extra_T.Append(1e-3);
    e.particles.Store_Fp(true);
    e.particles.Store_Lame(true);
    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    LOG::cout<<"GRID dx: "<<e.grid.dX<<std::endl;
    RANGE<TV> boxymin(TV(-10,-10,-10)*e.m,TV(10,0.03,10)*e.m);
    if(e.use_penalty_collisions) PHYSBAM_FATAL_ERROR();
    else e.Add_Collision_Object(boxymin,MPM::BC_SLIP,0.7);
    T density=(T)2200*e.unit_rho*e.scale_mass;
    T E=35.37e6*e.unit_p*e.scale_E,nu=.3;
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;

    const SPHERE<TV> ball(TV(0.5,0.75,0.5)*e.m,.1*e.m);
    const SPHERE<TV> padded_ball(ball.center,0.11*e.m);
    const T ball_volume=ball.Size();
    const int number_of_particles=e.particles_per_cell*ball_volume/e.grid.dX.Product();
    const T volume_per_particle=ball_volume/number_of_particles;

    ARRAY<SPHERE<TV> > spheres;
    for(int i=1;i<=500;i++){
        TV offset=e.random.template Get_Vector_In_Unit_Sphere<TV>()*padded_ball.radius;
        if(offset.Magnitude()<padded_ball.radius*0.5 || offset.Magnitude()>padded_ball.radius*0.9)
        {i--;continue;}
        T radius=e.random.Get_Uniform_Number(0,padded_ball.radius-offset.Magnitude());
        if(radius<padded_ball.radius*0.1)
        {i--;continue;}
        spheres.Append(SPHERE<TV>(offset+padded_ball.center,radius));}

    const T h=10;
    const T a=1.025;
    const T ea=exp((a-1)*h); 
    //e.particles.Preallocate(number_of_particles);
    e.particles.Resize(number_of_particles);
    ARRAY<int> sand_particles(number_of_particles);
    for(int p=0;p<e.particles.X.m;p++) sand_particles(p)=p;
    ARRAY_VIEW<VECTOR<T,3> >* color_attribute=e.particles.template Get_Array<VECTOR<T,3> >("color");
    for(int p=0;p<number_of_particles;p++){
        e.particles.valid(p)=true;
        e.particles.X(p)=e.random.template Get_Vector_In_Unit_Sphere<TV>()*padded_ball.radius+padded_ball.center;
        e.particles.F(p)=MATRIX<T,TV::m>()+1;
        if(e.particles.store_Fp) e.particles.Fp(p).Set_Identity_Matrix();

        bool inside=ball.Inside(e.particles.X(p),0);
        for(int j=0;j<spheres.m;j++)
            inside|=spheres(j).Inside(e.particles.X(p),0);
        if(!inside){p--;continue;}

        e.particles.V(p)=TV(0,-2.7,0);
        e.particles.mass(p)=density*volume_per_particle*e.Perturb(0.5);
        e.particles.volume(p)=volume_per_particle;

        (*color_attribute)(p)=VECTOR<T,3>(1,1,1);
        T E_local=E*e.Perturb(0.5);
        T nu_local=nu*e.Perturb(0.1);
        e.particles.mu(p)=E_local/(2*(1+nu_local));
        e.particles.lambda(p)=E_local*nu_local/((1+nu_local)*(1-2*nu_local));

        T mult=1;
        for(int i=0;i<10;i++)
            if((e.particles.X(p)-ball.center).Magnitude()>(i+1)*ball.radius/10){
                e.particles.mass(p)*=a;
                e.particles.lambda(p)*=ea;
                e.particles.lambda0(p)*=ea;
                e.particles.mu(p)*=ea;
                e.particles.mu0(p)*=ea;}
            else mult/=a;

        if(e.Uniform(0,1)>mult) p--;}

    //water
    T porosity=0.3;
    T saturation_level=1;
    T water_density=(T)1000*e.unit_rho;
    T water_E=E*e.extra_T(0);
    ARRAY<int> lambda_particles;
    T volume_lambda=e.particles.volume(0)*porosity*saturation_level;
    T mass_lambda=water_density*volume_lambda;
    T lambda_water=water_E*nu/((1+nu)*(1-2*nu));
    for(int k=0;k<sand_particles.m;k++){
        int p=e.particles.Add_Element();
        lambda_particles.Append(p);
        e.particles.valid(p)=true;
        e.particles.mass(p)=mass_lambda;
        int i=sand_particles(k);
        e.particles.X(p)=e.particles.X(i);
        e.particles.V(p)=e.particles.V(i);
        e.particles.F(p)=e.particles.F(i);
        if(e.particles.store_Fp) e.particles.Fp(p)=e.particles.Fp(i); 
        if(e.particles.store_C) e.particles.C(p)=e.particles.C(i);
        if(e.particles.store_S) e.particles.S(p)=e.particles.S(i);
        e.particles.volume(p)=volume_lambda;
        e.particles.mu(p)=(T)0;
        e.particles.mu0(p)=(T)0;
        e.particles.lambda(p)=lambda_water;
        e.particles.lambda0(p)=lambda_water;}
    for(int p=number_of_particles;p<e.particles.X.m;p++){
        T mult=1;
        for(int i=0;i<10;i++)
            if((e.particles.X(p)-ball.center).Magnitude()>(i+1)*ball.radius/10){
                e.particles.mass(p)*=a;
                e.particles.lambda(p)*=ea;
                e.particles.lambda0(p)*=ea;}
            else mult/=a;

        if(e.Uniform(0,1)>mult) p--;}
    //Add constitutive model
    e.Add_Drucker_Prager(E,nu,(T)35,&sand_particles,false,0);
    e.Add_Fixed_Corotated(water_E,nu,&lambda_particles,true);
    if(!e.use_theta_c) e.theta_c=0.01;
    if(!e.use_theta_s) e.theta_s=.00001;
    if(!e.use_hardening_factor) e.hardening_factor=80;
    if(!e.use_max_hardening) e.max_hardening=5;
    e.Add_Clamped_Plasticity(*new COROTATED_FIXED<T,TV::m>(water_E,nu),e.theta_c,e.theta_s,e.max_hardening,e.hardening_factor,&lambda_particles); 
    for(int k=0;k<500;k++){
        const TV offset=e.random.template Get_Vector_In_Unit_Sphere<TV>()*padded_ball.radius*5;
        const T radius=e.random.Get_Uniform_Number(padded_ball.radius,padded_ball.radius*10);
        SPHERE<TV> sphere_big(offset+padded_ball.center,radius+e.grid.dX.Min()/2*2);
        SPHERE<TV> sphere_small(offset+padded_ball.center,radius-e.grid.dX.Min()/2*2);
        const TV offset_new=e.random.template Get_Vector_In_Unit_Sphere<TV>()*padded_ball.radius*5;
        const T radius_new=e.random.Get_Uniform_Number(padded_ball.radius,padded_ball.radius*10);
        SPHERE<TV> sphere_new(offset_new+padded_ball.center,radius_new);
        if(offset_new.Magnitude()+ball.radius<radius_new) continue;
        if((offset-offset_new).Magnitude()+radius<radius_new) continue;
        for(int i=0;i<e.particles.X.m;i++)
            if(sphere_big.Inside(e.particles.X(i),0)&&!sphere_small.Inside(e.particles.X(i),0)&&sphere_new.Inside(e.particles.X(i),0)){
                e.particles.lambda(i)*=0.5;
                e.particles.mu(i)*=0.5;
                (*color_attribute)(i)(2)*=0.5;}}

    LOG::printf("check number_of_particles=%P\n",number_of_particles);
    LOG::printf("check particles.X.m=%P\n",e.particles.X.m);
    LOG::printf("particles count=%P\n",e.particles.X.m);
    //shift the particles
    for(int p=0;p<e.particles.X.m;p++) e.particles.X(p)+=TV(0,-0.60,0);
    e.Add_Gravity(TV(0,-9.81,0));

}
template void Function_69(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_69(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
