//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 7
// skew impact of two elastic spheres
template<class T> void Function_7(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>(TV(),TV(30,30,30))*e.m);
    T density=5*e.unit_rho*e.scale_mass;
    SPHERE<TV> sphere1(TV(10,13,15)*e.m,2*e.m);
    e.Seed_Particles(sphere1,
        [&e](const TV& X){return TV(0.75,0,0)*(e.m/e.s);},
        [&e](const TV&){return MATRIX<T,3>();},
        density,e.particles_per_cell,r0);
    SPHERE<TV> sphere2(TV(20,15,15)*e.m,2*e.m);
    e.Seed_Particles(sphere2,
        [&e](const TV& X){return TV(-0.75,0,0)*(e.m/e.s);},
        [&e](const TV&){return MATRIX<T,3>();},
        density,e.particles_per_cell,r0);
    e.Add_Neo_Hookean(31.685*e.unit_p*e.scale_E,0.44022); //solve({E/(2*(1+r))=11,E*r/((1+r)*(1-2*r))=81},{E,r});
}
template void Function_7(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_7(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
