//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/ORIENTED_BOX.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include <Deformables/Constitutive_Models/COROTATED_FIXED.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 72
template<class T> void Function_72(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    //flip .95, max_dt 0.005, cfl 0.05
    e.particles.Store_Fp(true);
    e.particles.Store_Lame(true);
    e.resolution=150;
    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);

    ORIENTED_BOX<TV> wedge(RANGE<TV>(TV(0,0,-0.1),TV(0.2,0.2,1.1))*e.m,ROTATION<TV>::From_Euler_Angles(TV(0,0,0.25*M_PI)),TV(0.5,0.4-sqrt(2.0)*0.1,0)*e.m);
    RANGE<TV> ground(TV(-1,0,-1)*e.m,TV(2,0.1,2)*e.m);
    e.Add_Collision_Object(ground,MPM::BC_SEP,1);
    e.Add_Collision_Object(new ANALYTIC_IMPLICIT_OBJECT<ORIENTED_BOX<TV> >(wedge),MPM::BC_SEP,1);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-2.5,0));

    T density=(T)3*e.unit_rho*e.scale_mass;
    T E=360*e.unit_p*e.scale_E,nu=0.3;
    const int number_of_particles=337500;
    const RANGE<TV> block(TV(.3,.7,.4),TV(.7,.9,.6));
    const T block_volume=block.Size();
    const T volume_per_particle=block_volume/number_of_particles;
    e.particles.Resize(number_of_particles);

    ARRAY_VIEW<VECTOR<T,3> >* color_attribute=e.particles.template Get_Array<VECTOR<T,3> >("color");
    for(int p=0;p<number_of_particles;p++){
        e.particles.valid(p)=true;
        e.random.Fill_Uniform(e.particles.X(p),block);
        e.particles.F(p)=MATRIX<T,TV::m>()+1;
        if(e.particles.store_Fp) e.particles.Fp(p).Set_Identity_Matrix();
        e.particles.V(p)=TV();
        e.particles.mass(p)=density*volume_per_particle*e.Perturb(0.5);
        e.particles.volume(p)=volume_per_particle;
        e.particles.mass(p)=e.scale_mass*density*volume_per_particle;
        (*color_attribute)(p)=VECTOR<T,3>(1,1,1);}
    for(int k=0;k<number_of_particles;k++) 
        if(e.particles.F(k)==MATRIX<T,TV::m>()){
        LOG::cout<<"F("<<k<<")="<<e.particles.F(k)<<std::endl;
        PHYSBAM_FATAL_ERROR();}
    e.Set_Lame_On_Particles(E,nu);
    if(!e.use_theta_c) e.theta_c=0.025;
    if(!e.use_theta_s) e.theta_s=.005;
    if(!e.use_hardening_factor) e.hardening_factor=10;
    e.Add_Fixed_Corotated(E*e.unit_p*e.scale_E,nu);
    e.Add_Clamped_Plasticity(*new COROTATED_FIXED<T,TV::m>(E,nu),
        e.theta_c,e.theta_s,e.max_hardening,e.hardening_factor); 
}
template void Function_72(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_72(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
