//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Geometry/Meshes/TETRAHEDRALIZED_VOLUME.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 74
// Reflection bc test
template<class T> void Function_74(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>(TV(),TV(4,1,1))*e.m,IV(4,1,1));
    e.reflection_bc=true;
    e.use_full_reflection=true;
    T E=1e4*e.unit_p*e.scale_E,nu=.3;
    T density=(T)2*e.unit_rho*e.scale_mass;
    TV gravity=TV(0,-9.8*e.m/(e.s*e.s),0);
    T vel=2*e.m/e.s;
    RANGE<TV> box(TV(.25,0,.25)*e.m,TV(1.25,.5,.75)*e.m);
    VECTOR<TV,8> C;
    box.Corners(C);
    TETRAHEDRALIZED_VOLUME<T> cube;
    cube.Initialize_Cube_Mesh_And_Particles(GRID<TV>(IV(box.Edge_Lengths()*e.grid.one_over_dX*2),box));
    e.Seed_Lagrangian_Particles(cube,
        [=](const TV& X){return TV(vel,0,0);},0,density,r0,false,false);
    e.Add_Fixed_Corotated(E,nu);
    e.Add_Gravity(gravity);
    e.Add_Walls(-1,MPM::BC_SEP,.3,0,false);
    T a=e.reflection_bc_friction*-gravity.y;
    PHYSBAM_ASSERT(vel>=0);
    e.write_output_files.Append([&e,vel,a,C]()
        {
            TV v,x;
            if(a*e.time<=vel)
            {
                v.x=vel-a*e.time;
                x.x=vel*e.time-a*e.time*e.time/2;
            }
            else x.x=vel*vel/(a*2);
            for(auto X:C) Add_Point(X+x,VECTOR<T,3>(1,0,0));
        });
}
template void Function_74(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_74(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
