//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Tools/Interpolation/INTERPOLATION_CURVE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 79
// smash sphere between plates
template<class T> void Function_79(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    if(e.extra_T.m<=0) e.extra_T.Append(0);
    if(e.extra_T.m<=0) e.extra_T.Append(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    SPHERE<TV> sphere(TV()*e.m,.5*e.m);
    T density=2*e.unit_rho*e.scale_mass;
    e.Seed_Particles(sphere,0,0,density,e.particles_per_cell,r0);
    e.Add_St_Venant_Kirchhoff_Hencky_Strain(1e3*e.unit_p*e.scale_E,0.3);
    INTERPOLATION_CURVE<T,FRAME<TV> > ic0,ic1;
    ic0.Add_Control_Point(0,FRAME<TV>());
    ic0.Add_Control_Point(e.extra_T(1),FRAME<TV>(TV(0,e.extra_T(0),0)));
    ic1.Add_Control_Point(0,FRAME<TV>());
    ic1.Add_Control_Point(e.extra_T(1),FRAME<TV>(TV(0,-e.extra_T(0),0)));
    e.Add_Collision_Object(RANGE<TV>(TV(-2,-1.5,-2),TV(2,-.5,2))*e.m,
        MPM::BC_SEP,.1,
        [=](T time){return ic0.Value(time);},
        [=](T time){return ic0.Derivative(time);});
    e.Add_Collision_Object(RANGE<TV>(TV(-2,.5,-2),TV(2,1.5,2))*e.m,
        MPM::BC_SEP,.1,
        [=](T time){return ic1.Value(time);},
        [=](T time){return ic1.Derivative(time);});
}
template void Function_79(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_79(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
