//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Grid_Tools/Grids/NODE_ITERATOR.h>
#include <Geometry/Basic_Geometry/TORUS.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 8
// torus into a box
template<class T> void Function_8(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    // TODO: fix crash ./mpm -3d 8 -affine -last_frame 500 -midpoint -resolution 40 -newton_tolerance 1e-3 
    e.Set_Grid(RANGE<TV>(TV(),TV(1,2,1))*e.m,IV(1,2,1),IV()+1);

    // e.Add_Walls(8,MPM::BC_SEP,.3,.1*e.m,false);

    T thickness=.1;
    e.Add_Collision_Object(RANGE<TV>(TV(.2,0,.2),TV(.8,.1,.8))*e.m,MPM::BC_SEP,.3);
    e.Add_Collision_Object(RANGE<TV>(TV(.2-thickness,0,0.2-thickness),TV(.8+thickness,.4,.2))*e.m,MPM::BC_SEP,.3);
    e.Add_Collision_Object(RANGE<TV>(TV(.2-thickness,0,0.8),TV(.8+thickness,.4,0.8+thickness))*e.m,MPM::BC_SEP,.3);
    e.Add_Collision_Object(RANGE<TV>(TV(0.2-thickness,0,0.2),TV(.2,.4,.8))*e.m,MPM::BC_SEP,.3);
    e.Add_Collision_Object(RANGE<TV>(TV(.8,0,0.2),TV(.8+thickness,.4,.8))*e.m,MPM::BC_SEP,.3);

    T density=5*e.unit_rho*e.scale_mass;
    ANALYTIC_IMPLICIT_OBJECT<TORUS<T> > torus(TORUS<T>(TV(),TV(1,0,0)*e.m,0.02*2*e.m,0.03*2*e.m));
    e.Seed_Particles(torus,0,0,density,e.particles_per_cell,r0);
    int n=e.particles.number;
    GRID<TV> torus_grid(IV(2,3,2)-1,RANGE<TV>(TV(.4,1,.4),TV(.6,1.5,.6))*e.m);
    for(NODE_ITERATOR<TV> iterator(torus_grid);iterator.Valid();iterator.Next()){
        TV center=iterator.Location();
        T angle=e.random.Get_Uniform_Number((T)0,(T)pi*2);
        ROTATION<TV> rotation(angle,TV(0,1,0));
        for(int k=0;k<n;k++)
            e.Add_Particle(center+rotation.Rotate(e.particles.X(k)),0,0,
                e.particles.mass(k),e.particles.volume(k),r0);
        break;
    }
    for(int k=0;k<n;k++) e.particles.Add_To_Deletion_List(k);
    e.particles.Delete_Elements_On_Deletion_List(true);
    e.Add_Fixed_Corotated(150*e.unit_p*e.scale_E,0.4);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.8,0));

}
template void Function_8(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_8(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
