//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/CYLINDER.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 948
// column collapse wedge firction angle
template<class T> void Function_948(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    e.particles.Store_Fp(true);

    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    if(e.use_penalty_collisions){
        e.Add_Penalty_Collision_Object(RANGE<TV>(TV(-0.5,-1,-0.5),TV(1.5,.1,1.5))*e.m,0.9); // ground 
        e.Add_Penalty_Collision_Object(RANGE<TV>(TV(0.06-0.3,0.6-0.5,0.5-0.5),TV(0.06+0.3,0.6+0.5,0.5+0.5))*e.m,0); // xmin
        e.Add_Penalty_Collision_Object(RANGE<TV>(TV(0.94-0.3,0.6-0.5,0.5-0.5),TV(0.94+0.3,0.6+0.5,0.5+0.5))*e.m,0); // xmax
        e.Add_Penalty_Collision_Object(RANGE<TV>(TV(0.5-0.14,0.6-0.5,0.13-0.3),TV(0.5+0.14,0.6+0.5,0.13+0.3))*e.m,0); // zmin
        e.Add_Penalty_Collision_Object(RANGE<TV>(TV(0.5-0.14,0.6-0.5,0.86-0.3),TV(0.5+0.14,0.6+0.5,0.86+0.3))*e.m,0);} // zmax
    else
        e.Add_Collision_Object(RANGE<TV>(TV(-0.5,-1,-0.5),TV(1.5,.1,1.5))*e.m,MPM::BC_NOSLIP,0);

    T density=(T)2200*e.unit_rho*e.scale_mass;
    T E=35.37e6*e.unit_p*e.scale_E,nu=.3;
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
    e.Add_Drucker_Prager_Case(E,nu,0);
    T gap=e.grid.dX(1)*0.1;
    T l0=0.05*e.m;
    T h0=l0*8;
    CYLINDER<T> cylinder(TV(.5*e.m,.1+gap,.5*e.m),TV(.5*e.m,.1*e.m+gap+h0,.5*e.m),l0);
    e.Seed_Particles(cylinder,0,0,density,e.particles_per_cell,r0);
    e.Set_Lame_On_Particles(E,nu);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.81,0));

}
template void Function_948(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_948(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
