//#####################################################################
// Copyright 2015, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __STANDARD_TESTS_BASE__
#define __STANDARD_TESTS_BASE__

#include <Core/Matrices/DIAGONAL_MATRIX.h>
#include <Tools/Interpolation/INTERPOLATION_CURVE.h>
#include <Core/Vectors/TWIST.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Tools/Parsing/PARSE_ARGS.h>
#include <Deformables/Deformable_Objects/DEFORMABLES_STANDARD_TESTS.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_EXAMPLE.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_PARTICLES.h>
#include <functional>

namespace PhysBAM{

template<class TV> class STANDARD_TESTS;
template<class TV> class POISSON_DISK;
template<class TV> class MPM_PARTICLE_SOURCE;
template<class T,int d> class ISOTROPIC_CONSTITUTIVE_MODEL;
template<class TV>
class STANDARD_TESTS_BASE:public MPM_EXAMPLE<TV>
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
    typedef MPM_EXAMPLE<TV> BASE;
    typedef MESH_OBJECT<TV,TV::m> T_VOLUME;

public:
    using BASE::last_frame;using BASE::grid;using BASE::particles;
    using BASE::mass;using BASE::force_helper;
    using BASE::frame_title;using BASE::write_substeps_level;using BASE::gather_scatter;
    using BASE::collision_objects;using BASE::substeps_delay_frame;
    using BASE::viewer_dir;using BASE::plasticity_models;
    using BASE::restart;using BASE::dt;using BASE::time;
    using BASE::frame_dt;using BASE::min_dt;using BASE::max_dt;
    using BASE::compute_sound_speed;using BASE::reflection_bc;
    using BASE::ghost;using BASE::cfl_F;using BASE::use_strong_cfl;
    using BASE::use_symplectic_euler;using BASE::test_output_prefix;
    using BASE::print_stats;using BASE::only_write_particles;using BASE::use_test_output;
    using BASE::cfl;using BASE::newton_tolerance;
    using BASE::newton_iterations;using BASE::solver_tolerance;using BASE::solver_iterations;
    using BASE::test_diff;using BASE::threads;using BASE::weights;
    using BASE::stream_type;
    using BASE::Add_Force;using BASE::Set_Weights;using BASE::solid_body_collection;
    using BASE::Add_Collision_Object;
    using BASE::data_directory;using BASE::reflection_bc_flags;
    using BASE::use_sound_speed_cfl;using BASE::cfl_sound;
    using BASE::side_bc_type;using BASE::Add_Callbacks;using BASE::random;
    using BASE::periodic_test_shift;using BASE::use_periodic_test_shift;
    using BASE::fix_particles_location;using BASE::opt;
    using BASE::suppress_output;
    
    std::string test_name;
    int resolution;
    bool user_resolution;
    int stored_last_frame;
    bool user_last_frame;
    int order;
    int seed;
    T particles_per_cell;
    bool regular_seeding,no_regular_seeding;
    T scale_mass;
    T scale_E;
    T scale_speed;
    T penalty_collisions_stiffness,penalty_collisions_separation,penalty_collisions_length;
    T penalty_damping_stiffness;
    bool use_penalty_collisions;
    bool use_plasticity,use_theta_c,use_theta_s,use_hardening_factor,use_max_hardening;
    T theta_c,theta_s,hardening_factor,max_hardening;
    bool use_implicit_plasticity,no_implicit_plasticity;
    int hardening_mast_case;
    bool use_hardening_mast_case;
    bool override_output_directory;
    T m,s,kg;
    T unit_p,unit_rho,unit_mu;
    int forced_collision_type;
    T friction;
    bool friction_is_set;
    T sigma_Y;
    bool use_cohesion;
    bool use_vc=true;
    std::function<void ()> destroy;
    std::string sph_rel="1k";
    ARRAY<T> extra_T;
    ARRAY<int> extra_int;
    bool dump_collision_objects;
    ARRAY<MPM_PARTICLE_SOURCE<TV>*> sources;
    std::function<void(const ARRAY<int>&)> update_dp_func;
    bool no_mu=false;
    bool use_sand_color=false;
    bool test_dt=false;
    bool print_dim=false,print_order=false,print_pic=false,print_mass=false,print_mu=false,print_lambda=false,print_dx=false;
    T input_mu=0,input_lambda=0,input_E=-1,input_nu=-1;
    int num_substeps=0;


    ARRAY<VECTOR<T,3> > default_region_colors;
    
    POISSON_DISK<TV>& poisson_disk;
    DEFORMABLES_STANDARD_TESTS<TV> tests;

    HASHTABLE<std::string,void(*)(STANDARD_TESTS_BASE&)> test_list;

    STANDARD_TESTS_BASE(const STREAM_TYPE stream_type_input,PARSE_ARGS& parse_args);
    virtual ~STANDARD_TESTS_BASE();

    T Perturb(T a);
    T Uniform(T a,T b);

    void Seed_Particles_Poisson(IMPLICIT_OBJECT<TV>& object,
        std::function<TV(const TV&)> V,
        std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,
        T particles_per_cell,REGION_ID r,
        ARRAY_VIEW<const TV> seeds=ARRAY_VIEW<const TV>(0,0),bool add_seeds=true);

    template<class T_OBJECT>
        typename enable_if<!is_base_of<IMPLICIT_OBJECT<TV>,T_OBJECT>::value>::type
    Seed_Particles_Poisson(const T_OBJECT& object,std::function<TV(const TV&)> V,
        std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,
        T particles_per_cell,REGION_ID r,
        ARRAY_VIEW<const TV> seeds=ARRAY_VIEW<const TV>(0,0),bool add_seeds=true)
    {
        ANALYTIC_IMPLICIT_OBJECT<T_OBJECT> obj(object);
        Seed_Particles_Poisson(obj,V,dV,density,particles_per_cell,r,seeds,add_seeds);
    }

    void Seed_Particles_Uniform(IMPLICIT_OBJECT<TV>& object,
        std::function<TV(const TV&)> V,
        std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,
        const GRID<TV>& seed_grid,REGION_ID r);

    template<class T_OBJECT>
        typename enable_if<!is_base_of<IMPLICIT_OBJECT<TV>,T_OBJECT>::value>::type
    Seed_Particles_Uniform(const T_OBJECT& object,std::function<TV(const TV&)> V,
        std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,
        const GRID<TV>& seed_grid,REGION_ID r)
    {
        ANALYTIC_IMPLICIT_OBJECT<T_OBJECT> obj(object);
        Seed_Particles_Uniform(obj,V,dV,density,seed_grid,r);
    }

    void Seed_Particles(IMPLICIT_OBJECT<TV>& object,std::function<TV(const TV&)> V,
        std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,
        T particles_per_cell,REGION_ID r);

    template<class T_OBJECT>
        typename enable_if<!is_base_of<IMPLICIT_OBJECT<TV>,T_OBJECT>::value>::type
    Seed_Particles(const T_OBJECT& object,std::function<TV(const TV&)> V,
        std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,
        T particles_per_cell,REGION_ID r)
    {
        ANALYTIC_IMPLICIT_OBJECT<T_OBJECT> obj(object);
        Seed_Particles(obj,V,dV,density,particles_per_cell,r);
    }

    template<class T_STRUCTURE>
    T_STRUCTURE& Seed_Lagrangian_Particles(T_STRUCTURE& object,
        std::function<TV(const TV&)> V,
        std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,
        REGION_ID r,bool use_constant_mass=true,bool destroy_after=true);

    template<class T_OBJECT,class... Args> MESH_OBJECT<TV,TV::m-1>&
    Seed_Particles_To_Analytic_Surface(const T_OBJECT& object,
        std::function<TV(const TV&)> V,
        std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,
        T particles_per_cell,REGION_ID r,bool add_structure,Args&&... args)
    {
        auto* tri=Generate_Triangles(object,args...);
        auto& ts=Seed_Lagrangian_Particles(*tri,V,dV,density,r,true,false);
        Seed_Particles_Poisson(object,V,dV,density,particles_per_cell,r,
            tri->particles.X,false);
        delete tri;
        return ts;
    }

    void Add_Penalty_Collision_Object(IMPLICIT_OBJECT<TV>* io,
        const T coefficient_of_friction=(T)0);

    template<class OBJECT> typename enable_if<!is_pointer<OBJECT>::value>::type
    Add_Penalty_Collision_Object(const OBJECT& object,
        const T coefficient_of_friction=(T)0)
    {
        Add_Penalty_Collision_Object(new ANALYTIC_IMPLICIT_OBJECT<OBJECT>(object),
            coefficient_of_friction);
    }

    void Initialize() override;

    int Add_Particle(const TV& X,std::function<TV(const TV&)> V,
        std::function<MATRIX<T,TV::m>(const TV&)> dV,const T mass,
        const T volume,REGION_ID r);
    int Add_Particle(const TV& X,const TV& V,const MATRIX<T,TV::m>& dV,
        const T mass,const T volume,REGION_ID r);
    void Add_Lambda_Particles(ARRAY<int>* affected_particles,T E,T nu,
        T density,REGION_ID r,bool no_mu=false,
        T porosity=(T)1,T saturation_level=(T)1);
    int Add_Gravity(TV g,ARRAY<int>* affected_particles=0);
    int Add_Fixed_Corotated(T E,T nu,
        ARRAY<int>* affected_particles=0,bool no_mu=false);
    int Add_Corotated(T E,T nu,
        ARRAY<int>* affected_particles=0,bool no_mu=false);
    int Add_Neo_Hookean(T E,T nu,ARRAY<int>* affected_particles=0);
    int Add_St_Venant_Kirchhoff_Hencky_Strain(T E,T nu,
        ARRAY<int>* affected_particles=0,bool no_mu=false);
    int Add_Drucker_Prager(T E,T nu,T a0,T a1,T a3,T a4,
        ARRAY<int>* affected_particles=0,bool no_mu=false,T sigma_Y=0);
    int Add_Drucker_Prager(T E,T nu,T phi_F,
        ARRAY<int>* affected_particles=0,bool no_mu=false,T sigma_Y=0);
    int Add_Drucker_Prager_Case(T E,T nu,int case_num_F,
        ARRAY<int>* affected_particles=0,bool no_mu=false);
    // -x +x -y +y [ -z +z ], as bit flags
    void Add_Walls(int flags,MPM::BC_TYPE type,T friction,T inset,
        bool penalty);
    int Add_Clamped_Plasticity(ISOTROPIC_CONSTITUTIVE_MODEL<T,TV::m>& icm,
        T theta_c,T theta_s,T max_hardening,T hardening_factor,
        ARRAY<int>* affected_particles=0);
    void Set_Lame_On_Particles(T E,T nu,ARRAY<int>* affected_particles=0);
    void Set_Grid(const RANGE<TV>& domain,IV resolution_scale=IV()+1,
        int default_resolution=32);
    void Set_Grid(const RANGE<TV>& domain,IV resolution_scale,
        IV resolution_padding,int resolution_multiple=1,
        int default_resolution=32);
    void Add_Source(const TV& source_location,const TV& source_normal,
        T source_radius,T source_speed,const TV& gravity,T density,T E,T nu,
        T start_time,T stop_time,REGION_ID r);

    void Test_Dt();
//#####################################################################
};
}

#endif
