//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Hybrid_Methods/Iterators/PARTICLE_GRID_ITERATOR.h>
#include <Hybrid_Methods/Iterators/PARTICLE_GRID_WEIGHTS_SPLINE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 131singleparticle
template<class TV> void Function_131singleparticle(STANDARD_TESTS_BASE<TV>& e)
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
    REGION_ID r0(0);
    // ./search_max_stable.sh 131singleparticle 3 -no_affine -2d -r 6
    // ./mpm 131singleparticle -explicit -single_particle_cfl -cfl_p 1.5 -order 2 -T .5 -T .5 -resolution 6 -last_frame 10000
    // ./mpm 131singleparticle -explicit -single_particle_cfl -cfl_p 1.5 -order 3 -T .3187942869 -T .3187942869 -resolution 6 -last_frame 10000
    e.random.Set_Seed(0);
    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    TV Xp;
    if(e.extra_T.m>0) Xp(0)=e.extra_T(0)*e.grid.dX(0);
    if(e.extra_T.m>1) Xp(1)=e.extra_T(1)*e.grid.dX(0);
    else Xp(1)=e.extra_T(0)*e.grid.dX(0);
    if constexpr(TV::m==3) if(e.extra_T.m>2) Xp(2)=e.extra_T(2)*e.grid.dX(0);
    T density=1000*e.unit_rho*e.scale_mass; T volume=e.grid.dX.Product()/e.particles_per_cell; T mass=density*volume;
    // oscillation particle
    e.Add_Particle(Xp,0,0,mass,volume,r0);
    MATRIX<T,TV::m> F,C;
    T eps=1e-5;
    e.random.Fill_Uniform(F,-eps,eps);
    e.particles.F.Fill(F+1);
    if(e.opt.use_affine){
        e.random.Fill_Uniform(C,-eps,eps);
        e.particles.C.Fill(C);
    }
    T E=1e7; if(e.input_E!=-1) E=e.input_E;
    T nu=.3; if(e.input_nu!=-1) nu=e.input_nu;
    if(e.extra_int.m==1){
        if(e.extra_int(0)==0)
            e.Add_Corotated(E*e.unit_p*e.scale_E,nu);
        else if(e.extra_int(0)==1)
            e.Add_Fixed_Corotated(E*e.unit_p*e.scale_E,nu);
        else if(e.extra_int(0)==2)
            e.Add_Neo_Hookean(E*e.unit_p*e.scale_E,nu);
        else if(e.extra_int(0)==3)
            e.Add_Drucker_Prager(E*e.unit_p*e.scale_E,nu,(T)35,9,0.3,10);
    }
    else
        e.Add_Neo_Hookean(E*e.unit_p*e.scale_E,nu);

    // e.reflection_bc=true;
    // e.reflection_bc_friction=0;
    // e.Add_Walls(-1,MPM::BC_PERIODIC,0,0,false);
    // TODO use magnitude of C F to analyze instability
    // bool exit_early=true;
    // for(int p=0;p<e.particles.number;p++) if(e.particles.valid(p)) e.simulated_particles.Append(p);
    // T max_ke=e.Total_Particle_Kinetic_Energy();
    // // if(e.extra_int.m>0) exit_early=e.extra_int(0);
    // e.Add_Callbacks(false,"time-step",
    //     [&e,max_ke,exit_early]()
    //     {
    //         T ke=e.Total_Particle_Kinetic_Energy();
    //         LOG::printf("ke %P vs %P\n",ke,max_ke);
    //         if(ke<=1.0000000001*max_ke) return;
    //         LOG::printf("UNSTABLE %P vs %P\n",ke,max_ke);
    //         if(exit_early) exit(0);
    //     }
    // );
    // e.destroy=[&e]()
    // {
    //     LOG::printf("TEST seed %P dt %P particlenum %P FINISHED\n",e.seed,e.max_dt,e.particles.number);
    //     LOG::printf("FINISHED\n");
    // };
}
template void Function_131singleparticle(STANDARD_TESTS_BASE<VECTOR<float,2> >& e); 
template void Function_131singleparticle(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
template void Function_131singleparticle(STANDARD_TESTS_BASE<VECTOR<float,3> >& e); 
template void Function_131singleparticle(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
