//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Core/Math_Tools/pow.h>
#include <Core/Matrices/LAPACK.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Tools/Read_Write/OCTAVE_OUTPUT.h>
#include <Geometry/Seeding/POISSON_DISK.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 131spread 131matrix-1p 131matrix-mixeddt
template<class TV> void Function_131spread(STANDARD_TESTS_BASE<TV>& e)
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
    typedef MATRIX<T,TV::m> TM;
    REGION_ID r0(0);
    // ./search_max_stable.sh -t 131spread -s 3 -a affine -d 2d -T .2
    // ./search_max_stable.sh -t 131spread -s 3 -a affine -d 2d -I 5
    // TODO add ./search_max_stable.sh adjustment for regular_seeding
    // input dt-max_dt  resolution k mu lam (poisson or not)-regular_seeding (periodic needs -explicit)
    // ./mpm 131spread -explicit -fix_particles_loc -mu $mu -lam $lam -T 2 -T 2 -I 5 -resolution 2 -regular_seeding
    // ./mpm 131spread -explicit -fix_particles_loc -mu $mu -lam $lam -T .1 -I 5 -resolution 2
    // ./mpm 131spread -explicit -fix_particles_loc -mu $mu -lam $lam -T poisson_mindist  -I k -resolution res (-regular_seeding)
    // ./mpm 131spread -explicit -fix_particles_loc -mu $mu -lam $lam -T .25 -I 5 -resolution 2 // poisson disk seems wrong under this command
    RANGE<TV> box=RANGE<TV>::Centered_Box()*e.m;
    e.Set_Grid(box);
    T density=1000*e.unit_rho*e.scale_mass;

    int k=0;
    T dt0=.5*e.max_dt,dt1=.5*e.max_dt; int n0=1,n1=1;
    T mass=density*e.grid.dX.Product()/(TV::m==2?4:8); T volume=e.grid.dX.Product()/(TV::m==2?4:8);
    if(e.test_name=="131matrix-1p")
    {
        RANDOM_NUMBERS<T> rand;
        TV Xp; Xp(0)=.25; Xp(1)=.25;
        if(e.extra_T.m>=2) {Xp(0)=e.extra_T(0); Xp(1)=e.extra_T(1);}
        // Xp(0)=rand.Get_Uniform_Number(-1,1);
        // Xp(1)=rand.Get_Uniform_Number(-1,1);
        // if(TV::m==3) Xp(2)=rand.Get_Uniform_Number(-1,1);
        LOG::printf("xploc %P\n",Xp);
        Xp=Xp*e.grid.dX(0);
        e.Add_Particle(Xp,0,0,mass,volume,r0);
    }
    else if(e.test_name=="131matrix-mixeddt")
    {
        // 4 inputs - dt0 dt1 n0 n1
        // currently n0=n1=1 output dt0 dt1 maxroot; matrix
        RANDOM_NUMBERS<T> rand;
        PHYSBAM_ASSERT(e.extra_T.m>=2);
        if(e.extra_T.m>=2) {
            dt0=e.extra_T(0);
            dt1=e.extra_T(1);
            e.max_dt=dt0;}
        LOG::printf("TEST DT0 %P DT1 %P MAXDT %P\n",dt0,dt1,e.max_dt);
        if(e.extra_int.m>=2) {
            n0=e.extra_int(0);
            n1=e.extra_int(1);}
        TV Xp; Xp(0)=.25; Xp(1)=.25;
        rand.Fill_Uniform(Xp,-.5,.5);
        // Xp(0)=rand.Get_Uniform_Number(-1,1);
        // Xp(1)=rand.Get_Uniform_Number(-1,1);
        // if(TV::m==3) Xp(2)=rand.Get_Uniform_Number(-1,1);
        LOG::printf("TEST xploc %P\n",Xp);
        Xp=Xp*e.grid.dX(0);
        e.Add_Particle(Xp,0,0,mass,volume,r0);
    }
    else if(e.test_name=="131spread")
    {
        if(e.extra_int.m)k=e.extra_int(0);
        e.poisson_disk.periodic_boundary=box;
        if(e.extra_T.m)e.poisson_disk.Set_Min_Distance(e.extra_T(0));
        if(k!=0) e.particles_per_cell=pow<TV::m>((T)k/(T)e.resolution);
        // LOG::printf("resolution %P parpcell %P box %P\n",e.resolution,e.particles_per_cell,box);
        // LOG::printf("poisson disk is_periodic %P min_dist %P \n",e.poisson_disk.is_periodic,e.poisson_disk.Min_Distance());
        // LOG::printf("mass %P par_per_cell %P\n",density*e.grid.dX.Product()/4,e.particles_per_cell);
        e.random.Set_Seed(1);
        e.Seed_Particles(box,[&e](const TV& X){TV v;e.random.Fill_Uniform(v,(T)-.1,(T).1);return v;}, 0,density,e.particles_per_cell,r0);
    }

    e.particles.mass.Fill(density*e.grid.dX.Product()/(TV::m==2?4:8));
    e.particles.volume.Fill(e.grid.dX.Product()/(TV::m==2?4:8));
    T E=1e6,nu=.3;
    if(e.input_E!=-1) E=e.input_E;
    if(e.input_nu!=-1) nu=e.input_nu;
    e.Add_Neo_Hookean(E*e.unit_p*e.scale_E,nu);
    e.reflection_bc=true;
    e.reflection_bc_friction=0;
    e.Add_Walls(-1,MPM::BC_PERIODIC,0,0,false);

    int* column_count=new int{};
    int* column_count_extended=new int{};
    int point_dim=TV::m*(1+TV::m+(e.opt.use_affine?TV::m:0));
    int matrix_dim=e.particles.X.m*point_dim;
    // e.frame_dt=e.max_dt*(n0+n1)*(matrix_dim+10);
    e.frame_dt=e.max_dt;
    e.min_dt=e.max_dt;
    MATRIX_MXN<T>* output_matrix=new MATRIX_MXN<T>(matrix_dim,matrix_dim);
    T eps=1e-6;
    e.Add_Callbacks(true,"time-step",
        // [&e,eps,column_count,output_matrix,point_dim,matrix_dim](){
        [=,&e](){
            auto cc=*column_count;
            auto cce=*column_count_extended;
            auto om=*output_matrix;
            if(e.test_name=="131matrix-mixeddt")
            {
                // if cce not moded don't set column
                if(cce%(n0+n1)!=0)
                    return;
            }
            if(cc>=om.m)
            {
                ARRAY<std::complex<T>> roots(matrix_dim);
                OCTAVE_OUTPUT<T>("matrix.txt").Write("M",om);
                Compute_Eigenvalues(om,roots);
                std::complex<T> max_root=0;
                for(auto i:roots){
                    if(abs(max_root)<abs(i))
                        max_root=i;
                }
                LOG::printf("output_matrix\n %P\n roots %P\n TEST abs_root %P max_root %P \n",om,roots,abs(max_root),max_root);
                if(abs(max_root)>1+1e-5){
                    LOG::printf("FAIL");
                }
                else{
                    LOG::printf("FINISHED");
                }
                delete column_count;
                delete output_matrix;
                exit(0);
            }
            // init all value to be 0s with one 1
            int p_index=cc/point_dim;
            int ind_in_p=cc%point_dim;
            TV tvel; TM tF,tC;
            if(ind_in_p<TV::m) tvel(ind_in_p)=eps;
            else if(ind_in_p<TV::m+TV::m*TV::m) tF.x[ind_in_p-TV::m]=eps;
            else tC.x[ind_in_p-(TV::m+TV::m*TV::m)]=eps;// /e.max_dt;
            e.particles.V.Fill(TV());
            e.particles.F.Fill(TM()+1);
            if(e.opt.use_affine)e.particles.C.Fill(TM());
            // LOG::printf("p_index %P ind_in_p %P column_count %P output_matrix->m %P  \n",p_index,ind_in_p,cc,output_matrix->m);
            e.particles.V(p_index)=tvel;
            e.particles.F(p_index)=tF+1;
            if(e.opt.use_affine)e.particles.C(p_index)=tC;
            if(e.opt.use_affine) LOG::printf("TEST input %P %P %P\n",e.particles.V,e.particles.F,e.particles.C);
            else LOG::printf("TEST input %P %P\n",e.particles.V,e.particles.F);
        });

    e.Add_Callbacks(false,"time-step",
        // [&e,eps,column_count,output_matrix,point_dim,matrix_dim](){
        [=,&e](){
            auto& cc=*column_count;
            auto& cce=*column_count_extended;
            auto& om=*output_matrix;
            ARRAY<T> output_column(matrix_dim);
            for(int i=0;i<e.particles.X.m;i++)
            {
                // for each for loop we extract particles value to matrix
                TV tvel=e.particles.V(i)/eps;
                TM tF=(e.particles.F(i)-1)/eps,tC;
                if(e.opt.use_affine)tC=e.particles.C(i)/eps; // *e.max_dt;
                int p_start_index=i*point_dim;
                LOG::printf("TEST matrix_dim %P p_start_index %P point_dim %P particlenum %P  \n",matrix_dim,p_start_index,point_dim,e.particles.X.m);
                for(int j=0;j<TV::m;j++){output_column(p_start_index+j)=tvel(j);}
                for(int j=0;j<TV::m*TV::m;j++){output_column(p_start_index+TV::m+j)=tF.x[j];}
                if(e.opt.use_affine) for(int j=0;j<TV::m*TV::m;j++){output_column(p_start_index+TV::m+TV::m*TV::m+j)=tC.x[j];}
            }
            if(e.test_name=="131matrix-mixeddt")
            {
                cce++;
                LOG::printf("TEST cce %P output column %P \n",cce,output_column);
                if(cce%(n0+n1)==0)
                {
                    om.Set_Column(cc,output_column);
                    cc++;
                }
                // after cce change change timestep for next timestep
                if(cce%(n0+n1)<n0)
                    e.max_dt=dt0;
                else
                    e.max_dt=dt1;
                e.min_dt=e.max_dt;
                e.frame_dt=e.max_dt;
                LOG::cout<<"TEST dt:"<<e.max_dt<<std::endl;
                return;
            }
            om.Set_Column(cc,output_column);
            cc++;
        });

}
template void Function_131spread(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_131spread(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
template void Function_131spread(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_131spread(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
