//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Core/Matrices/MATRIX_MXN.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 132
// reuse test 131 parameters
// dt=xxx; ./mpm 132 -test_dt -no_affine -order 2 -explicit -T 1e-6 -min_dt $dt -max_dt $dt -mu $mu -lam $lam | grep -i test
template<class TV> void Function_132(STANDARD_TESTS_BASE<TV>& e)
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
    REGION_ID r0(0);
    // v*2 F*4 C*4
    typedef MATRIX<T,TV::m> TM;

    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    T density=1010*e.unit_rho*e.scale_mass;
    T volume=e.grid.dX.Product()/e.particles_per_cell;
    T mass=density*volume;
    TV j0=-.1*e.grid.dX;
    // e.Add_Particle(j0,[](const TV& ){return -TV();},0,mass,volume,r0);
    // e.Add_Particle(-j0,[](const TV& ){return TV();},0,mass,volume,r0);
    e.Add_Particle(TV(),[](const TV& ){return TV();},0,mass,volume,r0);
    e.Add_Neo_Hookean(0.073e9*e.unit_p*e.scale_E,0.4);
    // e.Add_Neo_Hookean(0*e.scale_E,0.4);
    e.frame_dt=100;
    e.use_symplectic_euler=true;
    // e.Add_Walls(-1,MPM::BC_SLIP,0.01,e.grid.dX.Min()*e.m,false);

    int* column_count=new int{};
    int size=TV::m+TV::m*TV::m*(1+e.opt.use_affine);
    MATRIX_MXN<T>* output_matrix=new MATRIX_MXN<T>(size,size);
    T eps=1e-9;
    if(e.extra_T.m>0) eps=e.extra_T(0);
    PHYSBAM_ASSERT(sizeof(T)==sizeof(double));
    // column row are different TODO
    // j*m+i

    e.Add_Callbacks(true,"time-step",[&e,column_count,output_matrix,eps]()
    {
        LOG::printf("test in callback before timestep\n");
        auto cc=*column_count;
        if(cc>=output_matrix->m)
        {
            delete column_count;
            delete output_matrix;
            exit(0);
        }
        // LOG::printf("test in callback before timestep\ntest input %P \n",input);
        // input to v f c
        TV tvel; TM tF,tC;
        if(cc<TV::m) tvel(cc)=eps;
        else if(cc<TV::m+TV::m*TV::m) tF.x[cc-TV::m]=eps;
        else tC.x[cc-(TV::m+TV::m*TV::m)]=eps;
        
        // tvel=TV(1,1)*eps;
        // tF.x[0]=  6.8857e-03*eps;
        // tF.x[1]=-4.2571e-02 *eps;
        // tF.x[2]= 7.0609e-02 *eps;
        // tF.x[3]=  4.4748e-03*eps;
        // manipulate parameters for particle 0 - v F C
        e.particles.V(0) = tvel;
        e.particles.F(0) = tF+1;
        if(e.opt.use_affine)e.particles.C(0) = tC;
        e.particles.V(1) = -tvel;
        e.particles.F(1) = tF+1;
        if(e.opt.use_affine)e.particles.C(1) = tC;

        LOG::printf("test pV %P pF  %P  pC %P\n",e.particles.V(0),e.particles.F(0),(e.opt.use_affine)?e.particles.C(0):MATRIX<T,TV::m>());
    });

    e.Add_Callbacks(false,"time-step",[&e,column_count,output_matrix,eps](){
// calculate the output vector
        auto& cc=*column_count;
        auto& om=*output_matrix;
        LOG::printf("test in callback after timestep \ncolumn count %P particles num %P \n",cc,e.particles.X.m);
        LOG::printf("test end timestep pV %P pF  %P  pC %P\n",e.particles.V(0),e.particles.F(0),(e.opt.use_affine)?e.particles.C(0):MATRIX<T,TV::m>());
        // record column
        TV tvel=e.particles.V(0)/eps;
        TM tF=(e.particles.F(0)-1)/eps,tC;
        if(e.opt.use_affine) tC=e.particles.C(0)/eps;
        ARRAY<T> output_column(om.m);
        for(int i=0;i<TV::m;i++) output_column(i)=tvel(i);
        for(int i=0;i<TV::m*TV::m;i++) output_column(i+TV::m)=tF.x[i];
        if(e.opt.use_affine) for(int i=0;i<TV::m*TV::m;i++) output_column(i+TV::m+TV::m*TV::m)=tC.x[i];
        LOG::printf("test output_column %P\n",output_column);
        // write column to matrix
        om.Set_Column(cc,output_column);
        LOG::printf("test column_count %P output_matrix %.16P\n",cc,om);

        cc++;
    });
}
template void Function_132(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_132(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
template void Function_132(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_132(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
