//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 1test 1testsand
// rotating circle
template<class TV> void Function_1test(STANDARD_TESTS_BASE<TV>& e)
{
    // running commands
    // #cs && ./mpm 1test -explicit -last_frame 500 -print_stats -T .47 -T .2| grep energy     # shows instability is near .3
    // cs && ./mpm 1test -explicit -last_frame 500 -print_stats -T  .3 -T .02| grep energy     # shows instability always along x/y axis

    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    T raid=.3,cut=.02;
    if(e.extra_T.m==2){
        raid=e.extra_T(0);
        cut=e.extra_T(1);
    }
    TV center; center.Fill(.5);
    SPHERE<TV> sphere(center*e.m,raid*e.m);
    VECTOR<T,TV::m==2?1:3> angular_velocity;
    angular_velocity(0)=.4/e.s;
    T density=2000*e.unit_rho*e.scale_mass;
    e.Seed_Particles(sphere,
        [=](const TV& X){return angular_velocity.Cross(X-sphere.center);},
        [=](const TV&){return MATRIX<T,TV::m>::Cross_Product_Matrix(angular_velocity);},
        density,e.particles_per_cell,r0);
    for(int i=0;i<e.particles.X.m;i++){
        if((e.particles.X(i)-sphere.center).Magnitude()<sphere.radius-cut){
            e.particles.deletion_list.Append(i);
        }
    }
    e.particles.Delete_Elements_On_Deletion_List();
    T total_mass=e.particles.mass.Sum();
    TV total_momentum=e.particles.V.Weighted_Sum(e.particles.mass);
    TV dV=total_momentum/total_mass;
    e.particles.V-=dV;
    T E=1e3*e.unit_p*e.scale_E;
    T nu=0.3;
    if(e.input_nu!=-1) nu=e.input_nu;
    LOG::printf("input nu %P nu %P\n",e.input_nu,nu);

    if(e.test_name=="1testsand")
    {
        // if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
        if(!e.friction_is_set) e.friction=.3;
        E=35.37e6*e.unit_p*e.scale_E;nu=.3;
        ARRAY<int> sand_particles(e.particles.X.m);
        for(int p=0;p<e.particles.X.m;p++){
            sand_particles(p)=p;
            e.particles.X(p)(1)-=.2;
            e.particles.V(p)(1)-=.2;
        }
        e.Add_Drucker_Prager(E,nu,(T)35,9,0.3,10,&sand_particles);
        e.Add_Walls(-1,MPM::BC_SEP,e.friction,0,false);
        // TV gravity; gravity(1)-=9.81;
        // e.Add_Gravity(e.m/(e.s*e.s)*gravity);
    }
    else
    {
        e.Add_Neo_Hookean(E,nu);
        e.reflection_bc=true;
        e.reflection_bc_friction=0;
        // e.Add_Walls(-1,MPM::BC_PERIODIC,0,0,false);
        e.Add_Walls(-1,MPM::BC_SEP,0,0,false);
    }

    int exit_early=0;
    if(e.extra_int.m) exit_early=e.extra_int(0);
    T* global_max_ui=new T();
    T* last_max_up=new T();
    *last_max_up=e.particles.V.Maximum_Magnitude();
    bool* flag=new bool(false);
    e.Add_Callbacks(false,"g2p",[&e,global_max_ui,last_max_up,flag,exit_early](){
                             T max_up=0;
                             auto& lmup=*last_max_up;
                             max_up=e.particles.V.Maximum_Magnitude();
                             T mul=2;
                             if(e.test_name=="1testsand") mul=1.05;
                             if(max_up>mul*lmup){
                                 LOG::printf("red unstable\n");
                                 if(exit_early==1) {
                                     delete global_max_ui;
                                     delete flag;
                                     exit(0);
                                 }
                             }
                             if(e.test_name=="1testsand") lmup=max_up;
                         });
    e.destroy=[&e]()
              {
                  LOG::printf("FINISHED\n");
              };
}
template void Function_1test(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_1test(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
template void Function_1test(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_1test(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
