//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 51
// periodic, full domain, perturbed
template<class TV> void Function_51(STANDARD_TESTS_BASE<TV>& e)
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
    T eps=1e-4;
    REGION_ID r0(0);
    e.Set_Grid(RANGE<TV>::Unit_Box()*e.m);
    T density=pow<TV::m>(10)*e.unit_rho*e.scale_mass;
    e.Seed_Particles(e.grid.domain,0,0,density,e.particles_per_cell,r0);
    e.random.Fill_Uniform(e.particles.V,-eps,eps);
    e.random.Fill_Uniform(e.particles.C,-eps,eps);
    e.random.Fill_Uniform(e.particles.F,-eps,eps);
    for(auto&f:e.particles.F) f+=1;

    T E=1e3*e.unit_p*e.scale_E;
    T nu=0.3;
    if(e.input_nu!=-1) nu=e.input_nu;

    e.Add_Neo_Hookean(E,nu);
    e.reflection_bc=true;
    e.reflection_bc_friction=0;
    e.Add_Walls(-1,MPM::BC_PERIODIC,0,0,false);

    int exit_early=0;
    if(e.extra_int.m) exit_early=e.extra_int(0);
    e.Add_Callbacks(false,"g2p",
        [&e,exit_early,eps]()
        {
            T max_up=e.particles.V.Maximum_Magnitude();
            if(max_up>eps*10)
            {
                LOG::printf("red unstable\n");
                if(exit_early==1) exit(0);
            }
        });
    e.destroy=[&e](){LOG::printf("FINISHED\n");};
}
template void Function_51(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_51(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
template void Function_51(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_51(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
