#!/bin/bash

get_actual_sound_cfl()
{
    local CMD="$1"
    local ns="$2"
    local output=$($CMD | head -n 50)
    # echo output is "$output"
    local sound_dt=$(echo "$output" | grep -m 1 -e dx/soundspeed | awk '{print $2}')
    local actual_dt=$(echo "$output" | grep -m 1 -e actual | awk '{print $3}')
    local single_dt=$(echo "$output" | grep -m 1 -e Ratio | awk '{print $6}')
    ascfl=$(echo "$actual_dt/$sound_dt" | bc -l )
    sscfl=$(echo "$single_dt/$sound_dt" | bc -l )
    a1pcfl=$(echo "$actual_dt/$single_dt" | bc -l )
    echo ns $ns ascfl $ascfl a1pcfl $a1pcfl sscfl $sscfl
}

repeat()
{
    (
    for i in {1..80}; do echo -n $1;done
    echo
    )
}

# render them
renderA2G_indir()
{
    (
        DIR=$1 # should be out_* with many output folder inside
        scale=; scale=$2
        if [[ -z $scale ]] ; then scale=50000; fi
        cd $DIR
        RENDER=$PHYSBAM/Tools/dump_particle/render_sim_2d

        for s in * ; do
            $RENDER -i $s -clip '-1 -1 1 1' \
                    -use_part -part_size .003 -part_v_scale .5 -part_v_arrow -part_v_col_prog 'm=sqrt(u*u+v*v);r=m*10;g=0;b=0;' -part_col_prog 'm=sqrt(u*u+v*v);r=m*10;g=0;b=0;' -part_v_min_len 1e-7 \
                    -use_grid -use_u  -u_scale ${scale}  -u_col '0 1 0' -u_arrow & # -u_col_prog 'm=sqrt(u*u+v*v);r=0;g=m*10000000;b=0;'
            # break
        done
        wait
        echo Finished generating latex files

        for s in */tex ; do
            (
                cd $s
                ls frame-*.tex | xargs -n 1 -P 32 pdflatex -interaction=nonstopmode >/dev/null
                #  -halt-on-error
                echo Finished pdflatex $s
                for i in *.pdf ; do
                    convert -density 175.6 -resize 720 -background '#e0e0e0' -colorspace RGB -flatten $i ${i/pdf/png} &
                done
                wait
                echo Finished convert $s
                ffmpeg -y -i frame-%04d.png -pix_fmt yuv420p -vcodec libx264 -b 4096k movie.mov
                # -f mov
                # ffmpeg -y -i frame-%04d.png -pix_fmt yuv420p -vcodec libx264 -b:v 4096k movie.mov #
                echo Finished rendering $s
            )
            # break
        done
    )

}

function play_videos()
{
    (
        DIR=$1 # should be out_* with many output folder inside
        cd $DIR
        for i in $(find -iname "*.mov"); do repeat -; echo  $i; mplayer -really-quiet $i; sleep 1; done # DEBUG
    )
}

function make_mov()
{
    (
    ffmpeg -y -i $1 -pix_fmt yuv420p -vcodec libx264 -b:v 4096k $2
    )
}

