//#####################################################################
// Copyright 2017, Ounan Ding, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Math_Tools/RANGE_ITERATOR.h>
namespace PhysBAM{
// These tables are generated by this program:
// Projects/grid_iterators/gen_range_iterator_table.cpp
static const long default_table[2]=
{
    0000000000000666000L
};
static const long ghost_table[2]=
{
    0244022624002662000L,
    0644422664042666004L
};
static const long delay_corners_table[2]=
{
    0266000426200442220L,
    0666400466240446224L
};
static const long duplicate_corners_table[2]=
{
    0266000626000662000L,
    0666400666040666004L
};
static const long omit_corners_table[2]=
{
    0244022424202442220L,
    0644422464242446224L
};
static const long clamp_mask=0000001000001000001L;
//#####################################################################
// Constructor
//#####################################################################
template<int d> RANGE_ITERATOR<d>::
RANGE_ITERATOR(const RANGE<IV>& outer,const RANGE<IV>& inner,
    RI flags,int side_input)
    :vecs{outer.min_corner,inner.min_corner,inner.max_corner,outer.max_corner}
{
    Initialize(flags,side_input);
}
//#####################################################################
// Constructor
//#####################################################################
template<int d> RANGE_ITERATOR<d>::
RANGE_ITERATOR(const RANGE<IV>& range,int outer_ghost,int inner_ghost,
    RI flags,int side_input)
    :vecs{range.min_corner-outer_ghost,range.min_corner-inner_ghost,
        range.max_corner+inner_ghost,range.max_corner+outer_ghost}
{
    Initialize(flags,side_input);
}
//#####################################################################
// Constructor
//#####################################################################
template<int d> RANGE_ITERATOR<d>::
RANGE_ITERATOR(const IV& counts,int outer_ghost,int inner_ghost,
    RI flags,int side_input)
    :vecs{IV()-outer_ghost,IV()-inner_ghost,
        counts+inner_ghost,counts+outer_ghost}
{
    Initialize(flags,side_input);
}
//#####################################################################
// Constructor
//#####################################################################
template<int d> RANGE_ITERATOR<d>::
RANGE_ITERATOR(const RANGE<IV>& range,RI flags)
    :vecs{range.min_corner,range.min_corner,range.max_corner,range.max_corner}
{
    Initialize(flags,-1);
}
//#####################################################################
// Constructor
//#####################################################################
template<int d> RANGE_ITERATOR<d>::
RANGE_ITERATOR(const IV& counts,int outer_ghost,RI flags)
    :vecs{IV()-outer_ghost,IV()-outer_ghost,
        counts+outer_ghost,counts+outer_ghost}
{
    Initialize(flags,-1);
}
//#####################################################################
// Function Set_Range
//#####################################################################
template<int d> void RANGE_ITERATOR<d>::
Set_Range(const RANGE<IV>& outer,const RANGE<IV>& inner)
{
    vecs[0]=outer.min_corner;
    vecs[1]=inner.min_corner;
    vecs[2]=inner.max_corner;
    vecs[3]=outer.max_corner;
}
//#####################################################################
// Function Set_Range
//#####################################################################
template<int d> void RANGE_ITERATOR<d>::
Set_Range(const RANGE<IV>& range,int outer_ghost,int inner_ghost)
{
    vecs[0]=range.min_corner-outer_ghost;
    vecs[1]=range.min_corner-inner_ghost;
    vecs[2]=range.max_corner+inner_ghost;
    vecs[3]=range.max_corner+outer_ghost;
}
//#####################################################################
// Function Set_Range
//#####################################################################
template<int d> void RANGE_ITERATOR<d>::
Set_Range(const IV& counts,int outer_ghost,int inner_ghost)
{
    vecs[0]=IV()-outer_ghost;
    vecs[1]=IV()-inner_ghost;
    vecs[2]=counts+inner_ghost;
    vecs[3]=counts+outer_ghost;
}
//#####################################################################
// Function Initialize
//#####################################################################
template<int d> void RANGE_ITERATOR<d>::
Initialize(RI flags,int side_input)
{
    Encode(flags,side_input);
    Reset(flags);
}
//#####################################################################
// Function Reset
//#####################################################################
template<int d> void RANGE_ITERATOR<d>::
Reset(RI flags)
{
    if(!(flags&RI::reverse)){
        side=-1;
        Next_Side();}
    else{
        side=2*d;
        Prev_Side();}
}

// Search for x&~1 and x|1 together.
static long find_pair(long in,long mask,int x)
{
    long y=~(in^(mask*x)); // if octal character matches, put 7 at that spot.
    return (y>>1)&(y>>2)&mask; // Test for octal 7.
};

//#####################################################################
// Function Encode
//#####################################################################
template<int d> void RANGE_ITERATOR<d>::
Encode(RI flags,int side_input)
{
    if(any(flags&RI::side_mask)) side_mask=side_input&((1<<2*d)-1);
    else side_mask=side_input<0?(1<<2*d)-1:(1<<side_input);
    flags=flags&~RI::reverse;
    if(!any(flags&RI::ghost)) side_mask=1;

    // Explicitly check for flag combinations that do not make sense.
    if(!any(flags&RI::ghost)){
        assert(!(flags&~(RI::ghost)));}
    if(any(flags&RI::duplicate_corners)){
        assert(!(flags&~(RI::duplicate_corners|RI::ghost|RI::side_mask)));}
    if(any(flags&RI::omit_corners)){
        assert(!(flags&~(RI::omit_corners|RI::ghost|RI::side_mask)));}
    if(any(flags&RI::delay_corners)){
        assert(!(flags&~(RI::delay_corners|RI::ghost|RI::side_mask)));}

    const long* table=0;
    if(!(flags&RI::ghost)) table=default_table;
    else if(any(flags&RI::delay_corners)) table=delay_corners_table;
    else if(any(flags&RI::duplicate_corners)) table=duplicate_corners_table;
    else if(any(flags&RI::omit_corners)) table=omit_corners_table;
    else table=ghost_table;
    for(int i=0;i<2;i++) inst[i]=table[i];

    for(int a=0;a<d;a++)
    {
        if(vecs[0](a)>=vecs[1](a))
            for(auto&m:inst)
                m-=2*find_pair(m,clamp_mask<<3*a,2); // 2 -> 0, 3 -> 1
        if(vecs[2](a)>=vecs[3](a))
            for(auto&m:inst)
                m+=2*find_pair(m,clamp_mask<<3*(a+3),4); // 4 -> 6, 5 -> 7
    }
}
//#####################################################################
// Function Next_Helper
//#####################################################################
template<int d> void RANGE_ITERATOR<d>::
Next_Helper()
{
    // next row of range
    index(d-1)=current[0](d-1);
    for(int i=d-2;i>=0;i--){
        if(++index(i)<current[1](i)) return;
        index(i)=current[0](i);}
    Next_Side();
}
//#####################################################################
// Function Prev_Helper
//#####################################################################
template<int d> void RANGE_ITERATOR<d>::
Prev_Helper()
{
    // next row of range
    index(d-1)=current[1](d-1)-1;
    for(int i=d-2;i>=0;i--){
        if(--index(i)>=current[0](i)) return;
        index(i)=current[1](i)-1;}
    Prev_Side();
}
//#####################################################################
// Function Next_Side
//#####################################################################
template<int d> void RANGE_ITERATOR<d>::
Next_Side()
{
    // finished current block; next block
    side++;
    if(!(side_mask>>side)){
        index(d-1)=current[1](d-1); // Make invalid
        return;} // Done
    while(!(side_mask&(1<<side))) side++;
    Fill_Current();
    index=current[0];
    if(!current[0].All_Less(current[1])) Next_Side();
}
//#####################################################################
// Function Prev_Side
//#####################################################################
template<int d> void RANGE_ITERATOR<d>::
Prev_Side()
{
    // finished current block; next block
    side--;
    if(side<0 || !(side_mask&((2<<side)-1))){
        index(d-1)=current[0](d-1)-1; // Make invalid
        return;} // Done
    while(!(side_mask&(1<<side))) side--;
    Fill_Current();
    index=current[1]-1;
    if(!current[0].All_Less(current[1])) Prev_Side();
}
//#####################################################################
// Function Fill_Current
//#####################################################################
template<int d> void RANGE_ITERATOR<d>::
Fill_Current()
{
    // Assemble next block from rules.
    long mask=inst[side%2]>>(side/2*18);
    for(int k=0;k<2;k++)
        for(int a=0;a<d;a++)
        {
            int m=(mask>>(3*(3*k+a)))&7;
            current[k](a)=vecs[m>>1](a)+(m&1);
        }
}
//#####################################################################
template class RANGE_ITERATOR<1>;
template class RANGE_ITERATOR<2>;
template class RANGE_ITERATOR<3>;
}
