//#####################################################################
// Copyright 2006-2009, Jon Gretarsson, Geoffrey Irving, Nipun Kwatra, Craig Schroeder, Andrew Selle, Tamar Shinar, Eftychios Sifakis, Jonathan Su, Wen Zheng.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class DEFORMABLES_STANDARD_TESTS
//#####################################################################
#include <Deformables/Deformable_Objects/DEFORMABLES_STANDARD_TESTS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Geometry/Basic_Geometry/CYLINDER.h>
#include <Geometry/Basic_Geometry/RANGE_TESSELLATION.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include <Geometry/Basic_Geometry/TETRAHEDRON.h>
#include <Geometry/Basic_Geometry/TORUS.h>
#include <Geometry/Meshes/MESH_OBJECT.h>
#include <Deformables/Deformable_Objects/DEFORMABLE_PARTICLES.h>
#include <Geometry/Grids_Uniform_Computations/SEGMENTED_CURVE_2D_SIGNED_DISTANCE.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_TRANSFORMED.h>
#include <Geometry/Implicit_Objects/LEVELSET_IMPLICIT_OBJECT.h>
#include <Geometry/Intersections/SEGMENT_3D_TRIANGLE_3D_INTERSECTION.h>
#include <Geometry/Meshes/SEGMENTED_CURVE.h>
#include <Geometry/Meshes/TETRAHEDRALIZED_VOLUME.h>
#include <Geometry/Spatial_Acceleration/MESH_HIERARCHY.h>
#include <Deformables/Deformable_Objects/DEFORMABLE_BODY_COLLECTION.h>
using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class TV> DEFORMABLES_STANDARD_TESTS<TV>::
DEFORMABLES_STANDARD_TESTS(DEFORMABLE_BODY_COLLECTION<TV>& deformable_body_collection_input)
    :deformable_body_collection(deformable_body_collection_input)
{
}
//#####################################################################
// Function Copy_And_Add_Structure
//#####################################################################
template<class TV> template<class T_STRUCTURE> T_STRUCTURE& DEFORMABLES_STANDARD_TESTS<TV>::
Copy_And_Add_Structure(T_STRUCTURE& structure,ARRAY<int>* particle_indices,bool delete_structure)
{
    deformable_body_collection.Add_Structure(structure.Append_Particles_And_Create_Copy(deformable_body_collection.particles,particle_indices));
    deformable_body_collection.structures.Last()->Update_Number_Nodes();
    if(delete_structure) delete &structure;
    return dynamic_cast<T_STRUCTURE&>(*deformable_body_collection.structures.Last());
}
//#####################################################################
// Function Set_Mass_Of_Particles 
//#####################################################################
template<class TV> template<class T_OBJECT> void DEFORMABLES_STANDARD_TESTS<TV>::
Set_Mass_Of_Particles(const T_OBJECT& object,const T density,const bool use_constant_mass)
{
    PHYSBAM_ASSERT(density>0);
    if(!object.mesh.elements.m) return;
    DEFORMABLE_PARTICLES<TV>& particles=dynamic_cast<DEFORMABLE_PARTICLES<TV>&>(object.particles);
    particles.Store_Mass();
    ARRAY<int> nodes;
    Get_Unique(nodes,object.mesh.elements.Flattened());
    if(use_constant_mass&&nodes.m){
        T mass_per_node=density*object.Total_Size()/nodes.m;
        particles.mass.Subset(nodes).Fill(mass_per_node);}
    else{
        particles.mass.Subset(nodes).Fill((T)0);
        int nodes_per_element=object.mesh.elements(0).m;
        T density_scaled=density/(T)nodes_per_element;
        for(int t=0;t<object.mesh.elements.m;t++){
            T mass_scaled=density_scaled*object.Signed_Size(t);
            assert(mass_scaled>0);
            particles.mass.Subset(object.mesh.elements(t))+=mass_scaled;}}
}
//#####################################################################

#define INSTANTIATION_HELPER(T) \
    template TETRAHEDRALIZED_VOLUME<T>& DEFORMABLES_STANDARD_TESTS<VECTOR<T,3> >::Copy_And_Add_Structure(TETRAHEDRALIZED_VOLUME<T>&,ARRAY<int>*,bool); \
    template TRIANGULATED_SURFACE<T>& DEFORMABLES_STANDARD_TESTS<VECTOR<T,3> >::Copy_And_Add_Structure(TRIANGULATED_SURFACE<T>&,ARRAY<int>*,bool); \
    template TRIANGULATED_AREA<T>& DEFORMABLES_STANDARD_TESTS<VECTOR<T,2> >::Copy_And_Add_Structure(TRIANGULATED_AREA<T>&,ARRAY<int>*,bool); \
    template SEGMENTED_CURVE<VECTOR<T,2> >& DEFORMABLES_STANDARD_TESTS<VECTOR<T,2> >::Copy_And_Add_Structure(SEGMENTED_CURVE<VECTOR<T,2> >&,ARRAY<int>*,bool); \
    template void DEFORMABLES_STANDARD_TESTS<VECTOR<T,2> >::Set_Mass_Of_Particles<SEGMENTED_CURVE<VECTOR<T,2> > >(SEGMENTED_CURVE<VECTOR<T,2> > const&,T,bool); \
    template void DEFORMABLES_STANDARD_TESTS<VECTOR<T,3> >::Set_Mass_Of_Particles<SEGMENTED_CURVE<VECTOR<T,3> > >(SEGMENTED_CURVE<VECTOR<T,3> > const&,T,bool);

INSTANTIATION_HELPER(float);
template void DEFORMABLES_STANDARD_TESTS<VECTOR<float,3> >::Set_Mass_Of_Particles<TRIANGULATED_SURFACE<float> >(TRIANGULATED_SURFACE<float> const&,float,bool);
template void DEFORMABLES_STANDARD_TESTS<VECTOR<float,2> >::Set_Mass_Of_Particles<TRIANGULATED_AREA<float> >(TRIANGULATED_AREA<float> const&,float,bool);
template void DEFORMABLES_STANDARD_TESTS<VECTOR<float,3> >::Set_Mass_Of_Particles<TETRAHEDRALIZED_VOLUME<float> >(TETRAHEDRALIZED_VOLUME<float> const&,float,bool);
template void DEFORMABLES_STANDARD_TESTS<VECTOR<float,1> >::Set_Mass_Of_Particles<SEGMENTED_CURVE<VECTOR<float,1> > >(SEGMENTED_CURVE<VECTOR<float,1> > const&,float,bool);
INSTANTIATION_HELPER(double);
template void DEFORMABLES_STANDARD_TESTS<VECTOR<double,3> >::Set_Mass_Of_Particles<TRIANGULATED_SURFACE<double> >(TRIANGULATED_SURFACE<double> const&,double,bool);
template void DEFORMABLES_STANDARD_TESTS<VECTOR<double,2> >::Set_Mass_Of_Particles<TRIANGULATED_AREA<double> >(TRIANGULATED_AREA<double> const&,double,bool);
template void DEFORMABLES_STANDARD_TESTS<VECTOR<double,3> >::Set_Mass_Of_Particles<TETRAHEDRALIZED_VOLUME<double> >(TETRAHEDRALIZED_VOLUME<double> const&,double,bool);
template void DEFORMABLES_STANDARD_TESTS<VECTOR<double,1> >::Set_Mass_Of_Particles<SEGMENTED_CURVE<VECTOR<double,1> > >(SEGMENTED_CURVE<VECTOR<double,1> > const&,double,bool);
template DEFORMABLES_STANDARD_TESTS<VECTOR<double,1> >::DEFORMABLES_STANDARD_TESTS(DEFORMABLE_BODY_COLLECTION<VECTOR<double,1> >&);
template DEFORMABLES_STANDARD_TESTS<VECTOR<double,2> >::DEFORMABLES_STANDARD_TESTS(DEFORMABLE_BODY_COLLECTION<VECTOR<double,2> >&);
template DEFORMABLES_STANDARD_TESTS<VECTOR<double,3> >::DEFORMABLES_STANDARD_TESTS(DEFORMABLE_BODY_COLLECTION<VECTOR<double,3> >&);
template DEFORMABLES_STANDARD_TESTS<VECTOR<float,1> >::DEFORMABLES_STANDARD_TESTS(DEFORMABLE_BODY_COLLECTION<VECTOR<float,1> >&);
template DEFORMABLES_STANDARD_TESTS<VECTOR<float,2> >::DEFORMABLES_STANDARD_TESTS(DEFORMABLE_BODY_COLLECTION<VECTOR<float,2> >&);
template DEFORMABLES_STANDARD_TESTS<VECTOR<float,3> >::DEFORMABLES_STANDARD_TESTS(DEFORMABLE_BODY_COLLECTION<VECTOR<float,3> >&);
template TRIANGULATED_SURFACE<double> const& DEFORMABLES_STANDARD_TESTS<VECTOR<double,3> >::Copy_And_Add_Structure<TRIANGULATED_SURFACE<double> const>(TRIANGULATED_SURFACE<double> const&,ARRAY<int,int>*,bool);
template TRIANGULATED_SURFACE<float> const& DEFORMABLES_STANDARD_TESTS<VECTOR<float,3> >::Copy_And_Add_Structure<TRIANGULATED_SURFACE<float> const>(TRIANGULATED_SURFACE<float> const&,ARRAY<int,int>*,bool);
