//#####################################################################
// Copyright 2004-2009, Ronald Fedkiw, Geoffrey Irving, Nipun Kwatra, Michael Lentine, Andrew Selle, Tamar Shinar, Eftychios Sifakis, Joseph Teran.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class DEFORMABLE_PARTICLES
//#####################################################################
#ifndef __DEFORMABLE_PARTICLES__
#define __DEFORMABLE_PARTICLES__

#include <Core/Arrays/ARRAY.h>
#include <Tools/Clone/CLONEABLE.h>
#include <Tools/Particles/PARTICLES.h>
#include <Geometry/Geometry_Particles/GEOMETRY_PARTICLES.h>

namespace PhysBAM{

template<class TV>
class DEFORMABLE_PARTICLES:public CLONEABLE<DEFORMABLE_PARTICLES<TV>,GEOMETRY_PARTICLES<TV> > // X, V
{
    typedef typename TV::SCALAR T;
    typedef CLONEABLE<DEFORMABLE_PARTICLES<TV>,GEOMETRY_PARTICLES<TV> > BASE;
public:
    using BASE::X;using BASE::V;using BASE::Remove_Array;using BASE::Get_Attribute_Index;using BASE::Remove_Array_Using_Index;

    ARRAY_VIEW<T> mass;
    ARRAY_VIEW<T> one_over_mass;
    ARRAY_VIEW<T> effective_mass;
    ARRAY_VIEW<T> one_over_effective_mass;
    bool store_mass;

    DEFORMABLE_PARTICLES();
    virtual ~DEFORMABLE_PARTICLES();

    void Store_Mass(bool store=true)
    {store_mass=store;if(store) this->Add_Array("mass",&mass);else Remove_Array("mass");}

    T Min_Mass() const 
    {return mass.Size()?mass.Min():FLT_MAX;}

    T Max_Mass() const 
    {return mass.Size()?mass.Max():0;}

    void Euler_Step_Position(const T dt)
    {X+=dt*V;}

    template<class T_INDICES>
    void Euler_Step_Position(const T_INDICES& indices,const T dt)
    {X.Subset(indices)+=dt*V.Subset(indices);}

    void Clone_Helper(const DEFORMABLE_PARTICLES& particles)
    {Store_Mass(particles.store_mass);BASE::Clone_Helper(particles);}

//#####################################################################
    TV Center_Of_Mass() const;
//#####################################################################
};
}
#endif
