//#####################################################################
// Copyright 2002-2006, Ronald Fedkiw, Geoffrey Irving, Avi Robinson-Mosher, Joseph Teran.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Geometry/Basic_Geometry/HEXAHEDRON.h>
using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class T> HEXAHEDRON<T>::    
HEXAHEDRON()
    :x0(-1,-1,-1),x1(-1,-1,1),x2(-1,1,-1),x3(-1,1,1),x4(1,-1,-1),x5(1,-1,1),x6(1,1,-1),x7(1,1,1)
{}
//#####################################################################
// Constructor
//#####################################################################
template<class T> HEXAHEDRON<T>::    
HEXAHEDRON(const TV& x1_input,const TV& x2_input,const TV& x3_input,const TV& x4_input,
    const TV& x5_input,const TV& x6_input,const TV& x7_input,const TV& x8_input)
    :x0(x1_input),x1(x2_input),x2(x3_input),x3(x4_input),x4(x5_input),x5(x6_input),x6(x7_input),x7(x8_input)
{
}
template<class T> T HEXAHEDRON<T>::
Volume(const TV& x0,const TV& x1,const TV& x2,const TV& x3,
    const TV& x4,const TV& x5,const TV& x6,const TV& x7)
{
    return abs(Signed_Volume(x0,x1,x2,x3,x4,x5,x6,x7));
}
//#####################################################################
// Function Signed_Volume
//#####################################################################
template<class T> T HEXAHEDRON<T>::
Signed_Volume(const TV& x0,const TV& x1,const TV& x2,const TV& x3,
    const TV& x4,const TV& x5,const TV& x6,const TV& x7)
{
    TV x17=x7-x1,x12=x2-x1,x13=x3-x1,x24=x4-x2,x47=x7-x2,x46=x6-x4,x14=x4-x1,x15=x5-x1,x04=x4-x0;
    return ((T)1/6)*(TV::Triple_Product(x17,x12,x13)-TV::Triple_Product(x24,x47,x46)+TV::Triple_Product(x14,x17,x15)+TV::Triple_Product(x12,x17,x14)-TV::Triple_Product(x14,x24,x04));
}
//#####################################################################
// Function Volume
//#####################################################################
template<class T> T HEXAHEDRON<T>::
Volume() const
{
    return Volume(x0,x1,x2,x3,x4,x5,x6,x7);
}
//#####################################################################
// Function Signed_Volume
//#####################################################################
template<class T> T HEXAHEDRON<T>::
Signed_Volume() const
{
    return Signed_Volume(x0,x1,x2,x3,x4,x5,x6,x7);
}

//#####################################################################
// Function Scaled
//#####################################################################
template<class T,class TV> HEXAHEDRON<T>
Scaled(const HEXAHEDRON<T>& b,const TV& center,T scale)
{
    return HEXAHEDRON<T>((b.x0-center)*scale+center,
        (b.x1-center)*scale+center,
        (b.x2-center)*scale+center,
        (b.x3-center)*scale+center,
        (b.x4-center)*scale+center,
        (b.x5-center)*scale+center,
        (b.x6-center)*scale+center,
        (b.x7-center)*scale+center);
}
//#####################################################################
// Function Transformed
//#####################################################################
template<class T,class TV> HEXAHEDRON<T>
Transformed(const HEXAHEDRON<T>& b,const FRAME<TV>& f)
{
    return HEXAHEDRON<T>(f*b.x0,f*b.x1,f*b.x2,f*b.x3,f*b.x4,f*b.x5,f*b.x6,f*b.x7);
}
//#####################################################################
namespace PhysBAM{
template class HEXAHEDRON<float>;
template class HEXAHEDRON<double>;
}
