//#####################################################################
// Copyright 2020, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Geometry/Basic_Geometry/TRANSFORMED_OBJECT.h>
#include <Geometry/Basic_Geometry/ORIENTED_BOX.h>
#include <Geometry/Basic_Geometry/SMOOTH_GEAR.h>
#include <iostream>
namespace PhysBAM{
//#####################################################################
// Function Bounding_Box
//#####################################################################
template<class TV,class OBJECT> RANGE<TV> TRANSFORMED_OBJECT<TV,OBJECT>::
Bounding_Box() const
{
    return ORIENTED_BOX<TV>(object.Bounding_Box(),frame).Bounding_Box();
}
//#####################################################################
// Function Signed_Distance
//#####################################################################
template<class TV,class OBJECT> auto TRANSFORMED_OBJECT<TV,OBJECT>::
Signed_Distance(const TV& X) const -> T
{
    return object.Signed_Distance(frame.Inverse_Times(X));
}
//#####################################################################
// Function Surface
//#####################################################################
template<class TV,class OBJECT> TV TRANSFORMED_OBJECT<TV,OBJECT>::
Surface(const TV& X) const
{
    return frame*object.Surface(frame.Inverse_Times(X));
}
//#####################################################################
// Function Normal
//#####################################################################
template<class TV,class OBJECT> TV TRANSFORMED_OBJECT<TV,OBJECT>::
Normal(const TV& X) const
{
    return frame.r*object.Normal(frame.Inverse_Times(X));
}
//#####################################################################
// Function Normal
//#####################################################################
template<class TV,class OBJECT> TV TRANSFORMED_OBJECT<TV,OBJECT>::
Normal(const TV& X,const int aggregate) const
{
    return Normal(X);
}
//#####################################################################
// Function Hessian
//#####################################################################
template<class TV,class OBJECT> auto TRANSFORMED_OBJECT<TV,OBJECT>::
Hessian(const TV& X) const -> SYMMETRIC_MATRIX<T,TV::m>
{
    auto H=object.Hessian(frame.Inverse_Times(X));
    auto R=frame.r.Rotation_Matrix();
    return SYMMETRIC_MATRIX<T,TV::m>::Conjugate(R,H);
}
//#####################################################################
// Function Lazy_Inside
//#####################################################################
template<class TV,class OBJECT> bool TRANSFORMED_OBJECT<TV,OBJECT>::
Lazy_Inside(const TV& X) const
{
    return object.Lazy_Inside(frame.Inverse_Times(X));
}
//#####################################################################
// Function Lazy_Outside
//#####################################################################
template<class TV,class OBJECT> bool TRANSFORMED_OBJECT<TV,OBJECT>::
Lazy_Outside(const TV& X) const
{
    return object.Lazy_Outside(frame.Inverse_Times(X));
}
//#####################################################################
// Function Inside
//#####################################################################
template<class TV,class OBJECT> bool TRANSFORMED_OBJECT<TV,OBJECT>::
Inside(const TV& X,const T thickness_over_two) const
{
    return object.Inside(frame.Inverse_Times(X),thickness_over_two);
}
//#####################################################################
// Function Outside
//#####################################################################
template<class TV,class OBJECT> bool TRANSFORMED_OBJECT<TV,OBJECT>::
Outside(const TV& X,const T thickness_over_two) const
{
    return object.Outside(frame.Inverse_Times(X),thickness_over_two);
}
//#####################################################################
// Function Boundary
//#####################################################################
template<class TV,class OBJECT> bool TRANSFORMED_OBJECT<TV,OBJECT>::
Boundary(const TV& X,const T thickness_over_two) const
{
    return object.Boundary(frame.Inverse_Times(X),thickness_over_two);
}
//#####################################################################
// Function Name
//#####################################################################
template<class TV,class OBJECT> std::string TRANSFORMED_OBJECT<TV,OBJECT>::
Name()
{
    return "TRANSFORMED_OBJECT<TV,"+OBJECT::Name()+">";
}
template class TRANSFORMED_OBJECT<VECTOR<double,2>,SMOOTH_GEAR<VECTOR<double,2> > >;
template class TRANSFORMED_OBJECT<VECTOR<double,3>,SMOOTH_GEAR<VECTOR<double,3> > >;
template class TRANSFORMED_OBJECT<VECTOR<float,2>,SMOOTH_GEAR<VECTOR<float,2> > >;
template class TRANSFORMED_OBJECT<VECTOR<float,3>,SMOOTH_GEAR<VECTOR<float,3> > >;
}
