//#####################################################################
// Copyright 2002-2008, Robert Bridson, Ronald Fedkiw, Geoffrey Irving, Andrew Selle, Joseph Teran.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class HEXAHEDRALIZED_VOLUME
//#####################################################################
#ifndef __HEXAHEDRALIZED_VOLUME__
#define __HEXAHEDRALIZED_VOLUME__

#include <Core/Arrays/ARRAYS_FORWARD.h>
#include <Geometry/Geometry_Particles/GEOMETRY_PARTICLES.h>
#include <Geometry/Meshes/HEXAHEDRON_MESH.h>
#include <Geometry/Meshes/MESH_OBJECT_BASE.h>
namespace PhysBAM{

template<class TV> class GRID;
template<class T> class HEXAHEDRON;

template<class T>
class HEXAHEDRALIZED_VOLUME:public MESH_OBJECT_BASE<VECTOR<T,3>,HEXAHEDRON_MESH>
{
    typedef VECTOR<T,3> TV;typedef VECTOR<int,3> IV;
public:
    typedef MESH_OBJECT_BASE<TV,HEXAHEDRON_MESH> BASE;
    using BASE::mesh;using BASE::particles;

    ARRAY<HEXAHEDRON<T> >* hexahedron_list;
    TETRAHEDRALIZED_VOLUME<T>* tetrahedralized_volume;
    TRIANGULATED_SURFACE<T>* triangulated_surface;

    HEXAHEDRALIZED_VOLUME();
    HEXAHEDRALIZED_VOLUME(HEXAHEDRON_MESH& mesh_input,GEOMETRY_PARTICLES<TV>& particles_input);
    virtual ~HEXAHEDRALIZED_VOLUME();

    virtual std::string Name() const override {return Static_Name();}
    static std::string Static_Name()
    {return "HEXAHEDRALIZED_VOLUME<T>";}

//#####################################################################
    void Clean_Memory() override;
    void Update_Hexahedron_List(); // updates the hexahedrons assuming the particle positions are already updated
    void Initialize_Tetrahedralized_Volume();
    void Initialize_Triangulated_Surface();
    void Initialize_Cube_Mesh_And_Particles(const GRID<TV>& grid);
    T Total_Volume() const;
private:
    void Refresh_Auxiliary_Structures_Helper() override {PHYSBAM_NOT_IMPLEMENTED();}
//#####################################################################
};
}
#endif
