//#####################################################################
// Copyright 2002-2009, Robert Bridson, Ronald Fedkiw, Jon Gretarsson, Geoffrey Irving, Sergey Koltakov, Nipun Kwatra, Frank Losasso, Neil Molino, Eftychios Sifakis, Jerry Talton.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class SEGMENT_MESH
//#####################################################################
#include <Geometry/Meshes/SEGMENT_MESH.h>
#include <Core/Data_Structures/UNION_FIND.h>
#include <Geometry/Meshes/POINT_SIMPLEX_MESH.h>
using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
SIMPLEX_MESH<1>::
SIMPLEX_MESH()
    :connected_segments(0),ordered_loop_nodes(0),boundary_mesh(0)
{}
//#####################################################################
// Constructor
//#####################################################################
SIMPLEX_MESH<1>::
SIMPLEX_MESH(const int number_nodes_input,const ARRAY<VECTOR<int,2> >& segment_list)
    :SIMPLEX_MESH_BASE<1>(number_nodes_input,segment_list),connected_segments(0),ordered_loop_nodes(0),boundary_mesh(0)
{}
//#####################################################################
// Constructor
//#####################################################################
SIMPLEX_MESH<1>::
SIMPLEX_MESH(const SIMPLEX_MESH& segment_mesh)
    :SIMPLEX_MESH_BASE<1>(segment_mesh),connected_segments(0),ordered_loop_nodes(0),boundary_mesh(0)
{}
//#####################################################################
// Destructor
//#####################################################################
SIMPLEX_MESH<1>::
~SIMPLEX_MESH()
{
    delete connected_segments;
    delete ordered_loop_nodes;
}
//#####################################################################
// Function Delete_Auxiliary_Structures
//#####################################################################
void SIMPLEX_MESH<1>::
Delete_Auxiliary_Structures()
{
    SIMPLEX_MESH_BASE<1>::Delete_Auxiliary_Structures();
    delete connected_segments;
    connected_segments=0;
    delete ordered_loop_nodes;
    ordered_loop_nodes=0;
}
//#####################################################################
// Function Refresh_Auxiliary_Structures
//#####################################################################
void SIMPLEX_MESH<1>::
Refresh_Auxiliary_Structures()
{
    SIMPLEX_MESH_BASE<1>::Refresh_Auxiliary_Structures();
    if(connected_segments) Initialize_Connected_Segments();
    if(ordered_loop_nodes) Initialize_Ordered_Loop_Nodes();
}
//#####################################################################
// Function Initialize_Connected_Segments
//#####################################################################
void SIMPLEX_MESH<1>::
Initialize_Connected_Segments()
{
    delete connected_segments;
    connected_segments=new ARRAY<ARRAY<VECTOR<int,2> > >;
    UNION_FIND<> union_find(number_nodes);
    Add_Connectivity(union_find);
    ARRAY<int> buckets(number_nodes);
    buckets.Fill(-1);
    for(int s=0;s<elements.m;s++){
        int bucket=union_find.Find(elements(s)(0));
        int& index=buckets(bucket);
        if(index<0) index=connected_segments->Append(ARRAY<VECTOR<int,2> >());
        (*connected_segments)(index).Append(elements(s));}
}
//#####################################################################
// Function Initialize_Ordered_Loop_Nodes
//#####################################################################
void SIMPLEX_MESH<1>::
Initialize_Ordered_Loop_Nodes()
{
    delete ordered_loop_nodes;
    ordered_loop_nodes=new ARRAY<ARRAY<int> >;

    bool created_neighbor_nodes=false,created_connected_segments=false;
    if(!neighbor_nodes){Initialize_Neighbor_Nodes();
        created_neighbor_nodes=true;} 
    if(!connected_segments){Initialize_Connected_Segments();
        created_connected_segments=true;}

    for(int i=0;i<connected_segments->m;i++){
        for(int j=0;j<(*connected_segments)(i).m;j++)for(int k=0;k<2;k++)if((*neighbor_nodes)((*connected_segments)(i)(j)(k)).m!=2) goto not_closed;
        ordered_loop_nodes->Append(ARRAY<int>());
        {int start_node=(*connected_segments)(i)(0)(0),curr_node=(*connected_segments)(i)(0)(1);
        ordered_loop_nodes->Last().Append(start_node);
        do{
            int previous_node=ordered_loop_nodes->Last().Last(),neighbor1=(*neighbor_nodes)(curr_node)(0),neighbor2=(*neighbor_nodes)(curr_node)(1);
            ordered_loop_nodes->Last().Append(curr_node);
            curr_node=previous_node==neighbor1?neighbor2:neighbor1;}
        while(curr_node!=start_node);}
      not_closed: continue;}

    if(created_neighbor_nodes){delete neighbor_nodes;neighbor_nodes=0;}
    if(created_connected_segments){delete connected_segments;connected_segments=0;}
}
//#####################################################################
// Function Initialize_Straight_Mesh
//#####################################################################
void SIMPLEX_MESH<1>::
Initialize_Straight_Mesh(const int number_of_points,bool loop,int start_offset)
{
    Clean_Memory();
    number_nodes=number_of_points+start_offset;
    elements.Exact_Resize(number_of_points-!loop);
    for(int i=0;i<number_of_points-1;i++)elements(i).Set(start_offset+i,start_offset+i+1);
    if(loop) elements(number_of_points-1).Set(number_nodes-1,start_offset);
}
//#####################################################################
// Function Initialize_Boundary_Mesh
//#####################################################################
void SIMPLEX_MESH<1>::
Initialize_Boundary_Mesh()
{
    delete boundary_mesh;
    boundary_mesh=new POINT_SIMPLEX_MESH();

    ARRAY<VECTOR<int,dimension> >& boundary_elements=boundary_mesh->elements;
    ARRAY<bool>& boundary_directions=boundary_mesh->directions;
    bool incident_elements_defined=incident_elements!=0;

    if(!incident_elements_defined) Initialize_Incident_Elements();
    for(int i=0;i<incident_elements->m;i++){
        ARRAY<int>& current_incident_elements=(*incident_elements)(i);
        if(current_incident_elements.m==1){
            boundary_elements.Append(VECTOR<int,dimension>(i));
            boundary_directions.Append(elements(current_incident_elements(0)).x!=i);}}
    boundary_elements.Compact();boundary_directions.Compact();
    boundary_mesh->number_nodes=number_nodes;

    if(incident_elements_defined){delete incident_elements;incident_elements=0;}
}
//#####################################################################
// Function Assert_Consistent
//#####################################################################
bool SIMPLEX_MESH<1>::
Assert_Consistent() const
{
    if(connected_segments && ordered_loop_nodes) assert(connected_segments->m==ordered_loop_nodes->m);
    return SIMPLEX_MESH_BASE<1>::Assert_Consistent();
}
//#####################################################################
