//#####################################################################
// Copyright 2002-2005, Ronald Fedkiw, Geoffrey Irving.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Geometry/Spatial_Acceleration/SPHERE_PARTITION.h>
using namespace PhysBAM;
//#####################################################################
// Function Normal
//#####################################################################
template<class T> VECTOR<T,3> SPHERE_PARTITION<T>::
Normal(const VECTOR<T,3>& location,const int aggregate) const
{
    return spheres(aggregate).Normal(location); 
}
//#####################################################################
// Function Inside
//#####################################################################
template<class T> bool SPHERE_PARTITION<T>::
Inside(const VECTOR<T,3>& location,const T thickness_over_two) const 
{
    if(box.Outside(location,thickness_over_two)) return false; 
    
    // find the voxel that contains the location - left borders
    int i,j,ij;Find_Voxel(location,i,j,ij);
    // check all the spheres in that voxel
    if(voxel_sphere_list(i,j,ij)) for(int k=0;k<voxel_sphere_list(i,j,ij)->m;k++) if(spheres((*voxel_sphere_list(i,j,ij))(k)).Inside(location,thickness_over_two)) return true;

    return false; // otherwise
}
//#####################################################################
// Function Outside
//#####################################################################
template<class T> bool SPHERE_PARTITION<T>::
Outside(const VECTOR<T,3>& location,const T thickness_over_two) const  
{
    if(box.Outside(location,thickness_over_two)) return true; 
    
    // find the voxel that contains the location - left borders
    int i,j,ij;Find_Voxel(location,i,j,ij);
    // check all the spheres in that voxel
    if(voxel_sphere_list(i,j,ij)) for(int k=0;k<voxel_sphere_list(i,j,ij)->m;k++) if(!spheres((*voxel_sphere_list(i,j,ij))(k)).Outside(location,thickness_over_two)) return false;

    return true; // otherwise
}
//#####################################################################
// Function Boundary
//#####################################################################
template<class T> bool SPHERE_PARTITION<T>::
Boundary(const VECTOR<T,3>& location,const T thickness_over_two) const
{
    return !Inside(location,thickness_over_two) && !Outside(location,thickness_over_two);
}
//#####################################################################
// Function Set_Up_Grid
//#####################################################################
template<class T> void SPHERE_PARTITION<T>::
Set_Up_Grid(const IV& size)
{
    box=RANGE<TV>::Empty_Box();
    for(int k=0;k<spheres.m;k++)box.Enlarge_To_Include_Box(spheres(k).Bounding_Box());
    grid.Initialize(size,box);
    
    for(int i=0;i<voxel_sphere_list.array.m;i++){
        delete voxel_sphere_list.array(i);
        voxel_sphere_list.array(i)=0;}
    voxel_sphere_list.Resize(grid.numbers_of_cells);
    for(int k=0;k<spheres.m;k++){
        // find the grid cell that contains the sphere - left borders
        int i_left,j_bottom,ij_front;Find_Voxel(spheres(k).center,i_left,j_bottom,ij_front);
        // find the starting and stopping values for the sphere extent
        T radius=spheres(k).radius;
        int i_range=(int)(radius/grid.dX.x+1),j_range=(int)(radius/grid.dX.y+1),ij_range=(int)(radius/grid.dX.z+1);
        int i_start=max(1,i_left-i_range),i_end=min(grid.numbers_of_cells.x,i_left+i_range),j_start=max(1,j_bottom-j_range),
             j_end=min(grid.numbers_of_cells.y,j_bottom+j_range),ij_start=max(1,ij_front-ij_range),ij_end=min(grid.numbers_of_cells.z,ij_front+ij_range);
        // map out the cells that get the sphere
        for(int i=i_start;i<i_end;i++) for(int j=j_start;j<j_end;j++) for(int ij=ij_start;ij<ij_end;ij++){
            if(!voxel_sphere_list(i,j,ij)) voxel_sphere_list(i,j,ij)=new ARRAY<int>();
            voxel_sphere_list(i,j,ij)->Append(k);}}
}
//#####################################################################
// Function Find_Voxel
//#####################################################################
template<class T> void SPHERE_PARTITION<T>::
Find_Voxel(const VECTOR<T,3>& location,int& i_left,int& j_bottom,int& ij_front) const   
{
    typedef VECTOR<int,3> IV;
    IV index=clamp(IV((location-grid.Node_X(IV::All_Ones_Vector()))/grid.dX)+1,IV::All_Ones_Vector(),grid.numbers_of_cells);
    i_left=index.x;j_bottom=index.y;ij_front=index.z;
}
//#####################################################################
namespace PhysBAM{
template class SPHERE_PARTITION<float>;
template class SPHERE_PARTITION<double>;
}
