//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CUBIC_SPLINE_INTERPOLATION__
#define __CUBIC_SPLINE_INTERPOLATION__

#include <Grid_PDE/Interpolation/GENERAL_INTERPOLATION_CELL.h>
#include <Grid_PDE/Interpolation/GENERAL_INTERPOLATION_FACE.h>
#include <Grid_PDE/Interpolation/GENERAL_INTERPOLATION_NODE.h>
#include <Grid_PDE/Interpolation/INTERPOLATION_UNIFORM_FORWARD.h>
namespace PhysBAM{

// C^1 continous
struct CUBIC_SPLINE_INTERPOLATION_STENCIL
{
    constexpr static int stencil_size=4;
    template<int a,int d,class T> static VECTOR<VECTOR<T,4>,1> Value(T t)
    {
        T s=1-t,u=s*t;
        VECTOR<T,4> r={u*s,-2*s+u*(3*t-2),-2*t+u*(1-3*t),u*t};
        return {-(T).5*r};
    }
    template<int a,int d,class T> static VECTOR<VECTOR<T,4>,1> Diff(T t)
    {
        T s=1-t,u=3*s*t;
        VECTOR<T,4> r={u-s,-3*u-t,3*u+s,-u+t};
        return {(T).5*r};
    }
    template<int a,int d,class T> static VECTOR<VECTOR<T,4>,1> Hess(T t)
    {
        return {{2-3*t,9*t-5,4-9*t,3*t-1}};
    }
};

template<class TV,class T2>
using CUBIC_SPLINE_INTERPOLATION_NODE=GENERAL_INTERPOLATION_NODE<TV,T2,CUBIC_SPLINE_INTERPOLATION_STENCIL>;

template<class TV,class T2>
using CUBIC_SPLINE_INTERPOLATION_CELL=GENERAL_INTERPOLATION_CELL<TV,T2,CUBIC_SPLINE_INTERPOLATION_STENCIL>;

template<class TV,class T2,class T_FACE_ARRAY=ARRAY<T2,FACE_INDEX<TV::m> > >
using CUBIC_SPLINE_INTERPOLATION_FACE=GENERAL_INTERPOLATION_FACE<TV,T2,CUBIC_SPLINE_INTERPOLATION_STENCIL,false,T_FACE_ARRAY>;

template<class TV,class T2,class T_FACE_ARRAY>
struct CUBIC_SPLINE_INTERPOLATION_ALL:CUBIC_SPLINE_INTERPOLATION_NODE<TV,T2>,CUBIC_SPLINE_INTERPOLATION_CELL<TV,T2>,CUBIC_SPLINE_INTERPOLATION_FACE<TV,T2,T_FACE_ARRAY>
{
};

}
#endif
