//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __GENERAL_INTERPOLATION_NODE__
#define __GENERAL_INTERPOLATION_NODE__

#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays_Nd/ARRAYS_ND.h>
#include <Core/Log/DEBUG_UTILITIES.h>
#include <Tools/Interpolation/INTERPOLATION_FORWARD.h>
#include <Grid_Tools/Arrays/FACE_ARRAYS.h>
#include <Grid_PDE/Interpolation/GENERAL_INTERPOLATION_HELPERS.h>
#include <Grid_PDE/Interpolation/INTERPOLATION_HELPERS.h>
#include <Grid_PDE/Interpolation/INTERPOLATION_NODE.h>
namespace PhysBAM{

template<class TV,class T2,class STENCIL,bool const_extrap=false>
class GENERAL_INTERPOLATION_NODE:public INTERPOLATION_NODE<TV,T2>
{
    static_assert(TV::m>=1 && TV::m<=3);
    typedef typename TV::SCALAR T;typedef VECTOR<int,TV::m> IV;
    typedef VECTOR<T2,TV::m> TV2;typedef SYMMETRIC_MATRIX<T2,TV::m> SM2;
    typedef INTERPOLATION_NODE<TV,T2> BASE;
    constexpr static int ss=STENCIL::stencil_size;
public:
    STENCIL stencil;
    
    GENERAL_INTERPOLATION_NODE() = default;
    GENERAL_INTERPOLATION_NODE(const GENERAL_INTERPOLATION_NODE&) = delete;
    void operator=(const GENERAL_INTERPOLATION_NODE&) = delete;
    virtual ~GENERAL_INTERPOLATION_NODE() = default;

    T2 Interpolate_Node(const GRID<TV>& grid,const ARRAYS_ND_BASE<T2,IV>& u,const TV& X) const
    {
        auto p=INTERPOLATION_HELPERS::Clamped_Index<ss,const_extrap>(grid,at_nodes,u.domain,X);
        return Interpolate<-1>(stencil,grid,u,p);
    }
    VECTOR<T2,2> Extrema_Node(const GRID<TV>& grid,const ARRAYS_ND_BASE<T2,IV>& u0,const ARRAYS_ND_BASE<T2,IV>& u1,const TV& X) const
    {
        auto p=INTERPOLATION_HELPERS::Clamped_Index<ss,const_extrap>(grid,at_nodes,u0.domain,X);
        return Extrema<ss,-1>(grid,u0,u1,p);
    }
    TV2 Gradient_Node(const GRID<TV>& grid,const ARRAYS_ND_BASE<T2,IV>& u,const TV& X) const
    {
        auto p=INTERPOLATION_HELPERS::Clamped_Index<ss,const_extrap>(grid,at_nodes,u.domain,X);
        return Gradient<-1>(stencil,grid,u,p);
    }
    SM2 Hessian_Node(const GRID<TV>& grid,const ARRAYS_ND_BASE<T2,IV>& u,const TV& X) const
    {
        auto p=INTERPOLATION_HELPERS::Clamped_Index<ss,const_extrap>(grid,at_nodes,u.domain,X);
        return Hessian<-1>(stencil,grid,u,p);
    }

protected:
    virtual T2 Interpolate_Node_V(const GRID<TV>& grid,const ARRAYS_ND_BASE<T2,IV>& u,const TV& X) const
    {return Interpolate_Node(grid,u,X);}
    virtual VECTOR<T2,2> Extrema_Node_V(const GRID<TV>& grid,const ARRAYS_ND_BASE<T2,IV>& u0,const ARRAYS_ND_BASE<T2,IV>& u1,const TV& X) const
    {return Extrema_Node(grid,u0,u1,X);}
    virtual TV2 Gradient_Node_V(const GRID<TV>& grid,const ARRAYS_ND_BASE<T2,IV>& u,const TV& X) const
    {return Gradient_Node(grid,u,X);}
    virtual SM2 Hessian_Node_V(const GRID<TV>& grid,const ARRAYS_ND_BASE<T2,IV>& u,const TV& X) const
    {return Hessian_Node(grid,u,X);}
//#####################################################################
};
}
#endif
