//##################################################################### 
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __INTERPOLATION_SPLINES__
#define __INTERPOLATION_SPLINES__

namespace PhysBAM {
template<class T,int r,int d,int k>
VECTOR<T,r> Evaluate_Spline(const double (&A)[r][d],const T (&u)[k])
{
    static_assert(d<=k,"not enough powers available to evaluate spline.");
    VECTOR<T,r> v;
    for(int j=0;j<d;j++)
    {
        T x=u[j];
        for(int i=0;i<r;i++)
            v(i)+=A[i][j]*x;
    }
    return v;
}

template<class T,int r,int d,int k>
VECTOR<T,r> Evaluate_Spline_Diff(const double (&A)[r][d],const T (&u)[k])
{
    static_assert(d-1<=k,"not enough powers available to evaluate spline.");
    VECTOR<T,r> v;
    for(int j=1;j<d;j++)
    {
        T x=j*u[j-1];
        for(int i=0;i<r;i++)
            v(i)+=A[i][j]*x;
    }
    return v;
}

template<class T,int r,int d,int k>
VECTOR<T,r> Evaluate_Spline_Hess(const double (&A)[r][d],const T (&u)[k])
{
    static_assert(d-2<=k,"not enough powers available to evaluate spline.");
    VECTOR<T,r> v;
    for(int j=2;j<d;j++)
    {
        T x=j*(j-1)*u[j-2];
        for(int i=0;i<r;i++)
            v(i)+=A[i][j]*x;
    }
    return v;
}

namespace IS {
static constexpr double B0[1][1]=
{{1}};
static constexpr double B1[2][2]=
{{1,-1},{0,1}};
static constexpr double B2[3][3]=
{{.5,-1,.5},{.5,1,-1},{0,0,.5}};
static constexpr double B3[4][4]=
{{1./6,-.5,.5,-1./6},{2./3,0,-1,.5},{1./6,.5,.5,-.5},{0,0,0,1./6}};

static constexpr double C2[2][3]=
{{1,-4,3},{0,-2,3}};
static constexpr double C3[3][4]=
{{0,-1,2,-1},{0,1,-1},{0,0,-1,1}};
static constexpr double C4[4][5]=
{{-1./12,0,.5,-2./3,.25},{1./6,0,-1,1,-.25},{-1./12,0,.5,0,-.25},{0,0,0,-1./3,.25}};

static constexpr double D2[1][3]=
{{1,-6,6}};
static constexpr double D3[2][4]=
{{0,-1,3,-2},{0,1,-3,2}};
static constexpr double D4[3][5]=
{{0,0,.5,-1,.5},{0,0,-1,2,-1},{0,0,.5,-1,.5}};
static constexpr double D5[4][6]=
{{1./60,0,0,-1./6,.25,-.1},{-1./30,0,0,.5,-.75,.3},{1./60,0,0,-.5,.75,-.3},{0,0,0,1./6,-.25,.1}};

static constexpr double F4[3][5]=
{{-3,2,30,-54,25},{-3,-2,132,-260,130},{0,0,18,-46,25}};
static constexpr double F5[4][6]=
{{-.5,3,-1,-10,13.5,-5},{1,0,2,-34,51.5,-21},{-.5,-3,-1,38,-53.5,21},{0,0,0,6,-11.5,5}};

static constexpr double G4[3][5]=
{{27.5,-55,-172.5,400,-200},{27.5,55,-1095,2080,-1040},{0,0,-172.5,400,-200}};
static constexpr double G5[4][6]=
{{2.5,-27.5,27.5,57.5,-100,40},{2,0,-55,307.5,-420,168},{2.5,27.5,27.5,-307.5,420,-168},{0,0,0,-57.5,100,-40}};

static constexpr double H3[3][4]=
{{18,-50,46,-14},{18,50,-50},{0,0,4,14}};
static constexpr double H4[4][5]=
{{29./6,-18,25,-46./3,3.5},{79./3,0,-50,32,-3.5},{29./6,18,25,-18,-3.5},{0,0,0,4./3,3.5}};
}
}
#endif
