//#####################################################################
// Copyright 2005-2017, Eran Guendelman, Avi Robinson-Mosher, Craig Schroeder, Andrew Selle, Jonathan Su.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class CELL_ITERATOR
//#####################################################################
#ifndef __CELL_ITERATOR__
#define __CELL_ITERATOR__

#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Grid_Tools/Grids/FACE_INDEX.h>
#include <Grid_Tools/Grids/GRID.h>
namespace PhysBAM{

template<class TV>
class CELL_ITERATOR:public RANGE_ITERATOR<TV::m>
{
    typedef VECTOR<int,TV::m> IV;
public:
    typedef GRID_REGION T_REGION;
    typedef TV VECTOR_T;
    using RANGE_ITERATOR<TV::m>::index;

    GRID<TV> grid; // TODO GRID was const&

    CELL_ITERATOR(const GRID<TV>& grid_input,const int number_of_ghost_cells=0,
        const T_REGION& region_type=WHOLE_REGION,const int side=-1);
    CELL_ITERATOR(const GRID<TV>& grid_input,const RANGE<IV>& region_input);

    const IV& Cell_Index() const
    {return index;}

    TV Location() const
    {return grid.Cell_X(index);}

    RANGE<TV> Bounding_Box() const
    {return grid.Cell_Domain(index);}

    IV Cell_Node_Index(const int node) const
    {return index+GRID<TV>::Binary_Counts(index)(node);}

    IV First_Face_Index(const int axis) const
    {return grid.First_Face_Index_In_Cell(axis,index);}

    IV Second_Face_Index(const int axis) const
    {return grid.Second_Face_Index_In_Cell(axis,index);}

    FACE_INDEX<TV::m> Full_First_Face_Index(const int axis) const
    {return FACE_INDEX<TV::m>(axis,grid.First_Face_Index_In_Cell(axis,index));}

    FACE_INDEX<TV::m> Full_Second_Face_Index(const int axis) const
    {return FACE_INDEX<TV::m>(axis,grid.Second_Face_Index_In_Cell(axis,index));}

    IV Cell_Neighbor(const int i) const // 1 to 2*dimension
    {return GRID<TV>::Node_Neighbor(index,i);}
};
//#####################################################################
}
#endif
