//#####################################################################
// Copyright 2015-2020, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __MPM_DRIVER__
#define __MPM_DRIVER__
#include <Core/Vectors/VECTOR.h>
#include <Grid_Tools/Grids/FACE_INDEX.h>
#include <Hybrid_Methods/Collisions/MPM_COLLISION_OBJECT.h>
#include <Hybrid_Methods/Regions/REGION_ARRAY.h>
namespace PhysBAM{

template<class TV> class MPM_EXAMPLE;
template<class TV> class MPM_OBJECTIVE;
template<class TV> class PARTICLE_GRID_WEIGHTS;
template<class TV> class DIAGONALIZED_ISOTROPIC_STRESS_DERIVATIVE;
enum class RI;

template<class TV>
class MPM_DRIVER
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
public:

    int current_frame=0;

    MPM_EXAMPLE<TV>& example;
    
    MPM_DRIVER(MPM_EXAMPLE<TV>& example);
    virtual ~MPM_DRIVER();
    
    void Execute_Main_Program();
    void Initialize();

    void Advance_One_Time_Step();
    void Simulate_To_Frame(const int frame_input);
    void Write_Output_Files();
    void Write_Substep(const std::string& title);
    void Update_Particle_Weights();
    void Particle_To_Grid();
    void Grid_To_Particle();
    void Update_Plasticity_And_Hardening();
    void Apply_Forces();
    void Apply_Forces_Explicit();
    void Apply_Particle_Forces(REGION_ARRAY<TV,IV>& F);
    void Apply_Grid_Forces(REGION_ARRAY<TV,IV>& F);
    void Apply_Body_Collisions();
    void Apply_Particle_Collisions(int p);
    TV Examine_Collisions(TV X,TV V,T t);
    T Compute_Dt() const;
    T Max_Particle_Speed() const;
    T Grid_V_Upper_Bound() const;
    T Compute_Max_Sound_Speed() const;
    T Compute_Max_Sound_Speed_Async() const;
    T Max_Dt_Single_Particle_Pressure() const;
    T Max_Dt_Single_Particle() const;
    void Update_Simulated_Particles();
    void Print_Grid_Stats(const char* str,T dt,const REGION_ARRAY<TV,IV>& u,const REGION_ARRAY<TV,IV>* u0);
    void Print_Particle_Stats(const char* str,T dt);
    void Print_Energy_Stats(const char* str,const REGION_ARRAY<TV,IV>& u);
    T Grid_To_Particle_Limit_Dt();
    T Limit_Dt_Sound_Speed();
    void Reduce_Dt();
    void Reflect_Boundary_Mass_Momentum();
    void Reflect_Boundary_Friction(const REGION_ARRAY<T,IV>& mass,REGION_ARRAY<TV,IV>& u) const;
    void Reflect_Boundary_Velocity(REGION_ARRAY<TV,IV>& u);
    void Reflect_Boundary_Force(REGION_ARRAY<TV,IV>& force);
    void Step(std::function<void()> func,const char* name,bool dump_substep=true,bool do_step=true);
    void Compute_Regions();
    void Collide_Regions();
    template<class T2> void Fix_Periodic(ARRAY_VIEW<T2,IV> u) const;
    template<class T2,class OP> void Fix_Periodic_Accum(ARRAY_VIEW<T2,IV> u,const OP& op) const;
//#####################################################################
};
}
#endif
