//#####################################################################
// Copyright 2004-2009, Ronald Fedkiw, Geoffrey Irving, Nipun Kwatra, Michael Lentine, Andrew Selle, Tamar Shinar, Eftychios Sifakis, Joseph Teran.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class MPM_PARTICLES
//#####################################################################
#ifndef __MPM_PARTICLES__
#define __MPM_PARTICLES__

#include <Core/Arrays/ARRAY.h>
#include <Core/Matrices/MATRIX.h>
#include <Tools/Particles/PARTICLES.h>
#include <Deformables/Deformable_Objects/DEFORMABLE_PARTICLES.h>
#include <Hybrid_Methods/Examples_And_Drivers/REGION_ID.h>

namespace PhysBAM{

template<class TV>
class MPM_PARTICLES:public CLONEABLE<MPM_PARTICLES<TV>,DEFORMABLE_PARTICLES<TV> >
{
    typedef typename TV::SCALAR T;
    typedef CLONEABLE<MPM_PARTICLES<TV>,DEFORMABLE_PARTICLES<TV> > BASE;
public:
    using BASE::Add_Array;using BASE::Remove_Array;

    bool store_Fp=false,store_C=false,store_S=false;
    bool store_lame=false,store_lame0=false;
    bool store_plastic_def=false,store_region=false;
    ARRAY_VIEW<T> volume;
    ARRAY_VIEW<MATRIX<T,TV::m> > F,Fp,C;
    ARRAY_VIEW<SYMMETRIC_MATRIX<T,TV::m> > S;
    ARRAY_VIEW<bool> valid;
    ARRAY_VIEW<T> mu,lambda,mu0,lambda0;
    ARRAY_VIEW<int> region;
    ARRAY_VIEW<VECTOR<T,3> > color;

    MPM_PARTICLES();
    virtual ~MPM_PARTICLES();
    void Store_Fp(bool store);
    void Store_C(bool store);
    void Store_S(bool store);
    void Store_Lame(bool store);
    void Store_Lame0(bool store);
    void Store_Region(bool store);
    REGION_ID Region(int p) const {return REGION_ID(region(p));}

//#####################################################################
};
}
#endif
