//#####################################################################
// Copyright 2015, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class MPM_TRANSFERS
//#####################################################################
#ifndef __MPM_TRANSFERS__
#define __MPM_TRANSFERS__

#include <Core/Matrices/MATRIX.h>
#include <Core/Matrices/SYMMETRIC_MATRIX.h>
#include <Grid_Tools/Grids/GRID.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_TRANSFER_OPTIONS.h>
#include <Hybrid_Methods/Iterators/PARTICLE_GRID_ITERATOR.h>
namespace PhysBAM{
template<class T> std::pair<bool,bool>
Transfer_P2G(const MPM_TRANSFER_OPTIONS<T>& opt){return {};}

template<class TV,class T> void
Transfer_P2G(const GRID<TV>& grid,const MPM_TRANSFER_OPTIONS<T>& opt,
    const PARTICLE_GRID_ITERATOR<TV>& it_,
    ARRAYS_ND_BASE<T,VECTOR<int,TV::m> >& m,
    ARRAYS_ND_BASE<TV,VECTOR<int,TV::m> >& m_u,
    T mp,const TV& Vp,const TV& Xp,const MATRIX<T,TV::m>* Cp);

template<class T> std::pair<bool,bool>
Transfer_G2P(const MPM_TRANSFER_OPTIONS<T>& opt){return {!opt.use_cpic,opt.use_cpic};}

template<class TV,class T> void
Transfer_G2P(const GRID<TV>& grid,const MPM_TRANSFER_OPTIONS<T>& opt,
    const PARTICLE_GRID_ITERATOR<TV>& it_,
    const ARRAYS_ND_BASE<TV,VECTOR<int,TV::m> >& Vi_n,
    const ARRAYS_ND_BASE<TV,VECTOR<int,TV::m> >& Vi_u,
    const ARRAYS_ND_BASE<TV,VECTOR<int,TV::m> >& Vi_x,
    T Dp_inv,const TV& Xp_n,const TV& Vp_n,T dt,
    TV& Xp,TV& Vp,MATRIX<T,TV::m>* Cp,MATRIX<T,TV::m>& Fp,
    SYMMETRIC_MATRIX<T,TV::m>* Sp,MATRIX<T,TV::m>* dt_grad_u);

template<class TV,class T> MATRIX<T,TV::m>
Gather(const MPM_TRANSFER_OPTIONS<T>& opt,
    const PARTICLE_GRID_ITERATOR<TV>& it_,
    const ARRAYS_ND_BASE<TV,VECTOR<int,TV::m> >& u);

template<class TV,class T> void
Scatter(const MPM_TRANSFER_OPTIONS<T>& opt,
    const PARTICLE_GRID_ITERATOR<TV>& it_,
    ARRAYS_ND_BASE<TV,VECTOR<int,TV::m> >& f,
    const MATRIX<T,TV::m>& K);

// dF/dx(i)*du(i)
template<class T,int d> MATRIX<T,d>
Diff_Times_G2P_F(const MPM_TRANSFER_OPTIONS<T>& opt,
    const MATRIX<T,d>& Fp,const MATRIX<T,d>& dt_grad_u,
    const MATRIX<T,d>& grad_du);

// f(i)+=dM:dF/dx(i)
template<class T,int d> MATRIX<T,d>
Times_Diff_G2P_F(const MPM_TRANSFER_OPTIONS<T>& opt,
    const MATRIX<T,d>& Fp,
    const MATRIX<T,d>& dM,const MATRIX<T,d>& dt_grad_u);

template<class T,int d> MATRIX<T,d>
Times_Hess_Times_G2P_F(const MPM_TRANSFER_OPTIONS<T>& opt,
    const MATRIX<T,d>& Fp,
    const MATRIX<T,d>& dM,const MATRIX<T,d>& dt_grad_u,
    const MATRIX<T,d>& grad_du);

// dS/dx(i)*du(i)
template<class T,int d> SYMMETRIC_MATRIX<T,d>
Diff_Times_G2P_S(const MPM_TRANSFER_OPTIONS<T>& opt,
    const SYMMETRIC_MATRIX<T,d>& Sp,const MATRIX<T,d>& dt_grad_u,const MATRIX<T,d>& grad_du,T dt);

// f(i)+=dM:dS/dx(i)
template<class T,int d> MATRIX<T,d>
Times_Diff_G2P_S(const MPM_TRANSFER_OPTIONS<T>& opt,
    const SYMMETRIC_MATRIX<T,d>& Sp,
    const SYMMETRIC_MATRIX<T,d>& dM,const MATRIX<T,d>& dt_grad_u,T dt);

template<class T,int d> MATRIX<T,d>
Times_Hess_Times_G2P_S(const MPM_TRANSFER_OPTIONS<T>& opt,
    const SYMMETRIC_MATRIX<T,d>& Sp,
    const SYMMETRIC_MATRIX<T,d>& dM,const MATRIX<T,d>& dt_grad_u,
    const MATRIX<T,d>& grad_du,T dt);

template<class TV,class T> T
Transfer_G2P_Limit(const GRID<TV>& grid,const MPM_TRANSFER_OPTIONS<T>& opt,
    const PARTICLE_GRID_ITERATOR<TV>& it_,
    const ARRAYS_ND_BASE<TV,VECTOR<int,TV::m> >& Vi_n_,
    const ARRAYS_ND_BASE<TV,VECTOR<int,TV::m> >& Vi_x_,T dt,T s,T cfl,T cfl_F);
}
#endif
