//#####################################################################
// Copyright 2020, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class REGION_ARRAY
//#####################################################################
#ifndef __REGION_ARRAY__
#define __REGION_ARRAY__

#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays_Nd/ARRAYS_ND_VIEW.h>
#include <Hybrid_Methods/Examples_And_Drivers/REGION_ID.h>
namespace PhysBAM{

template<int d>
struct REGION_INDEX
{
    REGION_ID r;
    VECTOR<int,d> i;
};

template<class T,class ID> class REGION_ARRAY;

template<class T,int d>
class REGION_ARRAY<T,VECTOR<int,d> >
{
public:
    typedef VECTOR<int,d> IV;
    ARRAY<ARRAY_VIEW<T,IV>,REGION_ID> u;
    ARRAY<T,REGION_FLAT_ID> raw;

    T& operator()(const REGION_INDEX<d>& ri) {return u(ri.r)(ri.i);}
    const T& operator()(const REGION_INDEX<d>& ri) const {return u(ri.r)(ri.i);}

    T& operator()(REGION_FLAT_ID i) {return raw(i);}
    const T& operator()(REGION_FLAT_ID i) const {return raw(i);}

    REGION_FLAT_ID Flat_Index(const REGION_INDEX<d>& ri) const
    {return REGION_FLAT_ID(&(*this)(ri)-raw.Get_Array_Pointer());}

    REGION_FLAT_ID Flat_Size() const
    {return raw.m;}

    void Resize(const ARRAY<RANGE<IV>,REGION_ID>& region_ranges)
    {
        u.Resize(region_ranges.m);
        REGION_FLAT_ID space_needed(0);
        for(auto& r:region_ranges) space_needed+=r.Size();
        raw.Resize(space_needed,no_init);
        REGION_FLAT_ID next_offset(0);
        for(REGION_ID r(0);r<u.m;r++)
        {
            u(r).Set(&raw(next_offset),region_ranges(r));
            next_offset+=region_ranges(r).Size();
        }
    }

    template<class U>
    void Resize(const REGION_ARRAY<U,IV>& a)
    {
        raw.Resize(a.raw.m,no_init);
        u.Resize(a.u.m);
        REGION_FLAT_ID next_offset(0);
        for(REGION_ID r(0);r<u.m;r++)
        {
            u(r).Set(&raw(next_offset),a.u(r).domain);
            next_offset+=a.u(r).domain.Size();
        }
    }

    void Exchange(REGION_ARRAY& a)
    {
        u.Exchange(a.u);
        raw.Exchange(a.raw);
    }

    bool Valid_Index(const REGION_INDEX<d>& ri) const
    {
        return u.Valid_Index(ri.r) && u(ri.r).Valid_Index(ri.i);
    }
    
//#####################################################################
};
}
#endif
