//#####################################################################
// Copyright 2004-2009, Eran Guendelman, Geoffrey Irving, Andrew Selle.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <OpenGL/OpenGL/OPENGL_GRID_BASED_VECTOR_FIELD_3D.h>
#include <Core/Arrays/ARRAY.h>
#include <Core/Log/LOG.h>
#include <Core/Math_Tools/pow.h>
#include <Core/Math_Tools/RANGE.h>
#include <OpenGL/OpenGL/OPENGL_GRID_3D.h>
#include <OpenGL/OpenGL/OPENGL_UNIFORM_SLICE.h>
using namespace PhysBAM;
//#####################################################################
// OPENGL_GRID_BASED_VECTOR_FIELD_3D
//#####################################################################
template<class T> OPENGL_GRID_BASED_VECTOR_FIELD_3D<T>::
OPENGL_GRID_BASED_VECTOR_FIELD_3D(const GRID<TV> &grid, ARRAY<VECTOR<T,3> ,VECTOR<int,3> > &V)
    :OPENGL_VECTOR_FIELD_3D<T>(*new ARRAY<VECTOR<T,3> >,*new ARRAY<VECTOR<T,3> >,OPENGL_COLOR::Gray(.8f),.25f,true,false,false),grid(grid),V(V)
{
    max_vectors_3d=100000;
}
//#####################################################################
// ~OPENGL_GRID_BASED_VECTOR_FIELD_3D
//#####################################################################
template<class T> OPENGL_GRID_BASED_VECTOR_FIELD_3D<T>::
~OPENGL_GRID_BASED_VECTOR_FIELD_3D()
{
    delete &vector_field;
    delete &vector_locations;
}
//#####################################################################
// Bounding_Box
//#####################################################################
template<class T> RANGE<VECTOR<T,3> > OPENGL_GRID_BASED_VECTOR_FIELD_3D<T>::
Bounding_Box() const
{
    return World_Space_Box(grid.domain);
}
//#####################################################################
// Update
//#####################################################################
template<class T> void OPENGL_GRID_BASED_VECTOR_FIELD_3D<T>::
Update()
{
    vector_field.Remove_All();
    vector_locations.Remove_All();

    if(V.domain.Empty_Half_Open()) return;

    OPENGL_UNIFORM_SLICE<T>* slice=(OPENGL_UNIFORM_SLICE<T>*)this->slice;
    if(!slice || slice->mode==OPENGL_SLICE::NO_SLICE){
        int count=0;
        for(const auto&v:V.array) if(v!=VECTOR<T,3>()) count++;
        int inc=max(1,(int)ceil(pow<1,TV::m>((T)count/max_vectors_3d)));
        IV size=V.domain.Edge_Lengths();
        size=(size+1)/inc-1;
        for(RANGE_ITERATOR<TV::m> it(size);it.Valid();it.Next())
        {
            IV i=V.domain.min_corner+inc*it.index;
            TV v=V(i);
            if(v!=TV())
            {
                vector_field.Append(v);
                vector_locations.Append(grid.Node_X(i));
            }
        }
    }
    else
    {
        if(slice->mode==OPENGL_SLICE::CELL_SLICE)
            if(slice->index<V.domain.min_corner[slice->axis] || slice->index>=V.domain.max_corner[slice->axis])
                return;

        RANGE<IV> domain=V.domain;
        domain.min_corner(slice->axis)=slice->index;
        domain.max_corner(slice->axis)=slice->index+1;

        int idx=0;
        vector_field.Resize(domain.Size());
        vector_locations.Resize(domain.Size());
        for(RANGE_ITERATOR<TV::m> it(domain);it.Valid();it.Next()){
            vector_field(idx)=V(it.index);
            vector_locations(idx)=grid.Node_X(it.index);
            idx++;}
    }
}
//#####################################################################
// Function Print_Cell_Selection_Info
//#####################################################################
template<class T> void OPENGL_GRID_BASED_VECTOR_FIELD_3D<T>::
Print_Cell_Selection_Info(std::ostream& stream,const IV& cell) const
{
    if(V.Valid_Index(cell)) stream<<"velocity = "<<V(cell)<<std::endl;
    // TODO: interpolate to particles
}
//#####################################################################
// Function Print_Node_Selection_Info
//#####################################################################
template<class T> void OPENGL_GRID_BASED_VECTOR_FIELD_3D<T>::
Print_Node_Selection_Info(std::ostream& stream,const IV& node) const
{
    if(V.Valid_Index(node)) stream<<"velocity = "<<V(node)<<std::endl;
}
//#####################################################################
namespace PhysBAM{
template class OPENGL_GRID_BASED_VECTOR_FIELD_3D<float>;
template class OPENGL_GRID_BASED_VECTOR_FIELD_3D<double>;
}
