//#####################################################################
// Copyright 2004, Eran Guendelman.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class OPENGL_UNIFORM_SLICE
//##################################################################### 
#include <OpenGL/OpenGL/OPENGL_UNIFORM_SLICE.h>
#include <OpenGL/OpenGL/OPENGL_WORLD.h>
using namespace PhysBAM;
//#####################################################################
// Function Print_Slice_Info
//#####################################################################
template<class T> void OPENGL_UNIFORM_SLICE<T>::
Print_Slice_Info(std::ostream& output_stream)
{
    const char* axis_name[]={"","x","y","z"};
    if(mode==CELL_SLICE)
        output_stream<<"Slice: cell="<<index<<", "<<axis_name[axis]<<"="<<grid.Cell_X(IV()+index)[axis]<<std::endl;
}
//#####################################################################
// Function Update_Clip_Planes
//#####################################################################
template<class T> void OPENGL_UNIFORM_SLICE<T>::
Update_Clip_Planes()
{
    if(mode==NO_SLICE) {
        if(clip_plane_id1!=0) world.Remove_Clipping_Plane(clip_plane_id1); 
        if(clip_plane_id2!=0) world.Remove_Clipping_Plane(clip_plane_id2);
        clip_plane_id1=clip_plane_id2=0;
    }
    else {
        T pos=grid.Cell_X(IV()+index)[axis];
        PLANE<T> plane1(VECTOR<T,3>(0,0,0),VECTOR<T,3>(0,0,0)),plane2(VECTOR<T,3>(0,0,0),VECTOR<T,3>(0,0,0));
        plane1.normal[axis]=1;plane1.x0[axis]=pos-grid.dX[axis]/1.9;
        plane2.normal[axis]=-1;plane2.x0[axis]=pos+grid.dX[axis]/1.9;
        if(clip_plane_id1==0) clip_plane_id1=world.Add_Clipping_Plane(plane1);
        else world.Set_Clipping_Plane(clip_plane_id1,plane1);
        if(clip_plane_id2==0) clip_plane_id2=world.Add_Clipping_Plane(plane2);
        else world.Set_Clipping_Plane(clip_plane_id2,plane2);
    }
}
//#####################################################################
// Function Is_Slice_Mode
//#####################################################################
template<class T> bool OPENGL_UNIFORM_SLICE<T>::
Is_Slice_Mode()
{
    return mode==CELL_SLICE;
}
//#####################################################################
// Function Set_Slice_Mode
//#####################################################################
template<class T> void OPENGL_UNIFORM_SLICE<T>::
Set_Slice_Mode(OPENGL_SLICE::SLICE_MODE mode_input)
{
    mode=mode_input;
    if(Is_Slice_Mode()) {
        // keep current values of axis and slice number as long as they're in range
        axis=clamp(axis,0,2);
        index=clamp(index,0,Maximum_Slice_Index_In_Current_Mode());
    }
    Update_Clip_Planes();
}
//#####################################################################
// Function Toggle_Slice_Mode
//#####################################################################
template<class T> void OPENGL_UNIFORM_SLICE<T>::
Toggle_Slice_Mode()
{
    Set_Slice_Mode((OPENGL_SLICE::SLICE_MODE)(((int)mode+1)%3));
}
//#####################################################################
// Function Toggle_Slice_Axis
//#####################################################################
template<class T> void OPENGL_UNIFORM_SLICE<T>::
Toggle_Slice_Axis()
{
    if(Is_Slice_Mode()) {
        axis=(axis+1)%3;
        index=Maximum_Slice_Index_In_Current_Mode()/2+1;
        Update_Clip_Planes();
    }
}
//#####################################################################
// Function Increment_Slice
//#####################################################################
template<class T> void OPENGL_UNIFORM_SLICE<T>::
Increment_Slice()
{
    if(Is_Slice_Mode()) {
        index=min(index+1,Maximum_Slice_Index_In_Current_Mode());
        Update_Clip_Planes();
    }
}
//#####################################################################
// Function Decrement_Slice
//#####################################################################
template<class T> void OPENGL_UNIFORM_SLICE<T>::
Decrement_Slice()
{
    if(Is_Slice_Mode()) {
        index=max(index-1,0);
        Update_Clip_Planes();
    }
}
//#####################################################################
// Function Enable_Clip_Planes
//#####################################################################
template<class T> void OPENGL_UNIFORM_SLICE<T>::
Enable_Clip_Planes()
{
    glEnable(clip_plane_id1);
    glEnable(clip_plane_id2);
}
//#####################################################################
namespace PhysBAM{
template class OPENGL_UNIFORM_SLICE<double>;
template class OPENGL_UNIFORM_SLICE<float>;
}
