//#####################################################################
// Copyright 2022.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYREGHT.txt.
//#####################################################################

#include <Tools/Symbolics/EXECUTABLE_PROGRAM.h>
#include <Core/Data_Structures/HASHTABLE.h>
#include <Core/Vectors/VECTOR.h>
#include <Tools/Symbolics/CODE_BLOCK.h>
#include <Tools/Symbolics/CODE_BLOCK_NODE.h>
#include <Tools/Symbolics/INSTRUCTION.h>
#include <Tools/Symbolics/PROGRAM.h>
#include <Tools/Symbolics/PROGRAM_COMPILER.h>
#include <Tools/Symbolics/PROGRAM_CONTEXT.h>
#include <Tools/Symbolics/PROGRAM_PARSE_NODE.h>
#include <Tools/Symbolics/PROGRAM_SCALAR.h>
#include <map>
#include <string>
namespace PhysBAM{
//#####################################################################
// Constructor
//#####################################################################
template<class U> EXECUTABLE_PROGRAM<U>::
EXECUTABLE_PROGRAM()
    :prog(*new PROGRAM<T>),context(*new PROGRAM_CONTEXT<T>)
{
}
//#####################################################################
// Destructor
//#####################################################################
template<class U> EXECUTABLE_PROGRAM<U>::
~EXECUTABLE_PROGRAM()
{
    delete &prog;
    delete &context;
}
//#####################################################################
// Function Eval
//#####################################################################
template<class U> U EXECUTABLE_PROGRAM<U>::
Eval(const ARRAY<T>& in)
{
    context.data_in=in;
    prog.Execute(context);
    if constexpr(is_same<T,U>::value)
        return context.data_out(0);
    else if constexpr(is_same<VECTOR<T,U::m>,U>::value)
        return U(context.data_out);
    else PHYSBAM_FATAL_ERROR();
}
//#####################################################################
// Function Set
//#####################################################################
template<class U> void EXECUTABLE_PROGRAM<U>::
Set(const T_PROGRAM_SCALAR& ps)
{
    ARRAY_VIEW<const PROGRAM_SCALAR<T>> av;
    if constexpr(is_same<U,T>::value) av.Set(&ps,1);
    else if constexpr(is_same<VECTOR<T,U::m>,U>::value) av.Set(ps);
    else PHYSBAM_FATAL_ERROR();

    auto pc=*av(0).pc;
    int old_num_tmp=pc.num_tmp;
    for(int i=0;i<pc.var_out.m+pc.extra_out;i++)
        pc.Rename_Variable(i|mem_out,pc.num_tmp++);
    for(int i=0;i<av.m;i++)
    {
        int var=av(i).var;
        if((var&mem_mask)==mem_out)
            var=(var&~mem_mask)+old_num_tmp;
        pc.Append_Instruction(op_copy,i|mem_out,var,-1);
    }
    pc.var_out.Remove_All();
    pc.extra_out=av.m;
    pc.Optimize();
    pc.Finalize(prog);
    context.Initialize(prog);
}
template class EXECUTABLE_PROGRAM<float>;
template class EXECUTABLE_PROGRAM<VECTOR<float,3>>;
}
