//#####################################################################
// Copyright 2012.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYREGHT.txt.
//#####################################################################
#include <Tools/Symbolics/INSTRUCTION.h>
#include <Core/Arrays/ARRAY.h>
#include <Core/Log/LOG.h>
namespace PhysBAM{
int op_flags_table[op_last];
int Init_Instructions()
{
    op_flags_table[op_nop]=flag_none;
    op_flags_table[op_copy]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_add]=flag_reg_dest|flag_reg_src0|flag_reg_src1|flag_commute;
    op_flags_table[op_sub]=flag_reg_dest|flag_reg_src0|flag_reg_src1;
    op_flags_table[op_mul]=flag_reg_dest|flag_reg_src0|flag_reg_src1|flag_can_deduce|flag_commute;
    op_flags_table[op_div]=flag_reg_dest|flag_reg_src0|flag_reg_src1|flag_can_deduce;
    op_flags_table[op_mod]=flag_reg_dest|flag_reg_src0|flag_reg_src1|flag_can_deduce;
    op_flags_table[op_neg]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_inv]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_sqrt]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_exp]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_ln]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_pow]=flag_reg_dest|flag_reg_src0|flag_reg_src1|flag_can_deduce;
    op_flags_table[op_sin]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_cos]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_asin]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_acos]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_atan]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_atan2]=flag_reg_dest|flag_reg_src0|flag_reg_src1;
    op_flags_table[op_floor]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_ceil]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_rint]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_diff]=flag_reg_dest|flag_reg_src0|flag_reg_src1;
    op_flags_table[op_lt]=flag_reg_dest|flag_reg_src0|flag_reg_src1;
    op_flags_table[op_le]=flag_reg_dest|flag_reg_src0|flag_reg_src1;
    op_flags_table[op_gt]=flag_reg_dest|flag_reg_src0|flag_reg_src1;
    op_flags_table[op_ge]=flag_reg_dest|flag_reg_src0|flag_reg_src1;
    op_flags_table[op_eq]=flag_reg_dest|flag_reg_src0|flag_reg_src1|flag_commute;
    op_flags_table[op_ne]=flag_reg_dest|flag_reg_src0|flag_reg_src1|flag_commute;
    op_flags_table[op_not]=flag_reg_dest|flag_reg_src0;
    op_flags_table[op_or]=flag_reg_dest|flag_reg_src0|flag_reg_src1|flag_can_deduce|flag_commute;
    op_flags_table[op_and]=flag_reg_dest|flag_reg_src0|flag_reg_src1|flag_can_deduce|flag_commute;
    op_flags_table[op_br_z]=flag_reg_src0;
    op_flags_table[op_br_nz]=flag_reg_src0;
    op_flags_table[op_jmp]=flag_none;
    op_flags_table[op_label]=flag_none;
    op_flags_table[op_phi]=flag_reg_dest|flag_reg_src0|flag_reg_src1;
    return 0;
}
const char* messages[op_last]={
    "nop\n",
    "copy  %c%d, %c%d\n",
    "add   %c%d, %c%d, %c%d\n",
    "sub   %c%d, %c%d, %c%d\n",
    "mul   %c%d, %c%d, %c%d\n",
    "div   %c%d, %c%d, %c%d\n",
    "mod   %c%d, %c%d, %c%d\n",
    "neg   %c%d, %c%d\n",
    "inv   %c%d, %c%d\n",
    "sqrt  %c%d, %c%d\n",
    "exp   %c%d, %c%d\n",
    "ln    %c%d, %c%d\n",
    "pow   %c%d, %c%d, %c%d\n",
    "sin   %c%d, %c%d\n",
    "cos   %c%d, %c%d\n",
    "asin  %c%d, %c%d\n",
    "acos  %c%d, %c%d\n",
    "atan  %c%d, %c%d\n",
    "atan2 %c%d, %c%d, %c%d\n",
    "floor %c%d, %c%d\n",
    "ceil  %c%d, %c%d\n",
    "rint  %c%d, %c%d\n",
    "lt    %c%d, %c%d, %c%d\n",
    "le    %c%d, %c%d, %c%d\n",
    "gt    %c%d, %c%d, %c%d\n",
    "ge    %c%d, %c%d, %c%d\n",
    "eq    %c%d, %c%d, %c%d\n",
    "ne    %c%d, %c%d, %c%d\n",
    "not   %c%d, %c%d\n",
    "or    %c%d, %c%d, %c%d\n",
    "and   %c%d, %c%d, %c%d\n",
    "br_z  L%.0s%d, %c%d\n",
    "br_nz L%.0s%d, %c%d\n",
    "jmp   L%.0s%d\n",
    "label L%.0s%d\n",
    "phi   %c%d, %c%d, %c%d\n",
    "diff  %c%d, %c%d, %c%d\n"
};

//#####################################################################
// Function Print
//#####################################################################
void Print(const INSTRUCTION& I)
{
    printf(messages[I.type],"rcio"[I.dest>>mem_shift],I.dest&~mem_mask,
        "rcio"[I.src0>>mem_shift],I.src0&~mem_mask,
        "rcio"[I.src1>>mem_shift],I.src1&~mem_mask);
}
//#####################################################################
// Function Print
//#####################################################################
void Print(const ARRAY<INSTRUCTION>& code)
{
    for(int i=0;i<code.m;i++){
        printf("% 3d  ",i);
        Print(code(i));}
}
}
