//#####################################################################
// Copyright 2012.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYREGHT.txt.
//#####################################################################
#ifndef __PROGRAM_DEFINITION__
#define __PROGRAM_DEFINITION__
#include <Tools/Symbolics/PROGRAM.h>
#include <Tools/Symbolics/INSTRUCTION.h>
#include <Tools/Symbolics/PROGRAM_COMPILER.h>
#include <Tools/Symbolics/PROGRAM_CONTEXT.h>
#include <cmath>

namespace PhysBAM{
//#####################################################################
// Function Execute
//#####################################################################
template<class T> T
Evaluate_Op(int type,T in0,T in1)
{
    switch(type){
        case op_copy: return in0;
        case op_add: return in0+in1;
        case op_sub: return in0-in1;
        case op_mul: return in0*in1;
        case op_div: return in0/in1;
        case op_mod:{T x=fmod(in0,in1);return x<0?x+abs(in1):x;}
        case op_neg: return -in0;
        case op_inv: return (T)1/in0;
        case op_sqrt: return sqrt(in0);
        case op_exp: return exp(in0);
        case op_ln: return log(in0);
        case op_pow: return pow(in0,in1);
        case op_sin: return sin(in0);
        case op_cos: return cos(in0);
        case op_asin: return asin(in0);
        case op_acos: return acos(in0);
        case op_atan: return atan(in0);
        case op_atan2: return atan2(in0,in1);
        case op_floor: return floor(in0);
        case op_ceil: return ceil(in0);
        case op_rint: return rint(in0);
        case op_lt: return in0<in1;
        case op_le: return in0<=in1;
        case op_gt: return in0>in1;
        case op_ge: return in0>=in1;
        case op_eq: return in0==in1;
        case op_ne: return in0!=in1;
        case op_not: return !in0;
        case op_or: return in0||in1;
        case op_and: return in0&&in1;
        default: PHYSBAM_FATAL_ERROR("Cannot evaluate this instruction");}
}
//#####################################################################
// Function Execute
//#####################################################################
template<class T> void PROGRAM<T>::
Execute_Op(ARRAY<T>& reg,int& ip) const
{
    const INSTRUCTION& o=code(ip);
    if(op_flags_table[o.type]&flag_reg_dest){
        T in0=0,in1=0;
        if(op_flags_table[o.type]&flag_reg_src0) in0=reg(o.src0);
        if(op_flags_table[o.type]&flag_reg_src1) in1=reg(o.src1);
        reg(o.dest)=Evaluate_Op(o.type,in0,in1);
        return;}

    switch(o.type){
        case op_nop: case op_label: break;
        case op_br_z: if(!reg(o.src0)) ip=o.dest-1;break;
        case op_br_nz: if(!!reg(o.src0)) ip=o.dest-1;break;
        case op_jmp: ip=o.dest-1;break;
        default: PHYSBAM_FATAL_ERROR("Missing instruction");}
}
//#####################################################################
// Function Execute
//#####################################################################
template<class T> void PROGRAM<T>::
Execute(ARRAY<T>& reg) const
{
    for(int ip=0;ip<code.m;ip++)
        Execute_Op(reg,ip);
}
//#####################################################################
// Function Execute
//#####################################################################
template<class T> void PROGRAM<T>::
Execute(PROGRAM_CONTEXT<T>& context) const
{
    Execute(context.reg);
}
//#####################################################################
// Function Print
//#####################################################################
template<class T> void PROGRAM<T>::
Print() const
{
    PhysBAM::Print(code);
}
//#####################################################################
// Function Simple_Compile
//#####################################################################
template<class T> void
Simple_Compile(PROGRAM<T>& prog,const std::string& code,
    const ARRAY<std::string>& var_in,const ARRAY<std::string>& var_out)
{
    PROGRAM_COMPILER<T> pc;
    pc.var_out=var_out;
    pc.var_in=var_in;
    pc.Parse(code.c_str(),false);
    pc.Optimize();
    pc.Finalize(prog);
}
}
#endif
