//#####################################################################
// Copyright 2012.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYREGHT.txt.
//#####################################################################
#include <Tools/Symbolics/PROGRAM_SCALAR.h>
#include <Core/Vectors/VECTOR.h>
#include <Tools/Symbolics/CODE_BLOCK.h>
#include <Tools/Symbolics/INSTRUCTION.h>
#include <Tools/Symbolics/PROGRAM_COMPILER.h>

namespace PhysBAM{
namespace NS_PROGRAM_SCALAR
{
template<class T> PROGRAM_SCALAR<T>
PROGRAM_SCALAR<T>::Number(PROGRAM_COMPILER<T>& pc,T a)
{
    return {&pc,pc.Add_Constant(a)};
}

template<class T> PROGRAM_SCALAR<T>
operator+(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_copy,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
operator-(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_neg,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
operator+(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b)
{
    return {a.pc,a.pc->Append_Instruction(op_add,a.pc->num_tmp++,a.var,b.var)};
}

template<class T> PROGRAM_SCALAR<T>
operator-(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b)
{
    return {a.pc,a.pc->Append_Instruction(op_sub,a.pc->num_tmp++,a.var,b.var)};
}

template<class T> PROGRAM_SCALAR<T>
operator*(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b)
{
    return {a.pc,a.pc->Append_Instruction(op_mul,a.pc->num_tmp++,a.var,b.var)};
}

template<class T> PROGRAM_SCALAR<T>
operator/(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b)
{
    return {a.pc,a.pc->Append_Instruction(op_div,a.pc->num_tmp++,a.var,b.var)};
}

template<class T> PROGRAM_SCALAR<T>
operator<(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b)
{
    return {a.pc,a.pc->Append_Instruction(op_lt,a.pc->num_tmp++,a.var,b.var)};
}

template<class T> PROGRAM_SCALAR<T>
operator>(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b)
{
    return {a.pc,a.pc->Append_Instruction(op_gt,a.pc->num_tmp++,a.var,b.var)};
}

template<class T> PROGRAM_SCALAR<T>
operator<=(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b)
{
    return {a.pc,a.pc->Append_Instruction(op_le,a.pc->num_tmp++,a.var,b.var)};
}

template<class T> PROGRAM_SCALAR<T>
operator>=(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b)
{
    return {a.pc,a.pc->Append_Instruction(op_ge,a.pc->num_tmp++,a.var,b.var)};
}

template<class T> PROGRAM_SCALAR<T>
operator==(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b)
{
    return {a.pc,a.pc->Append_Instruction(op_eq,a.pc->num_tmp++,a.var,b.var)};
}

template<class T> PROGRAM_SCALAR<T>
operator!=(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b)
{
    return {a.pc,a.pc->Append_Instruction(op_ne,a.pc->num_tmp++,a.var,b.var)};
}

template<class T> PROGRAM_SCALAR<T>
op_if(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b,const PROGRAM_SCALAR<T>& c)
{
    int result=a.pc->num_tmp++;
    CODE_BLOCK* A=a.pc->code_blocks.Last();
    CODE_BLOCK* B=new CODE_BLOCK;
    B->id=a.pc->code_blocks.Append(B);
    a.pc->Append_Instruction(op_copy,result,b.var,-1);
    CODE_BLOCK* C=new CODE_BLOCK;
    C->id=a.pc->code_blocks.Append(C);
    a.pc->Append_Instruction(op_copy,result,c.var,-1);
    CODE_BLOCK* D=new CODE_BLOCK;
    D->id=a.pc->code_blocks.Append(D);
    INSTRUCTION in_A={op_br_z,C->id,a.var,B->id};
    A->Append(in_A);

    A->next[0]=B;
    B->prev[0]=A;
    A->next[1]=C;
    C->prev[0]=A;

    D->prev[0]=B;
    D->prev[1]=C;
    B->next[0]=D;
    C->next[0]=D;
    return {a.pc,result};
}

template<class T> PROGRAM_SCALAR<T>
abs(const PROGRAM_SCALAR<T>& a)
{
    return op_if(a<a.Number(*a.pc,0),-a,a);
}

template<class T> PROGRAM_SCALAR<T>
sqrt(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_sqrt,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
exp(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_exp,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
log(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_ln,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
sin(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_sin,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
cos(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_cos,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
tan(const PROGRAM_SCALAR<T>& a)
{
    return sin(a)/cos(a);
}

template<class T> PROGRAM_SCALAR<T>
asin(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_asin,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
acos(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_acos,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
atan(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_atan,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
floor(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_floor,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
ceil(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_ceil,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
rint(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_rint,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
Inverse(const PROGRAM_SCALAR<T>& a)
{
    return {a.pc,a.pc->Append_Instruction(op_inv,a.pc->num_tmp++,a.var,-1)};
}

template<class T> PROGRAM_SCALAR<T>
atan2(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b)
{
    return {a.pc,a.pc->Append_Instruction(op_atan2,a.pc->num_tmp++,a.var,b.var)};
}

template<class T> PROGRAM_SCALAR<T>
hypot(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b)
{
    return sqrt(a*a+b*b);
}

template<class T> PROGRAM_SCALAR<T>
pow(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b)
{
    return {a.pc,a.pc->Append_Instruction(op_pow,a.pc->num_tmp++,a.var,b.var)};
}

template<class T> std::ostream&
operator<<(std::ostream& o, const PROGRAM_SCALAR<T>& s)
{
    PHYSBAM_FATAL_ERROR();
}

template<class T> std::istream&
operator>>(std::istream& i, PROGRAM_SCALAR<T>& s)
{
    PHYSBAM_FATAL_ERROR();
}

template<class T> PROGRAM_SCALAR<T>
Magnitude_Squared(const PROGRAM_SCALAR<T>& s)
{
    return s*s;
}

template<class T> PROGRAM_SCALAR<T>
Dot_Product(const PROGRAM_SCALAR<T>& a1,const PROGRAM_SCALAR<T>& a2)
{
    return a1*a2;
}

template<class T> PROGRAM_SCALAR<T>
wrap(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b,const PROGRAM_SCALAR<T>& c)
{
    return op_if(a<b,b,op_if(a>c,c,a));
}

template<class T,int d> VECTOR<PROGRAM_SCALAR<T>,d>
Gradient(const PROGRAM_SCALAR<T>& a,const VECTOR<int,d>& vars)
{
    VECTOR<PROGRAM_SCALAR<T>,d> r;
    for(int i=0;i<d;i++)
        r(i)={a.pc,a.pc->Diff(a.var,vars(i))};
    return r;
}

template PROGRAM_SCALAR<float> operator+(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> operator-(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> operator+(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b);
template PROGRAM_SCALAR<float> operator-(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b);
template PROGRAM_SCALAR<float> operator*(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b);
template PROGRAM_SCALAR<float> operator/(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b);
template PROGRAM_SCALAR<float> operator<(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b);
template PROGRAM_SCALAR<float> operator>(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b);
template PROGRAM_SCALAR<float> operator<=(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b);
template PROGRAM_SCALAR<float> operator>=(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b);
template PROGRAM_SCALAR<float> operator==(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b);
template PROGRAM_SCALAR<float> operator!=(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b);
template PROGRAM_SCALAR<float> op_if(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b,const PROGRAM_SCALAR<float>& c);
template PROGRAM_SCALAR<float> abs(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> sqrt(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> exp(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> log(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> sin(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> cos(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> tan(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> asin(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> acos(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> atan(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> floor(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> ceil(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> rint(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> Inverse(const PROGRAM_SCALAR<float>& a);
template PROGRAM_SCALAR<float> atan2(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b);
template PROGRAM_SCALAR<float> hypot(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b);
template PROGRAM_SCALAR<float> pow(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b);
template std::ostream& operator<<(std::ostream& o, const PROGRAM_SCALAR<float>& s);
template std::istream& operator>>(std::istream& i, PROGRAM_SCALAR<float>& s);
template PROGRAM_SCALAR<float> Magnitude_Squared(const PROGRAM_SCALAR<float>& s);
template PROGRAM_SCALAR<float> Dot_Product(const PROGRAM_SCALAR<float>& a1,const PROGRAM_SCALAR<float>& a2);
template PROGRAM_SCALAR<float> wrap(const PROGRAM_SCALAR<float>& a,const PROGRAM_SCALAR<float>& b,const PROGRAM_SCALAR<float>& c);
template VECTOR<PROGRAM_SCALAR<float>,1> Gradient(const PROGRAM_SCALAR<float>& a,const VECTOR<int,1>& vars);
template VECTOR<PROGRAM_SCALAR<float>,2> Gradient(const PROGRAM_SCALAR<float>& a,const VECTOR<int,2>& vars);
template VECTOR<PROGRAM_SCALAR<float>,3> Gradient(const PROGRAM_SCALAR<float>& a,const VECTOR<int,3>& vars);

template PROGRAM_SCALAR<double> operator+(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> operator-(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> operator+(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b);
template PROGRAM_SCALAR<double> operator-(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b);
template PROGRAM_SCALAR<double> operator*(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b);
template PROGRAM_SCALAR<double> operator/(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b);
template PROGRAM_SCALAR<double> operator<(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b);
template PROGRAM_SCALAR<double> operator>(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b);
template PROGRAM_SCALAR<double> operator<=(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b);
template PROGRAM_SCALAR<double> operator>=(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b);
template PROGRAM_SCALAR<double> operator==(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b);
template PROGRAM_SCALAR<double> operator!=(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b);
template PROGRAM_SCALAR<double> op_if(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b,const PROGRAM_SCALAR<double>& c);
template PROGRAM_SCALAR<double> abs(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> sqrt(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> exp(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> log(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> sin(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> cos(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> tan(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> asin(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> acos(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> atan(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> floor(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> ceil(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> rint(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> Inverse(const PROGRAM_SCALAR<double>& a);
template PROGRAM_SCALAR<double> atan2(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b);
template PROGRAM_SCALAR<double> hypot(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b);
template PROGRAM_SCALAR<double> pow(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b);
template std::ostream& operator<<(std::ostream& o, const PROGRAM_SCALAR<double>& s);
template std::istream& operator>>(std::istream& i, PROGRAM_SCALAR<double>& s);
template PROGRAM_SCALAR<double> Magnitude_Squared(const PROGRAM_SCALAR<double>& s);
template PROGRAM_SCALAR<double> Dot_Product(const PROGRAM_SCALAR<double>& a1,const PROGRAM_SCALAR<double>& a2);
template PROGRAM_SCALAR<double> wrap(const PROGRAM_SCALAR<double>& a,const PROGRAM_SCALAR<double>& b,const PROGRAM_SCALAR<double>& c);
template VECTOR<PROGRAM_SCALAR<double>,1> Gradient(const PROGRAM_SCALAR<double>& a,const VECTOR<int,1>& vars);
template VECTOR<PROGRAM_SCALAR<double>,2> Gradient(const PROGRAM_SCALAR<double>& a,const VECTOR<int,2>& vars);
template VECTOR<PROGRAM_SCALAR<double>,3> Gradient(const PROGRAM_SCALAR<double>& a,const VECTOR<int,3>& vars);
}
}
