//#####################################################################
// Copyright 2012.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYREGHT.txt.
//#####################################################################
// Class PROGRAM_SCALAR
//#####################################################################
#ifndef __PROGRAM_SCALAR__
#define __PROGRAM_SCALAR__

#include <Core/Math_Tools/cube.h>
#include <Core/Math_Tools/Inverse.h>
#include <Core/Math_Tools/sqr.h>
#include <Core/Vectors/SCALAR_POLICY.h>
#include <cmath>
#include <iosfwd>

using ::std::sin;
using ::std::cos;
using ::std::tan;
using ::std::exp;
using ::std::log;
using ::std::abs;
using ::std::sqrt;
using ::PhysBAM::sqr;
using ::PhysBAM::cube;
using ::std::hypot;
namespace PhysBAM{

template<class T> class PROGRAM_COMPILER;

namespace NS_PROGRAM_SCALAR
{
using ::std::sin;
using ::std::cos;
using ::std::tan;
using ::std::asin;
using ::std::acos;
using ::std::atan;
using ::std::atan2;
using ::std::exp;
using ::std::log;
using ::std::abs;
using ::std::pow;
using ::std::sqrt;
using ::PhysBAM::sqr;
using ::PhysBAM::cube;
using ::PhysBAM::Inverse;
using ::std::hypot;
using ::std::floor;
using ::std::ceil;
using ::std::rint;

template<class T>
class PROGRAM_SCALAR
{
public:
    PROGRAM_COMPILER<T>* pc=nullptr;
    int var=0;

    static PROGRAM_SCALAR Number(PROGRAM_COMPILER<T>& pc,T a);
};

template<class T> PROGRAM_SCALAR<T> operator+(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> operator-(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> operator+(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b);
template<class T> PROGRAM_SCALAR<T> operator-(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b);
template<class T> PROGRAM_SCALAR<T> operator*(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b);
template<class T> PROGRAM_SCALAR<T> operator/(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b);

template<class T> PROGRAM_SCALAR<T>& operator+= (PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b) {return a=a+b;}
template<class T> PROGRAM_SCALAR<T>& operator-= (PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b) {return a=a-b;}
template<class T> PROGRAM_SCALAR<T>& operator*= (PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b) {return a=a*b;}
template<class T> PROGRAM_SCALAR<T>& operator/= (PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b) {return a=a/b;}

template<class T> PROGRAM_SCALAR<T> operator<(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b);
template<class T> PROGRAM_SCALAR<T> operator>(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b);
template<class T> PROGRAM_SCALAR<T> operator<=(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b);
template<class T> PROGRAM_SCALAR<T> operator>=(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b);
template<class T> PROGRAM_SCALAR<T> operator==(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b);
template<class T> PROGRAM_SCALAR<T> operator!=(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b);
template<class T> PROGRAM_SCALAR<T> op_if(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b,const PROGRAM_SCALAR<T>& c);

template<class T> PROGRAM_SCALAR<T> abs(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> sqrt(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> exp(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> log(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> sin(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> cos(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> tan(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> asin(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> acos(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> atan(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> floor(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> ceil(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> rint(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> Inverse(const PROGRAM_SCALAR<T>& a);
template<class T> PROGRAM_SCALAR<T> atan2(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b);
template<class T> PROGRAM_SCALAR<T> hypot(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b);
template<class T> PROGRAM_SCALAR<T> pow(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b);

template<class T> std::ostream& operator<<(std::ostream& o, const PROGRAM_SCALAR<T>& s);
template<class T> std::istream& operator>>(std::istream& i, PROGRAM_SCALAR<T>& s);

template<class T> PROGRAM_SCALAR<T> Magnitude_Squared(const PROGRAM_SCALAR<T>& s);
template<class T> PROGRAM_SCALAR<T> Dot_Product(const PROGRAM_SCALAR<T>& a1,const PROGRAM_SCALAR<T>& a2);
template<class T> PROGRAM_SCALAR<T> wrap(const PROGRAM_SCALAR<T>& a,const PROGRAM_SCALAR<T>& b,const PROGRAM_SCALAR<T>& c);

template<class T,int d> VECTOR<PROGRAM_SCALAR<T>,d>
Gradient(const PROGRAM_SCALAR<T>& a,const VECTOR<int,d>& vars);
}
using NS_PROGRAM_SCALAR::PROGRAM_SCALAR;

template<class T> struct SCALAR_POLICY<PROGRAM_SCALAR<T>> {typedef PROGRAM_SCALAR<T> TYPE;};
}
#endif
